%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           pxc/alpha-cmac.sl
% Title:          Patterns and predicates for DEC Alpha PSL cmacro expansion
% Author:         Winfried Neun
% Created:        December 24, 1992
% Modified:
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions

(on fast-integers)

(compiletime
  (progn (on eolinstringok)
     (dm * (u)  nil)
     (setq *syslisp t)
     nil))

(fluid '(AddressingUnitsPerItem
     CharactersPerWord
     StackDirection
     *ImmediateQuote
     AddressingUnitsPerFunctionCell bigcode!*))

(loadtime (off ImmediateQuote))

(flag '(Immediate UnImmediate Indirect displacement predecrement
    label extrareg postincrement Indexed absolute deferred reglist)
    'TerminalOperand)

% MkItem may be used when evaluating WConst expressions.

(de MkItem (TagPart InfPart)
  (lor (lsh TagPart 56) (land InfPart 16#ffffffffffffff)))

(de freep (asize) (and (eq 0 asize) (not &links-present&)))

(de evenp (x) (and (fixp x) (eq 0 (land x 1))))

(de indexedp (x) (eqcar x 'indexed))

(de DeferrableP (Expression)
  (and (PairP Expression)
       (not (memq (first Expression)
	      '(Deferred $Fluid $Global Fluid Global)))))

(de RegisterDeferredP (Expression)
  (and (EqCar Expression 'Deferred) (EqCar (second Expression) 'Reg)))

(de FakeRegisterNumberP (Expression)
  (and (IntP Expression) (GreaterP Expression 5)))

% NegativeImmediateP is for numbers whose negative is a one byte operand.

(de negp (value) (and (numberp value) (lessp value 0)))

(de displacementp (x) (and (pairp x) (eq (car x) 'displacement)))

(de NegativeImmediateP (Expression)
  (and (FixP Expression) (minusp Expression) (GreaterP Expression -64)))

% expansions of addressing modes.

(DefAnyreg CAR  ZIBanyregCAR) 

(de ZIBAnyregCAR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. caar
	       (member (cadr source) *nextreglist*))
	     (setq areg (cadr source))
	     (setq areg (nexttempreg)))
      (if (regp source)  % its a reg or a frame reg
	  (push `(zap ,source 16#80 ,areg) Resultingcode*)

	  (push `(*move ,source ,areg) Resultingcode*)
	  (push `(zap ,areg 16#80 ,areg) Resultingcode*))
      (return `(displacement ,areg 0))))

(DefAnyreg CDR  ZIBanyregCDR)  % ZIB

(de ZIBAnyregCDR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. cddar
	       (member (cadr source) *nextreglist*))
	     (setq areg (cadr source))
	     (setq areg (nexttempreg)))
      (if (regp source)  
	  (push `(zap ,source 16#80 ,areg) Resultingcode*)

	  (push `(*move ,source , areg) Resultingcode*)
	  (push `(zap ,areg 16#80 ,areg) Resultingcode*))
      (return `(displacement ,areg 8))))

(fluid '(*nextreg* *nextreglist*))

(setq *nextreg* 0        % Four registers should be sufficient
      *nextreglist* '((reg t2) (reg t3) (reg t4) ))

(de nexttempreg (x)   % ZIB  :This saves Regs , non static assignment
     (setq *nextreg* (wplus2 *nextreg* 1))
     (when (eq *nextreg* 4) (setq *nextreg* 1))
     (nth *nextreglist* *nextreg*))

(DefAnyreg Memory  ZIBanyregMemory)  % ZIB

(de ZIBAnyregMemory (reg source offset)
  (prog (aReg cadrsource)
   (setq source (resolveoperand reg source))
   (setq offset (resolveoperand reg offset))
   (cond ((and (registerp source) (intp offset))
      (return (if (eq 0 offset) `(Deferred ,source)
		   (if (alpha-disp-p offset)
			    `(displacement ,source ,offset)
			(progn
			  (setq aReg (nexttempreg))
			  (push `(*move ,offset ,aReg) Resultingcode*)
			  (push `(*wplus2 ,areg ,source)  Resultingcode*)
			 `(displacement ,aReg 0) )))))
      ((and (RegisterDeferredP source) (ZeroP offset))
	 (return `(deferred (Displacement (UnDeferred ,SOURCE) 0)))))
    (setq aReg (if (and (displacementp source)
		    (member (setq cadrsource (cadr source))
				 *nextreglist*))
	       cadrsource      (nexttempreg))) % take a new TempReg
    (push `(*move ,source ,areg) Resultingcode*)
    (push `(*wplus2 ,areg ,offset) Resultingcode*)
    (return `(displacement ,aReg 0))))


(DefAnyreg QUOTE
       AnyregQUOTE
       ((Null) (REG NIL))
       ((EqTP) (reg T))
       ((InumP) SOURCE)
       (       (QUOTE SOURCE)))

(de compiler_evenp (x) (and (fixp x) (eq 0 (land (quotient x 8) 1))))


(DefAnyreg FRAME ZIBAnyregFrame)

(de ZIBAnyregFRAME (dum number)
     (if *usefastframe (nth '((reg ff1) (reg ff2) (reg ff3) (reg ff4)) number)
	`(displacement (reg st) ,(plus 8 (times2 8 number)))))

(DefAnyreg xFRAME ZIBAnyregxFrame)

(de ZIBAnyregxFRAME (dum number)   % regardless of *usefastframe
    `(displacement (reg st) ,(plus 8 (times2 8 number))))

(DefAnyreg REG ZIBAnyregREG)

(de ZIBanyregReg (reg source)
   (cond ((Fakeregisternumberp source) % argumentblock lies just below SYMVAL
	     `(displacement (reg symvalcenter) ,(wtimes2 (wplus2 source -17) 8)))
	 (t   `(REG ,SOURCE))))

(loadtime (remflag '(fluid global $fluid $global) 'Terminaloperand))

(defanyreg Fluid ZIBanyregfluid)

(defanyreg $Fluid ZIBanyregfluid)

(de ZIBanyregFluid (reg source)
   (cond ((and (idp source) (get source 'registername))
			 `(reg ,(get source 'registername)))
	 (t       `($Fluid ,SOURCE))))

(defanyreg Global ZIBanyregglobal)

(defanyreg $Global ZIBanyregglobal)

(de ZIBanyregGlobal (reg source)
   (cond ((and (idp source) (get source 'registername))
			 `(reg ,(get source 'registername)))
	 (t       `($global ,SOURCE))))

%-----------------------------------------------------------------------------

(de kernelCallableP(u)(flagp u 'kernel))

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (if (FlagP FunctionName 'ForeignFunction)
      `((*ForeignLink ,FunctionName ,FunctionType ,NumberOfArguments)
	(*Exit 0))
      `((*DeAlloc ,DeAllocCount)
	(*JCall ,FunctionName))))

(DefCMacro *Call
       (((lambda (v) (and (InternallyCallableP v) (asmp))))
				(la (Reg r27) (InternalEntry ArgOne) )
				(jsr (reg returnaddr)
					 (displacement (Reg r27) nil))
				(ldq (reg r27) (displacement (reg st) 8)))
       ((InternallyCallableP) (lda (Reg r27) (InternalEntry ArgOne) )
				(jsr (reg returnaddr)
					 (displacement (Reg r27) nil))
				(ldq (reg r27) (displacement (reg st) 8)))
       ((asmp)  (lda (Reg t3) (idloc ArgOne))     
		 (s8addq  (reg t3) (reg SYMFNCcenter) (reg r27))
		 (ldq (reg r27) (displacement (reg r27) 0))
		 (jsr (reg returnaddr) (displacement (reg r27) nil)))
       (((lambda (v) (and (kernelCallableP v) bigcode!*)))
                 (ldq (reg r27) (displacement   (reg SYMFNCcenter)
                                            (xidloc ArgOne)))
                 (jsr (reg returnaddr) (displacement (reg r27) nil))
		 (lda (reg r27) (displacement (reg returnaddr) nearestentry*))
		 (ldah (reg r27) (displacement (reg r27) nearestentry*hi))
                 templabel )          
       ((kernelCallableP)
                 (ldq (reg r27) (displacement   (reg SYMFNCcenter)
                                            (xidloc ArgOne)))
                 (jsr (reg returnaddr) (displacement (reg r27) nil))
                 (lda (reg r27) (displacement (reg returnaddr) nearestentry*))
                 templabel )
        (((lambda (v) bigcode!*))
	         (lda (Reg t3) (idloc ArgOne))
		 (s8addq  (reg t3) (reg SYMFNCcenter) (reg r27))
                 (ldq (reg r27) (displacement (reg r27) 0))
                 (jsr (reg returnaddr) (displacement (reg r27) nil))
                 (lda (reg r27) (displacement (reg returnaddr) nearestentry*))
		 (ldah (reg r27) (displacement (reg r27) nearestentry*hi))
                 templabel )
	(       (lda (Reg t3) (idloc ArgOne))
		 (s8addq  (reg t3) (reg SYMFNCcenter) (reg r27))
		 (ldq (reg r27) (displacement (reg r27) 0))
		 (jsr (reg returnaddr) (displacement (reg r27) nil))
		 (lda (reg r27) (displacement (reg returnaddr) nearestentry*))
                 templabel ))

(DefCMacro *JCall
       (((lambda (v) (and (InternallyCallableP v) (asmp))))
			(la (Reg r27) (InternalEntry ArgOne))
                        (jmp (reg null) (displacement (reg r27) nil)))
       ((InternallyCallableP)  (lda (Reg r27) (InternalEntry ArgOne))
				(jmp (reg null) (displacement (reg r27) nil)))
	((asmp) (*move (idloc ArgOne) (Reg t3))     
		 (s8addq  (reg t3) (reg SYMFNCcenter) (reg r27))
		 (ldq (reg r27) (displacement (reg r27) 0))
		 (jmp (reg null) (displacement (reg r27) nil)))

       ((kernelCallableP)
                 (ldq (reg r27) (displacement   (reg SYMFNCcenter)
                                            (xidloc ArgOne)))
                 (jmp (reg null) (displacement (reg r27) nil)))

	(       (lda (Reg t3) (idloc ArgOne) )     
		 (s8addq  (reg t3) (reg SYMFNCcenter) (reg r27))
		 (ldq (reg r27) (displacement (reg r27) 0))
		 (jmp (reg null) (displacement (reg r27) nil))))

% At Alloc and Dealloc we have to add room for the returnaddr register, since
% this must be saved by the user

(de *ALLOC (framesize)

    (setq NAlloc!* framesize)
    (cond ((freep (wtimes2 AddressingUnitsPerItem framesize)) NIL)
	  (t  (when (eq 1 (wand framesize 1)) % set st on double word boundary
			 (setq framesize (iadd1 framesize)))
	      (if (asmp)
	      `((lda (reg st) 
		    (displacement (reg st) 
			 ,(minus (plus2 16 (times2 8 framesize))))) 
		(stq (reg returnaddr) (displacement (reg st) 0)))
	      `((lda (reg st) 
		    (displacement (reg st) 
			 ,(minus (plus2 16 (times2 8 framesize))))) 
		(stq (reg r27) (displacement (reg st) 8))
		(stq (reg returnaddr) (displacement (reg st) 0)))))
)   )

(defcmacro *ALLOC)

(de *xalloc (framesize)      % alloc in any case, regardless of freep
    (when (eq 1 (wand framesize 1)) % set st on double word boundary
                         (setq framesize (iadd1 framesize)))
    (setq framesize (wtimes2 8 framesize))
    `((lda (reg st) (displacement (reg st) ,(minus (plus2 16 framesize))))
      (stq (reg r27) (displacement (reg st) 8))
      (stq (reg returnaddr) (displacement (reg st) 0))))

(defcmacro *xALLOC)

(DefCMacro *DeAlloc
	   ((FreeP))
	   ((compiler_evenp) 
		    (ldq (reg returnaddr) (displacement (reg st) 0))
		    (lda (reg st) (displacement (reg st) (plus2 16 ArgOne))))
	   (    (ldq (reg returnaddr) (displacement (reg st) 0))
		(lda (reg st) (displacement (reg st) (plus2 24 ArgOne)))))

(de *xDeAlloc (framesize)
	    (when (eq 1 (wand framesize 1)) % set st on double word boundary
			 (setq framesize (iadd1 framesize)))
	   `((ldq (reg returnaddr) (displacement (reg st) 0))
	     (lda (reg st) (displacement (reg st)
			  (plus2 16 ,(wtimes2 8 framesize))))))

(DefCMacro *xDeAlloc)

(de asmp () *writingasmfile)

(DefCMacro *Exit
	 % leaf routine first
	((Freep)  (ret (reg null) (displacement (reg returnaddr) nil)))
	(((lambda (x) (and (asmp) (compiler_evenp x))))
		(ldq (reg returnaddr) (displacement (reg st) 0))
		(lda (reg st) (displacement (reg st) (plus2 16 ArgOne)))
		(ret  (reg null) (displacement (reg returnaddr) nil)))
	((asmp) (ldq (reg returnaddr) (displacement (reg st) 0)) 
		(lda (reg st) (displacement (reg st) (plus2 24 ArgOne))) 
		(ret (reg null)  (displacement (reg returnaddr) nil)))
	((compiler_evenp)
		(ldq (reg returnaddr) (displacement (reg st) 0))
		(lda (reg st) (displacement (reg st) (plus2 16 ArgOne)))
		(ret (reg null)  (displacement (reg returnaddr) nil)))
	(       (ldq (reg returnaddr) (displacement (reg st) 0))
		(lda (reg st) (displacement (reg st) (plus2 24 ArgOne)))  
		(ret (reg null)  (displacement (reg returnaddr) nil))))

(loadtime (remflag '(fluid global $fluid $global) 'Terminaloperand))

(defanyreg Fluid ZIBanyregfluid)

(defanyreg $Fluid ZIBanyregfluid)

%-----------------------------------------------------------------------------

(de quote-fix-p (x)
    (and (eqcar x 'quote) (fixp (cadr x))))

(de displacement-p (x)
    (and (eqcar x 'displacement)
     (eqn (length x) 3) (alpha-register-p (cadr x))))

(de deferred-p (x)
    (and (pairp x) (or (alpha-register-p (car x))
		   (eq (car x) 'deferred))))

(remprop '*move 'optfn)

(DefCMacro *Move
       (Equal)
       ((regp regp)          (bis ArgOne (reg r31) ArgTwo))
       ((displacementp regp) (ldq ArgTwo ArgOne))
       ((regp displacementp) (stq ArgOne ArgTwo))
       ((regp anyp)          (*movex ArgOne ArgTwo))
       ((anyp regp)          (*movex ArgOne ArgTwo))
       (                     (*move ArgOne (reg t5))
			     (*move (reg t5) ArgTwo)))
			  
(de deferredp (x) (and (pairp x) (eq (car x) 'deferred)))

(DefCMacro *Jump
       ((deferredp)        (*move (undeferred ArgOne) (reg t5))
			   (jmp (reg null) (deferred (reg t5)))) 
       ((displacementp)    (jmp (reg null) ArgOne))
       (                   (br (reg null) ArgOne)))
       
(DefCMacro *Lbl           (ArgOne))

(de *3op (a1 a2 instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	   (list (ResolveOperand '(Reg t1) A1)
		 (ResolveOperand '(Reg t2) A2)
		  Instruction)
		 (get '*3op 'CMacroPatternTable)))))

(DefCMacro *3op
       ((regp regp) (ARGTHREE ArgOne ArgTwo ArgOne))
       ((regp alpha-short-immediate-p)
		    (ARGTHREE ArgOne ArgTwo ArgOne))
       ((regp AnyP)  (*move ArgTwo (reg T5))
		     (*3op ArgOne (reg t5) ArgThree))
       ((Anyp regp)  (*move ArgOne (reg T5))
		     (ARGTHREE (reg t5) ArgTwo (reg t5))
		     (*move (reg t5) ArgOne))
       ((Anyp alpha-short-immediate-p)
		     (*move ArgOne (reg T6))
		     (ArgThree (reg t6) ArgTwo (reg t5))
		     (*move (reg t5) ArgOne))
       (            (*move ArgTwo (reg T5))
		     (*move ArgOne (reg T6))
		     (ArgThree (reg t6) (reg t5) (reg t5))
		     (*move (reg t5) ArgOne)))

(DefCMacro *WPlus2 
    ((*3op ArgOne ArgTwo Addq)))

(de *wplus3(arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wplus3))

(DefCMacro *WPlus3
	   ((regp regp regp) (addq ArgThree ArgTwo ArgOne))
	   ((regp alpha-short-immediate-p regp)
			     (addq ArgThree ArgTwo ArgOne))
	   ((regp regp alpha-short-immediate-p)
			     (addq ArgTwo ArgThree ArgOne))
	   ((regp alpha-short-immediate-p anyp)
			     (*move ArgThree (reg t5))
			      (addq (reg t5) ArgTwo ArgOne))
	   ((regp anyp alpha-short-immediate-p)
			     (*move ArgTwo (reg t5))
			      (addq (reg t5) ArgThree ArgOne))
           ((regp regp anyp) (*move ArgThree (reg t5))
                              (*wplus3 ArgOne ArgTwo (reg t5)))
	   ((regp anyp anyp) (*move ArgTwo (reg t6))
                              (*wplus3 ArgOne (reg t6) argthree))
	   ((anyp regp anyp) (*move ArgThree (reg t5))
			      (*wplus2 (reg t5) ArgTwo)
			      (*move (reg t5) ArgOne))
	   (                 (*move ArgTwo (reg t5))
			      (*move ArgThree (reg t6))
			      (addq (reg t5) (reg t6) (reg t5))
			      (*move (reg t5) ArgOne)))

(DefCMacro *WDifference
       ((*3op ArgOne ArgTwo Subq)))

(de *wdifference3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wdifference3))

% format ArgOne = ArgTwo - ArgThree

(DefCMacro *Wdifference3
	   ((regp regp regp) (subq ArgTwo ArgThree ArgOne))
	   ((regp regp alpha-short-immediate-p)
			     (subq ArgTwo ArgThree ArgOne))
	   ((regp anyp regp) (*move ArgTwo (reg t5))
			      (subq (reg t5) ArgThree ArgOne))
	   ((regp anyp anyp) (*move ArgTwo (reg t5))
			      (*move ArgThree (reg t6))
			      (subq (reg t5) (reg t6) ArgOne))
	   ((anyp regp anyp) (*move ArgThree (reg t5))
			      (subq ArgTwo (reg t5) (reg t5))
			      (*move (reg t5) ArgOne))
	   (                 (*move ArgTwo (reg t5))
			      (*move ArgThree (reg t6))
			      (subq (reg t5) (reg t6) (reg t5))
			      (*move (reg t5) ArgOne)))

(defcmacro *wtimes2)

(de *wtimes2 (arg1 arg2) (*wtimes3 arg1 arg1 arg2))

(de *wtimes3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wtimes3))

% format ArgOne = ArgTwo * ArgThree

(DefCMacro *wtimes3
	   ((regp regp regp) (mulq ArgThree ArgTwo ArgOne))
	   ((regp alpha-short-immediate-p regp)
			     (mulq ArgThree ArgTwo ArgOne))
	   ((regp regp alpha-short-immediate-p)
			     (mulq ArgTwo ArgThree ArgOne))
	   ((regp alpha-short-immediate-p anyp)
			     (*move ArgThree (reg t5))
			      (mulq (reg t5) ArgTwo ArgOne))
	   ((regp anyp alpha-short-immediate-p)
			     (*move ArgTwo (reg t5))
			      (mulq (reg t5) ArgThree ArgOne))
	   ((regp regp anyp) (*move ArgThree (reg t6))
                              (*wtimes3 ArgOne ArgTwo (reg t6)))
	   ((regp anyp anyp) (*move ArgTwo (reg t5))
                              (*wtimes3 ArgOne (reg t5) ArgThree))
	   ((anyp regp anyp) (*move ArgThree (reg t5))
			      (*wtimes2 (reg t5) ArgTwo)
			      (*move (reg t5) ArgOne))
	   (                 (*move ArgTwo (reg t5))
			      (*move ArgThree (reg t6))
			      (mulq (reg t5) (reg t6) (reg t5))
			      (*move (reg t5) ArgOne)))

(DefCMacro *WAnd
       ((*3op ArgOne ArgTwo and)))

(DefCMacro *WOr
       ((*3op ArgOne ArgTwo bis)))

(DefCMacro *WXOr
       ((*3op ArgOne ArgTwo xor)))

(de fixplusp (x) (and (fixp x) (not (minusp x))))

(DefCMacro *AShift
    ((AnyP ZeroP))
    ((regp negp)     (sra ArgOne (minus ArgTwo) ArgOne))
    ((regp fixplusp) (sll ArgOne ArgTwo ArgOne))
    ((regp regp)    (subq (reg null) ArgTwo (reg t6))
                      (bgt (reg t6) templabel)
                      (sll ArgOne ArgTwo ArgOne)
                      (br (reg null) templabel2)
                     templabel
                      (sra ArgOne (reg t6) ArgOne)
		     templabel2)
    ((regp anyp)     (*move ArgTwo (reg t5))
		      (*ashift ArgOne (reg t5)))
    ((Anyp fixp)     (*move ArgOne (reg t5))
		      (*ashift (reg t5) ArgTwo)
		      (*move (reg t5) ArgOne))
    (                (*move ArgOne (reg t1))
		      (*ashift (reg t1) ArgTwo)
		      (*move (reg t1) ArgOne)))

(DefCMacro *WShift
    ((AnyP ZeroP))
    ((regp negp)     (srl ArgOne (minus ArgTwo) ArgOne))
    ((regp fixplusp) (sll ArgOne ArgTwo ArgOne))
    ((regp regp)    (subq (reg null) ArgTwo (reg t6))
		      (bgt (reg t6) templabel)
		      (sll ArgOne ArgTwo ArgOne)
		      (br (reg null) templabel2)
		     templabel
		      (srl ArgOne (reg t6) ArgOne)
		     templabel2)
    ((regp anyp)     (*move ArgTwo (reg t5))
		      (*wshift ArgOne (reg t5)))
    ((Anyp fixp)     (*move ArgOne (reg t5))
		      (*wshift (reg t5) ArgTwo)
		      (*move (reg t5) ArgOne))
    (                (*move ArgOne (reg t1))
		      (*wshift (reg t1) ArgTwo)
		      (*move (reg t1) ArgOne)))

(DefCMacro *WNot
       ((regp regp)  (ornot (reg null) ArgTwo ArgOne))
       ((regp Anyp)  (*move ArgTwo ArgOne)
		      (ornot (reg null) ArgOne ArgOne))
       (             (*move ArgTwo (reg t5))
		      (*wnot (reg t5) (reg t5))
		      (*move (reg t5) ArgOne)))

(DefCMacro *WMinus
	((regp regp) (subq (reg null) ArgTwo ArgOne))
	((regp Anyp) (*move ArgTwo ArgOne)
			(subq (reg null) ArgOne ArgOne))
	(            (*move ArgTwo (reg t5))
			(*wminus (reg t5) (reg t5))
			(*move (reg t5) ArgOne)))

(DefCMacro *MkItem
       ((regp fixp)     (sll ArgOne 8 ArgOne)
			 (srl ArgOne 8 ArgOne)
			 (*move ArgTwo (reg t5)) 
			 (sll (reg t5) 56 (reg t5))
			 (bis ArgOne (reg t5) ArgOne))
       ((regp regp)     (sll ArgOne 8 ArgOne)
			 (srl ArgOne 8 ArgOne)
			 (sll ArgTwo 56 (reg t5))
			 (bis ArgOne (reg t5) ArgOne))
       ((regp anyp)     (*move ArgTwo (reg t5))
			 (*Mkitem ArgOne (reg t5)))
       (                 (*move ArgOne (reg t1))
			 (*mkitem (reg t1) ArgTwo)
			 (*move (reg t1) ArgOne)))

(de smallinump (x) (alpha-short-immediate-p x))

(de immediate-p (x)    (fixp x))

(DefCmacro *tag
      ((regp regp) (srl ArgTwo 56 ArgOne))
      ((Anyp  regp)(*move ArgTwo (reg t1))
		    (srl (reg t1) 56 (reg t1))
		    (*move (reg t1) ArgOne))
      ((regp anyp) (*move ArgTwo (reg t1))
		    (srl (reg t1) 56 ArgOne))
      (             (*move ArgTwo (reg t1))
		    (*tag ArgOne (reg t1))))

(de *tag (ar1 ar2)
     (expand2operandCmacro ar1 ar2 '*tag))

(DefCMacro *JumpType
      (             (*Tag (Reg t3) ArgOne)
		    (*JumpEQ ArgThree (Reg t3) ArgTwo)))

(DefCMacro *JumpNotType
      (             (*Tag (Reg t3) ArgOne)
		    (*JumpNOTEQ ArgThree (Reg t3) ArgTwo)))

(DefCMacro *JumpInType
       ((regp Anyp) (srl ArgOne 56 (reg t5))
		     (subq (reg t5) ArgTwo (reg at))
		     (ble (reg at) ARGTHREE)
		     (subq (reg t5) 255 (reg at)) % negint-tag)
		     (beq (reg at) ARGTHREE))
       (             (*move ArgOne (reg t5))
		     (srl (Reg t5) 56 (reg t5))
		     (subq (reg t5) ArgTwo (reg at))
		     (ble (reg at) ARGTHREE)
		     (subq (reg t5) 255 (reg at)) % negint-tag)
		     (beq (reg at) ARGTHREE)))

(DefCMacro *JumpNotInType
       ((regp Anyp) (srl ArgOne 56 (reg t5))
		     (subq (reg t5) 255 (reg at))
		     (beq (reg at) templabel)
		     (subq (reg t5) ArgTwo (reg at))
		     (bgt (reg at) ARGTHREE)
		    templabel)
       (            (*move ArgOne (reg t5))
		     (srl (reg t5) 56 (reg t5))
		     (subq (reg t5) 255 (reg at))
		     (beq (reg at) templabel)
		     (subq (reg t5) ArgTwo (reg at))
		     (bgt (reg at) ARGTHREE)
		    templabel))

%---------------------------------------------------------------------
%   General Purpose Jumper.
%   FORMAT: Operand1 Operand2 Address JumpOpCode

(DefCMacro *JumpIf
     ((regp regp)      (ArgFour ArgOne ArgTwo (reg at))
			 (ArgFive (reg at) ArgThree))
     ((regp SmallInumP) (ArgFour ArgOne ArgTwo (reg at))
			 (ArgFive (reg at) ArgThree))
     ((regp anyp)       (*move ArgTwo (reg t5))
			 (ArgFour ArgOne (reg t5) (reg at))
			 (ArgFive (reg at) ArgThree))
     ((anyp  regp)      (*move ArgOne (reg t5))
			 (ArgFour (reg t5) ArgTwo (reg at))
			 (ArgFive (reg at)ArgThree))
     ((anyp  SmallInumP) (*move ArgOne (reg t5))
			 (ArgFour (reg t5) ArgTwo  (reg at))
			 (ArgFive (reg at) ArgThree))
     (                  (*move ArgOne (reg t5))
			 (*move ArgTwo (reg t4))
			 (ArgFour (reg t5) (reg t4) (reg at))
			 (ArgFive (reg at) ArgThree)))

(de *JumpIF (ArgOne ArgTwo Label Instruction Instruction2)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	   (list (ResolveOperand '(Reg t1) ArgOne)
		 (ResolveOperand '(Reg t2) ArgTwo)
		     (ResolveOperand '(Reg Error) Label)
		  Instruction Instruction2)
		 (get '*JumpIF 'CMacroPatternTable)))))


(defcmacro *jumpeq)

(de *jumpeq (Lbl ArgOne ArgTwo)
       (*jumpif ArgOne ArgTwo lbl 'CMPEQ 'BNE))

(defcmacro *jumpnoteq)

(de *jumpnoteq(Lbl ArgOne ArgTwo)
	(*jumpif ArgOne ArgTwo Lbl 'CMPEQ 'BEQ))

(defcmacro *jumpwlessp)

(de *jumpwlessp (Lbl ArgOne ArgTwo)
	(*jumpif ArgOne ArgTwo lbl 'CMPLT 'BNE))

(defcmacro *jumpwgreaterp)

(de *jumpwgreaterp (Lbl ArgOne ArgTwo)
	(*jumpif ArgOne ArgTwo lbl 'CMPLE 'BEQ))

(defcmacro *jumpwleq)

(de  *jumpwleq(Lbl ArgOne ArgTwo)
	(*jumpif ArgOne ArgTwo lbl 'CMPLE 'BNE))

(defcmacro *jumpwgeq)

(de *jumpwgeq (Lbl ArgOne ArgTwo)
	(*jumpif ArgOne ArgTwo lbl 'CMPLT 'BEQ))

% a new type of cmacro: set dest to T or NIL under condition

(de *setcond (ArgOne ArgTwo ArgThree instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg t1) ArgThree)
                  instruction)
                 (get '*setcond 'CMacroPatternTable)))))

(defcmacro *setcond
     ((regp regp regp)  (*wdifference3 (reg t5) ArgTwo ArgThree)
                         (*move (reg NIL) ArgOne)
			 (argfour (reg t5) (reg T) ArgOne))
     ((regp regp smallinump)
                        (*wdifference3 (reg t5) ArgTwo ArgThree)
			 (*move (reg NIL) ArgOne)
                         (ArgFour (reg t5) (reg T) ArgOne))
     ((regp anyp regp)  (*move argTwo (reg t5))
                         (*setcond ArgOne (reg t5) ArgThree ArgFour))
     ((regp anyp smallinump)
                        (*move argTwo (reg t5))
                         (*setcond ArgOne (reg t5) ArgThree ArgFour))
     ((regp anyp anyp)  (*move argTwo (reg t5))
                         (*move argThree ArgOne)
                         (*setcond ArgOne (reg t5) ArgOne ArgFour))
     (                  (*setcond (reg t5) ArgTwo ArgThree ArgFour)
                         (*move (reg t5) ArgOne))
)

(defcmacro *setcondeq)

(de *setcondeq (dest ArgOne ArgTwo)
       (*setcond dest ArgOne ArgTwo 'CMOVEQ))

(defcmacro *setcondnoteq)

(de *setcondnoteq(dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo 'CMOVNE))

(defcmacro *setcondwlessp)

(de *setcondwlessp (dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo 'CMOVLT))

(defcmacro *setcondwgreaterp)

(de *setcondwgreaterp (dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo 'CMOVGT))

(defcmacro *setcondwleq)

(de  *setcondwleq(dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo 'CMOVLE))

(defcmacro *setcondwgeq)

(de *setcondwgeq (dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo 'CMOVGE))


(de *setcondtype (ArgOne ArgTwo ArgThree)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg t1) ArgThree))
                 (get '*setcondtype 'CMacroPatternTable)))))

(de *setcondNOTtype (ArgOne ArgTwo ArgThree)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg t1) ArgThree))
                 (get '*setcondNOTtype 'CMacroPatternTable)))))

(DefCMacro *setcondType
      (             (*Tag (Reg t6) ArgTwo)
                    (*setcondEQ ArgOne (Reg t6) ArgThree)))

(DefCMacro *setcondNotType
      (             (*Tag (Reg t6) Argtwo)
                    (*setcondNOTEQ ArgOne (Reg t6) ArgThree)))

%---------------

(de fluid-arg-p (x)
    (and (pairp x)
     (member (car x) '(!$fluid !$global fluid global))))

(DefCMacro *Push
       ( (*cerror "NO push ")))

(DefCMacro *Pop
       ( (*cerror "NO pop ")))

(de take-car (x) (car x))
(de take-cadr (x) (cadr x))   % functions to strip out deferred and
			  % displacement parts.
(de take-caddr (x) (caddr x))


(de *loc (ArgOne ArgTwo)
   (let ((resolvedarg (resolveoperand '(reg t2) ArgTwo)))
      (cond ((deferred-p resolvedarg)
	 `((*move (undeferred ,resolvedarg) ,ArgOne))) % get the address.
	 ((displacement-p resolvedarg)
	    `((addq ,(cadr resolvedarg) ,(caddr resolvedarg) ,ArgOne)))
	 ((eqcar resolvedarg 'cdr)
	    `((addq ,(cadr resolvedarg) 4 ,ArgOne)))
	 (t `((*movex ,resolvedarg ,ArgOne))))))

(DefCMacro *loc)

%t6 is used, because it is not used in lap code, but it destroyed in cmacros.
%look at test file mask.s in ~burdorf/alpha for example.

(de Eightp (u) (eq u 8))

(de X56p (u) (eq u 56))

(DefCMacro *Field

  ((regp regp zerop fixp)  (srl ArgTwo (difference 64 ARGFOUR) ArgOne))
  ((regp regp Eightp X56p) (zap ArgTwo 128  ArgOne))
  ((regp regp fixp fixp)   (sll ArgTwo ArgThree ArgOne)
			    (srl ArgOne (difference 64 ARGFOUR) ArgOne))
  ((regp anyp fixp fixp)   (*move ArgTwo (reg t5))
			    (sll (reg t5) ArgThree (reg t5))
			    (srl (reg t5) (difference 64 ARGFOUR) ArgOne))
  ((anyp regp fixp fixp)   (*field (reg t5) ArgTwo ArgThree ArgFour)
			    (*move (reg t5) ArgOne))
  (                         (*move ArgTwo (reg t6))
			 (*field (reg t5) (reg t6) ArgThree ArgFour)
			 (*move (reg t5) ArgOne)))

(DefCMacro *SignedField

  ((regp regp fixp fixp) (sll ArgTwo ArgThree ArgOne)
			 (sra ArgOne (difference 64 ARGFOUR) ArgOne))
  ((regp anyp fixp fixp) (*move ArgTwo (reg t5))
			 (sll (reg t5) ArgThree (reg t5))
			 (sra (reg t5) (difference 64 ARGFOUR) ArgOne))
  ((anyp regp fixp fixp) (*Signedfield (reg t5) ArgTwo ArgThree ArgFour)
			 (*move (reg t5) ArgOne))
  (                      (*move ArgTwo (reg t6))
			 (*Signedfield (reg t5) (reg t6) ArgThree ArgFour)
			 (*move (reg t5) ArgOne)))

% following is used by garbage collector only. Not very useful to
% optimize other cases than put inf field or put tag field (with constant)


(DefCMacro *PutField

 ((regp regp Eightp X56P)     (zap ArgOne 16#80 (reg t6))
			       (zap ArgTwo 16#7f ArgTwo)
			       (bis  ArgTwo (reg t6) ArgTwo))
 ((regp regp Zerop Eightp)    (sll   ArgOne 56 (reg t5))
			       (zap ArgTwo  16#80 ArgTwo)
			       (bis (reg t5) ArgTwo ArgTwo))
 ((inump regp Zerop Eightp)   (lda (reg t5) ArgOne)
			       (sll (reg t5) 56 (reg t5))
			       (zap ArgTwo 16#80 ArgTwo)
			       (bis (reg t5) ArgTwo ArgTwo))
 (     (lda (reg t5) -1)
	(*move   ArgOne (reg t4))
	(sll     (reg t5) (difference 64 ArgFour) (reg t5))
	(*wshift (reg t5) (minus ArgThree)) % for 0 opt
	(sll   (reg t4) (difference 64 (plus ArgThree ArgFour)) (reg t4))
	(and   (reg t5) (reg t4) (reg t4))
	(*move ArgTwo (reg t6))
	(bic  (reg t6) (reg t5) (reg t6))
	(bis  (reg t4) (reg t6) (reg t5))
	(*move (reg t5) ArgTwo)))


 % Taken from DEC's architecture handbook, p 4-5o and 4-46

(deflist '((Byte    ((addq (reg 2) (reg 1) (reg t2))
		     (ldq_u (reg 1) (displacement (reg t2) 0))
		     (sll (reg t2) 3 (reg t2))
		     (sll (reg 1) (reg t2) (reg 1))
		     (sra (reg 1) 56 (reg 1))))
       (PutByte     ((addq (reg 2) (reg 1) (reg t1))
		     (ldq_u  (reg t2) (displacement (reg t1) 0))
		     (*wxor (reg t1) 7)
		     (insbl (reg 3) (reg t1) (reg t5))
		     (mskbl (reg t2) (reg t1) (reg t2))
		     (bis (reg t2) (reg t5) (reg t5))
		     (stq_u (reg t5) (displacement (reg t1) 0))))
       (HalfWord    ((sll (reg 2) 1 (reg t1))
		     (addq (reg t1) (reg 1) (reg t1))
		     (ldq_u  (reg 1) (displacement (reg t1) 0))
                     (sll (reg 2) 4 (reg t1))
		     (sll (reg 1) (reg t1) (reg 1))
                     (sra (reg 1) 48 (reg 1))))
       (PutHalfWord ((sll (reg 2) 1 (reg t1))
		     (addq (reg t1) (reg 1) (reg t1))
		     (ldq_u  (reg t2) (displacement (reg t1) 0))
		     (*move -1 (reg t4))
		     (sll (reg 2) 4 (reg t3))
		     (sll (reg t4) 48 (reg t4))
		     (srl (reg t4) (reg t3) (reg t4))
                     (!*wnot (reg t4) (reg t4))
		     (!*wand (reg t2) (reg t4))
		     (sll (reg 3) 48 (reg t4))
		     (srl (reg t4) (reg t3) (reg t4))
		     (!*wor (reg t2) (reg t4))
		     (stq_u (reg t2) (displacement (reg t1) 0)))))
  'OpenCode)

(&OneReg '(Byte PutByte HalfWord PutHalfWord))

(on fast-integers)

(for (in function '(idapply0 idapply1 idapply2 idapply3 idapply4))
     (from register 1)
     (do
      (put function 'opencode
        `((zap (reg ,register) 16#80 (reg t3))
          (s8addq (reg t3) (reg symfnccenter) (reg r27))
          (ldq (reg r27) (displacement (reg r27) 0))
          (jsr (reg returnaddr) (displacement (reg r27) nil))
          (ldq (reg r27) (displacement (reg st) 8))))
      (put function 'exitopencode
        `((zap (reg ,register) 16#80 (reg t3))
          (s8addq (reg t3) (reg symfnccenter) (reg r27))
          (ldq (reg r27) (displacement (reg r27) 0))
          (jmp (reg null) (displacement (reg r27) nil))))))

(for (in function '(addressapply0 addressapply1 addressapply2
		addressapply3 addressapply4))
     (from register 1)
     (do
      (put function 'opencode
	 `((*move (reg ,register) (reg r27))
	   (jsr (reg returnaddr) (displacement (reg r27) nil))
	   (ldq (reg r27) (displacement (reg st) 8))))
      (put function 'exitopencode
	 `((*move (reg ,register) (reg r27))
	   (jmp (reg null) (displacement (reg r27) 0)))))
      )

(defcmacro *fast-apply-load
   ( (*move ArgOne (reg t1))))

(put 'fast-idapply    'opencode
     '((zap (reg t1) 16#80 (reg t3))
       (s8addq (reg t3) (reg symfnccenter) (reg r27))
       (ldq (reg r27) (displacement (reg r27) 0))
       (jsr (reg returnaddr) (displacement (reg r27) nil))
       (ldq (reg r27) (displacement (reg st) 8)) ))

(put 'fast-codeapply    'opencode
     '((zap (reg t1) 16#80 (reg r27))
			    %codepointer passed, so just strip off tag
       (jsr (reg returnaddr) (displacement (reg r27) nil))
       (ldq (reg r27) (displacement (reg st) 8)) ))
       
(put 'fast-idapply    'exitopencode
     '((zap (reg t1) 16#80 (reg t3))
       (s8addq (reg t3) (reg symfnccenter) (reg r27))
       (ldq (reg r27) (displacement (reg r27) 0))
       (jmp (reg null) (displacement (reg r27) nil))))

(put 'fast-codeapply    'exitopencode
     '((zap (reg t1) 16#80 (reg r27))
			 %codepointer passed, so just strip off tag
       (jmp (reg null) (displacement (reg r27) nil))))

(De *LamBind (Regs Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cregs cfluids cadrcfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Regs (rest Regs))   % Remove "Registers" from the front
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Regs) 16)) % two words per BndStk entry *4

   %looking for free register to preload values of fluid s

      (setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
		   (function (lambda (x)
				(cond ((member x regs) nil)
				       (t (cons x nil))
      )              ) )         )  )
      (setq freeregs (nconc (list '(reg T3) '(reg T4)) freeregs))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
		(nconc cadrcfluids (cons (car freeregs) nil)))
	(nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist

      (setq list `((*wplus3 (reg t2) (Reg bndstkptr) ,lng )
		   (*move (Reg bndstkptr) (Reg t1))
                   (subq (reg BndstkUpperBound) (reg t2) (reg t5))
                   (bgt (reg t5) ,genlabel)
                   (*call Bstackoverflow)
                  ,genlabel  
		   (*move (Reg t2) (Reg bndstkptr)))) %start of code

      (setq list (append initload list))

  loop
      (setq cregs (car Regs))
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	   (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 16))
      (Setq list (append list
     `((*move ,(car freeregs) (displacement (Reg t1) ,n))
       (*move (quote ,Cadrcfluids)
		   (displacement (reg t1) ,(wplus2 n -8)))
       (*move ,cregs ,cfluids)
      )          ))
      (setq fluids (cdr Fluids))
      (setq freeregs (cdr freeregs))
      (cond ((setq regs (cdr Regs)) (go loop)))
      (return list)
)    )

(defcmacro !*lambind)

(De *ProgBind (Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cfluids cadrcfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 16)) % two words per BndStk entry
					 % * 8 addressingunits
      (setq freeregs '((reg t3)(reg T4)(reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
		(nconc cadrcfluids (cons (car freeregs) nil)))
	(nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist


	(setq list `((*wplus3 (reg t2) (Reg bndstkptr) ,lng )
		(*move  (Reg bndstkptr)  (Reg t1))
		(subq (reg BndstkUpperBound) (reg t2) (reg t5))
		(bgt (reg t5) ,genlabel)
		(*call Bstackoverflow)
	        ,genlabel
		(*move (Reg t2) (Reg bndstkptr)))) %start of code


     (setq list (append initload list))

 loop
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	   (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 16))
      (Setq list (append list
	     `((*move ,(car freeregs) (displacement (Reg t1) ,n))
	       (*move (quote ,Cadrcfluids)
			   (displacement (reg t1)  ,(wplus2 n -8)))
	       (*move (quote nil) ,cfluids)
      )          ))
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *progbind)

(De *FreeRstr (Fluids)

 (prog (n lng list cfluids listfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
      (setq freeregs '((reg t3)(reg t4)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
      (setq n (wtimes2 8 (wdifference 2 lng)))
      (setq lng (wtimes2 lng 8)) % * addressingunitperitem
      (setq initload (list '(*move (fluid Bndstkptr) (reg t1))))

preload  (setq initload
       (progn (setq listfluids
	       (if freeregs
		(nconc listfluids (cons (car freeregs) nil))
		(nconc listfluids (cons nil nil))) )
	(nconc initload
	 (if freeregs
	  `((*move (displacement (reg t1) ,n) ,(car freeregs))) nil)
       )   ))
       (setq n (wplus2 n 16))
       (when freeregs (setq freeregs (cdr freeregs)))
       (setq cfluids (cdr cfluids))

       (cond (cfluids (go preload)))

       (setq freeregs listfluids ) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and nil if not enough regs available

      (setq list `((*wdifference3 (Reg t2) (reg t1) ,lng)
		(subq (reg t2)(reg BndstkLowerBound) (reg t5))
		(bge (reg t5) ,genlabel)
		(*call  Bstackunderflow)
		,genlabel
		(*move (Reg t2) ($fluid BndstkPtr)))) %start of code

     (setq list (append initload list))
     (setq n 0)

 loop
      (setq cfluids (car Fluids))
      (setq n (wplus2 n 16))

  % insert reloaded register or memory reference

      (setq list (append list
	   (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
		`((*move (displacement (Reg t2) ,n) ,cfluids )))

      )          )
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )

(defcmacro *freerstr)

(setq *unsafebinder t)   % has to save Registers across calls

(global '(*writingasmfile))

(de !*xjumpon (register lowerbound upperbound labellist)
    (for (from x lowerbound upperbound)
	 (join
	  `((subq ,register ,x (reg t5))
	    (beq (reg t5) ,(pop labellist))
	    ))))

(de !*jumpon (register lowerbound upperbound labellist)
   (if (or *writingasmfile)
       (*xjumpon register lowerbound upperbound labellist)
     (PROG (X LL LL2)
       (setq ll  (gensym))
       (setq ll2 (stringgensym))
       (SETQ X
	(if (and (weq lowerbound 0) (weq upperbound 31) *syslisp)
		  % jumpon on tags (most probably)
      `((*wplus3 (reg t5) (difference 2 ,lowerbound) ,register)
	(sll (reg t5) 2 (reg t5))  % 4 bytes per jumptable entry
	(br (reg t1) ,ll2) % t1 contains address of following instr!
      ,ll2 (addq (reg t1) (reg t5) (reg t1))
	(jmp (reg null) (displacement (reg t1) nil)))
      `((subq ,register ,upperbound (reg at))
	(beq (reg at) ,(lastcar labellist))
	(bgt (reg at)  (label ,ll))
	(subq ,register ,Lowerbound (reg t5))
	(blt (reg t5) (label ,ll))
        (*wplus3 (reg t5) (difference 2 ,lowerbound) ,register)
        (sll (reg t5) 2 (reg t5))
        (br (reg t1) ,ll2) % t1 contains address of following instr!
      ,ll2 (addq (reg t1) (reg t5) (reg t1))
        (jmp (reg null) (displacement (reg t1) nil)))))

      Loop  (Setq x (nconc X `((Br (reg null) ,(car Labellist)) )) )
	    (setq Labellist (cdr Labellist))
	    (cond (Labellist (go loop)))

	    (setq x (nconc x `((*lbl (Label ,ll))) ))
	    (return x)
)  )  )

(defcmacro !*jumpon)


% Call foreign fns (loaded by oload) by putting SymFunc cell addr in a reg,
% then doing "calls" instr indirect via the target address of the jmp
% instruction stored there.  Use old direct link for asm (kernel) compile.

(de *foreignlink (functionname functiontype numberofarguments)

	`((*move (quote T) ($fluid *kernelmode))
	  (*move (reg st) (reg t1))
	  (LDA (reg st) (displacement (reg st) -100))
	  (bic  (reg st) 31 (reg st))
	  (stq (reg t1) (displacement (reg st) 64))
	  (!.extern (foreignentry ,functionname))
	  (bsr (reg returnaddr) (foreignentry ,functionname))
	  (la (reg symvalcenter) SYMVAL) % may be lost
	  (la (reg symfnccenter) SYMFNC) % too
	  (*move (reg r0) (reg 1))  % value
	  (ldq (reg st) (displacement (reg st) 64))
	  (*move (quote NIL) ($fluid *kernelmode))))

(defcmacro !*foreignlink)

(defcmacro *movex)
   % Movex is of type cmacro

(de *movex (src dst)
 (let (tmp)
 (cond ((alpha-register-p dst)
  (cond ((alpha-register-p src) `((bis ,src (reg null) ,dst)))
	((eqcar src 'idloc)     `((lda ,dst ,src )))
	%%%% ((eqcar src 'quote) `((ldq ,dst ,src)))
	((eqcar src 'quote) (*movex-quote src dst))
	((or (alpha-absolute-p src) (alpha-fluid-p src))
	 (cond 
           ( *writingasmfile  
	     `((lda (reg t1) ,(findidnumber (cadr src)))
		(s8addq  (reg t1) (reg symvalcenter) (reg t1))
		(ldq ,dst (displacement (reg t1) 0))) 
           )

           ((flagp (cadr src) 'kernel)
              `((ldq ,dst (displacement (reg symvalcenter) 
                                        (xidloc ,(cadr src))))
            )
           )

           (t `((lda (reg at) (idloc ,(cadr src)))
                (s8addq  (reg at) (reg symvalcenter) (reg at))
                (ldq ,dst (displacement (reg at) 0))))))

	((alpha-short-immediate-p src) `((bis (reg null) ,src ,dst)))
	((or (alpha-relocatable-immediate-p src)
	     (alpha-foreignentry-p src))
	 (if *writingasmfile           %% asmout and lap generate
	     `((sethi (hi ,src) ,dst)  %% different code.
	       (add ,dst (lo ,src) ,dst))
	     `((move-relocatable-object ,src ,dst))))
	((alpha-disp-p src) (if *writingasmfile
			`((la ,dst (displacement (reg r31) ,src)))
			`((lda ,dst (displacement (reg r31) ,src)))))
	((alpha-immediate-p src) (if *writingasmfile
			`((la ,dst (quote ,src))
			  (ldq ,dst (displacement ,dst 0)))
			`((ldq ,dst (quote ,src)))))
	((alpha-huge-displacement-p src)
	     `((sethi (hi ,(caddr src)) ,dst)
	       (add ,dst (lo ,(caddr src)) ,dst)
	       (ld (,(cadr src) ,dst) ,dst)))
	((and (alpha-deferred-p src) (alpha-indirect-p (cadr src)))
	       (car (load-src-arg src dst)))
	((alpha-indirect-p src) `((ldq ,dst ,src)))
	(t  (print-error-status)
	    (error 0 (list "Unknown addressing mode for *movex:" src dst)))))
     ((alpha-register-p src)
	 (cond ((or (alpha-absolute-p dst) (alpha-fluid-p dst)
		     (alpha-foreignentry-p src))
	 (cond
           ( *writingasmfile
	     `((lda (reg r28) ,(findidnumber (cadr dst)))
	       (s8addq (reg r28) (reg symvalcenter) (reg r28))
	       (stq ,src (displacement (reg r28) 0)))
           )
           ((flagp (cadr dst) 'kernel)
              `((stq ,src (displacement (reg symvalcenter) 
                                        (xidloc ,(cadr dst)))))
           )
           (t`((lda (reg r28) (idloc ,(cadr dst)))
               (s8addq (reg r28) (reg symvalcenter) (reg r28))
               (stq ,src (displacement (reg r28) 0))))))

	((and (alpha-deferred-p dst) (alpha-indirect-p (cadr dst)))
	 (setq tmp (load-src-arg (cadr dst) '(reg t6)))
	 (append (car tmp) `((st ,src (deferred (reg t6))))))
	((alpha-huge-displacement-p dst)
	 `((sethi (hi ,(caddr dst)) (reg t6))
	   (add (reg t6) (lo ,(caddr dst)) (reg t6))
	   (st ,src (,(cadr dst) (reg t6)))))
	((alpha-indirect-p dst) `((stq ,src ,dst)))
	(t   (print-error-status)
	     (error 0 (list "Unknown addressing mode:" src dst)))))
   (t (print (list " Unsaeglich: Move " src dst))))))

(de *movex-quote(src dest) 
   (if *writingasmfile 
		`((la ,dest ,src) (ldq ,dest (displacement ,dest 0)))
		`((ldq ,dest ,src))))



(commentoutcode

% code for ALPHA immediate quote

(fluid '(*alpha-quote-nodes*))

(setq *alpha-quote-nodes* 1000)

(de *movex-quote(src dest)
 (if (or (equal dest '(reg r28))
         (greaterp *alpha-quote-nodes* 500)
     )
     `((ldq ,dest ,src))
   (let ((immq (expandimmediatequote (cadr src))))
     (when (eqcar immq 'immediate)
           (setq immq (cadr immq)))
     (when (not (eqcar immq 'mkitem))
           (stderror (list "***** immediatequote:" immq "unknown")))
    `((*move ,(cadr immq) (reg r28))
      (*wshift (reg r28) 56)
      (lda ,dest ,(caddr immq))
      (*wor ,dest (reg r28))
))))       

(de alpha-test-immmediatequote(w)
  % look whether the number of nodes in quotes allows immediate quote
  % addressing
 (setq *alpha-quote-nodes* 0)
 (alpha-test-immmediatequote1 w nil)
)

(de alpha-test-immmediatequote1(w m)
  (cond ((atom w) nil)
        ((eq (car w)'quote) 
         (alpha-test-immmediatequote1 (cdr w) t))
        (t (when m (setq *alpha-quote-nodes*
                         (iadd1 *alpha-quote-nodes*)))
           (alpha-test-immmediatequote1(car w) m)
           (alpha-test-immmediatequote1(cdr w) m))))

)
