%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxc/alpha-instrs.sl
% Title:          DEC alpha instructions
% Author:         Winfried Neun
% Created:        19 Jan 1993
% Mode:           Lisp
% Package:        Compiler
%
% Copyright (c) 1993, Konrad-Zuse-Zentrum Berlin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

(compiletime (load fasl-decls))

(global '(*writingasmfile))

(fluid '(labeloffsets!* currentoffset* codesize* codebase* entries*
	 forwardinternalreferences* lapreturnvalue*  *writingfaslfile
	 initoffset* *pgwd *pwrds label-alist* nearestentry*))

%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%k
%%      Assembler code generator
%%
%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

(loadtime
(dm define-instruction (u)
 (prog (name length result)
  (setq u (cdr u)) 
  (setq name (first u))
  (setq length (second u))
  (push (list 'put (mkquote name) (mkquote '*op-length*) length) result)
  (setq u (cddr u))    
  (setq result
     (append (mapcar u (function 
	 (lambda (x) (list 'put (mkquote name)
				(mkquote (car x)) (mkquote (cdr x))))))
	result))
  (return(cons 'progn result))))
)

%%      Pseudo-Instructions

(define-instruction  *ENTRY 0 (*apply* entry-gen))
%
%       (*ENTRY name type argn)
%
(de entry-gen (instruction name type argn)
  (setq nearestentry* currentoffset*)
  (saveentry (list instruction name type argn)))

%
%       (WORD value ......)
%
(define-instruction WORD      8 (*apply* expand-word))
(define-instruction FullWORD  8 (*apply* expand-word))
(define-instruction STRING    0 (*apply* expand-string))
			    %will be treated by onelaplength
(define-instruction Float     8 (*apply* expand-float))

(de expand-word (in x)        (store-word-expression x))

(de expand-string (in x)
	(for (from i 0 (size x))
	     (do (setf (byte (wplus2 codebase* currentoffset*) i) (indx x i))))
	(setf (byte (wplus2 codebase* currentoffset* ) (wplus2 (size x) 1)) 0)
	(setq currentoffset* (wplus2 (wtimes2 addressingunitsperitem 
					(strpack (size x)) currentoffset*))))

(de expand-float (in x)         (store-double x ))

%%    The Instruction processor

(de depositinstruction (in)
   (prog (name format spec oldcurrent lng oldbto)
 
      (setq oldcurrent currentoffset*)
      (when *writingfaslfile (setq oldbto bittableoffset*))
      (when (atom in) (return nil)) % skip labels
      (setq name (first in))
      (setq format (get name '*format*))
      (setq lng (cond ((eq name 'string) 
			(wtimes2 addressingunitsperitem
				 (strpack (size (second in))) ))
		      ((eq name 'oddalign)
			(if (eq currentoffset* (wand currentoffset* (wnot 7)))
				addressingunitsperitem 0))
		      ((eq name 'wordalign)
			(if (eq currentoffset* (wand currentoffset* (wnot 7)))
				0 4))
		      (t (get name '*op-length*) )))
      (when *writingfaslfile (updatebittable lng 0)
% clear bittable 
			     (setq bittableoffset* oldbto))
      
      (cond ((get name '*apply*) (apply (car (get name '*apply*)) in))
			  % escape for special pseudo instructions
	    ((eqcar format '*memory_format*)
		 (format-matcher in (cdr format) nil '*memory_format*))
	    ((eqcar format '*operate_FORMAT*)
		 (format-matcher in (cdr format) nil '*operate_format*))
	    ((eqcar format '*memory_format_branch*)
		 (format-matcher in (cdr format) nil '*memory_format_branch*))
	    ((eqcar format '*branch_format*)
		 (format-matcher in (cdr format) nil '*branch_format*))
	    ((setq format (get name '*fi-format*))  %float instructions
		  (format-matcher in (cdr format)
			 (list (cons 'fmt (car format)))
			 '*fi-format*))
   (t (error 4711 (list "Wrong alpha instruction :" in))))
  % 
  % update currentoffset*
  %
    (setq currentoffset* (plus2 oldcurrent lng)) 
    (when *writingfaslfile (setq bittableoffset* (plus2 oldbto lng ))) 
) )

(de format-matcher (ins format initlist formfkt) 
   (prog (in name item fmtitem alist appl)
       (setq alist 
	     (append initlist
		`((ra . 0) (rb . 0) (dest . 0) (rc . 0)
		 (imm_16 . 0) (lit . 0) (bit0 . 0) (immediate . 0)
		 (function . 0) (imm_16 . 0) (imm_5 . 0) )))
       (setq name (car ins))
       (setq in (cdr ins)) 
	(while (pairp in)
	 (setq item (car in))
	 (setq fmtitem (car format))
	 (cond  ((numberp fmtitem)  nil)
		((and (idp fmtitem)
		      (setq appl (get fmtitem '*matchfn*)))
		      (setq alist (apply appl (list (car in) alist))))
		(t (error 4711 (list "wrong instruction " ins))))
	 (when (and (eq fmtitem 'rb)(or (eqcar (car in) 'entryaddress)
			(eqcar (car in) 'valueaddress)
			(eqcar (car in) 'displacement) 
			(eqcar (car in) 'deferred)
			(eqcar (car in) 'indexed)))
		(setq format (cdr format)))
	 (setq in (cdr in))
	 (setq format (cdr format))
	 (when (and (not (pairp format)) in)
		(error 4711 (list "alpha wrong instruction :" ins))))
	(while format
	   (when (idp (car format))
	       (error 4711 (list "alpha instruction too short :" ins)))
	   (setq format (cdr format)))

   (apply formfkt (list name alist))
)  )
   
(put 'dest '*matchfn* 'dest-match)

(de dest-match (r alist)
    (if (regp r) (cons (cons 'dest r) alist)
     (matcherror "Register" r)))

(put 'rb '*matchfn* 'rb-match)

(de rb-match (x alist)
  (if (or (eqcar x 'displacement) (eqcar x 'deferred) (eqcar x 'indexed)
	   (eqcar x 'entryaddress)(eqcar x 'valueaddress))
			 (based-match x alist)
    (if (regp x) 
	(cons (cons 'rb x) alist)
	(if (and (setq x (addressvalue x)) (posintp x) (wlessp x 16#100))
	    (cons (cons 'imm_8 x) (cons (cons 'rb x) alist))
     (matcherror "Register or Immediate rb" x)))))

(put 'ra '*matchfn* 'ra-match)

(put 'rc '*matchfn* 'rc-match)

(de ra-match (x alist)
    (if (regp x) (cons (cons 'ra x) alist)
     (matcherror "Register" x)))

(de rc-match (x alist)
    (if (regp x) (cons (cons 'rc x) alist)
     (matcherror "Register" x)))

(put 'based '*matchfn* 'based-match)

(de based-match (x alist)
   (if (atom x) 
      (cond ((stringp x) (cons (cons 'immediate
			 (wdifference (addressvalue x) nearestentry*))
			 (cons (cons 'rb '(reg r27)) alist)))
	    ((numberp x) (cons (cons 'immediate x) 
				(cons (cons 'rb '(reg r31)) alist))))
      (cond ((eq (car x) 'deferred)
		 (cons (cons 'immediate 0) (cons (cons 'rb (cadr x)) alist)))
	    ((eq (car x) 'internalentry)
		(cons (cons 'immediate 
			(wdifference (jumpaddressvalue x) nearestentry*))
				(cons (cons 'rb '(reg r27)) alist)))
	    ((eq (car x) 'displacement)
		     (cons (cons 'immediate (addressvalue (caddr x)))
			   (cons (cons 'rb (cadr x)) alist)))
	    ((eq (car x) 'indexed)
		     (cons (cons 'rb (caddr x))
			   (cons (cons 'rb (cadr x)) alist)))
	    ((eq (car x) 'idloc)
                (cons (cons 'immediate
			(wand (addressvalue (list 'idloc (cadr x))) 16#ffff))
                        (cons (cons 'rb '(reg null)) alist)))
	    (t (matcherror "based" x)))))

(put 'imm '*matchfn* 'imm-match)

(de imm-match (x alist)
  (prog (add)
     (return  
	(if (setq add (addressvalue x))  
		(cons (cons 'immediate (wand add 16#ffff)) alist)
	     (matcherror "immediate" x)))))

(put 'simm '*matchfn* 'simm-match)

(de simm-match (x alist)
  (prog (add)
     (return  
	(if (setq add (addressvalue x))  
			(cons (cons 'immediate (wand add 16#ffff)) alist)
     (matcherror "signed immediate" x))))) 

(put 'offset '*matchfn* 'offset-match)

(de offset-match (x alist) % small branch address
  (prog (add)
     (return  
      (if (setq add (wquotient (jumpaddressvalue x) 4))  
	(cons (cons 'immediate
		(wand (wplus2 -1 add) 16#ffff)) 
		alist) % word addresses
     (matcherror "offset" x)))))

(put 'brdisp '*matchfn* 'brdisp-match)

(de brdisp-match (x alist) % jump address
  (prog (add)
     (return  
	(if (setq add (wquotient (jumpaddressvalue x) 4)) 
		(cons (cons 'brdisp (wand (isub1 add) 16#1fffff))  alist) 
     (matcherror "brdisp" x))))) 

(put 'd_rb  '*matchfn* 'd_rb-match)
  
(de d_rb-match (x alist)
	(if (and (eqcar x 'displacement) (regp (cadr x)) (null (caddr x)))
		(cons (cons 'rb (cadr x)) alist)
		 (matcherror "branch " x)))

(de matcherror (expected got)
    (error 4711 (list "Mismatch of Arguments (" expected "expected) :" got))) 

(compiletime (ds valueofreg (x)
		(if (pairp (setq cdrass (cdr (atsoc x alist))))
		     (get-register-code cdrass) cdrass)))
(compiletime (ds rb_isreg()
		(pairp (setq cdrass (cdr (atsoc 'rb alist))))))

(de *operate_format* (name alist)
   (prog (cdrass)
     (return
       (if (Rb_isreg)
	(operate_instruction_format (get name '*opcode*) 
		(valueofreg 'Ra) (valueofReg 'Rb) 0
		(get name '*OPERATE*) (valueofreg 'rc))
	(operate_instruction_format_2 (get name '*opcode*) 
		(valueofreg 'Ra) (cdr (atsoc 'imm_8 alist)) 1
		 (get name '*OPERATE*) (valueofreg 'rc))))))

(de operate_instruction_format (opcode ra rb zero function rc)
       (setq opcode (wor (wshift opcode 26)
		     (wor (wshift Ra 21)
		      (wor (wshift Rb 16)
		       (wor (wshift zero 12)
			(wor (wshift function 5) Rc))))))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
			(prin2 bittableoffset* ) (prin2 "  ")
			(print (sys2int opcode)))) 
	(store-word opcode))

(de operate_instruction_format_2 (opcode Ra lit one function Rc)
       (setq opcode (wor (wshift opcode 26)
		     (wor (wshift Ra 21)
		      (wor (wshift (wand 16#ff lit) 13)
		       (wor (wshift one 12)
			(wor (wshift function 5) Rc))))))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
			(prin2 bittableoffset* ) (prin2 "  ")
			(print (sys2int opcode)))) 
	(store-word opcode))

(de *memory_format* (name alist)
   (prog (cdrass)
     (return
	(memory_format (get name '*opcode*) 
		(valueofreg 'Ra) (valueofreg 'Rb)
		(cdr (atsoc 'immediate alist))))))

(de memory_format (opcode Ra Rb imm_offset)
  %     (when (or (wlessp imm_offset (wminus 32768))
%		(wgreaterp imm_offset 32767))
%		(stderror "compiler bug: function too long"))
       (setq opcode (wor (wshift opcode 26)
			 (wor (wshift Ra 21)
			      (wor (wshift Rb 16)
				 (wand imm_offset 16#ffff)))))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
		     (prin2 bittableoffset* ) (prin2 "  ")
		     (print (sys2int opcode))))
	(store-word    opcode))

(de *memory_format_branch* (name alist)
   (prog (cdrass)
     (return
        (memory_format (get name '*opcode*)
                (valueofreg 'Ra) (valueofreg 'Rb)
                (wshift (get name '*branch-code*) 14)))))

(de *memory_function_format* (name alist)
   (prog (cdrass)
     (return
	(memory_format (get name '*opcode*) 
		(valueofreg 'Ra) (valueofreg 'Rb)
		(cdr (atsoc 'function alist))))))

(de memory_function_format (opcode Ra Rb function)
       (setq opcode (wor (wshift opcode 26)
			 (wor (wshift Ra 21)
			      (wor (wshift Rb 16)
				 (wand function 16#ffff)))))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
		     (prin2 bittableoffset* ) (prin2 "  ")
		     (print (sys2int opcode))))
	(store-word    opcode))

(de *branch_format* (name alist)
   (prog (cdrass)
	(branch_format (get name '*opcode*) 
		(valueofreg 'Ra) (cdr (atsoc 'brdisp alist)))))

(de branch_format (opcode ra brdisp)
	(setq opcode (wor (wshift opcode 26) 
			(wor (wshift ra 21) brdisp)))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
			(prin2 bittableoffset* ) (prin2 "  ")
			(print (sys2int opcode))))
	(store-word    opcode))
			
(define-instruction oddalign  0 (*apply* expand-oddalign))

(de expand-oddalign (in)
     (when (eq (wand currentoffset* (wnot 7)) currentoffset*)
	(setq currentoffset* (wplus2 currentoffset* addressingunitsperitem))))

(define-instruction wordalign  0 (*apply* expand-wordalign))

(de expand-wordalign (in)
     (when (neq (wand currentoffset* (wnot 7)) currentoffset*)
        (setq currentoffset* (wplus2 currentoffset* 4))))

%%      move-relocatable-object pseudo instructions

(define-instruction move-relocatable-object 8 (*apply* movex-gen))

(de movex-gen (name src reg)
	  (cond (*writingfaslfile
	     (store-word-expression src) %% when writing to faslfile, generate
	     (relocatable-movex-gen reg)) %% special code for relocation.
	   (t (depositinstruction `(orh (hi ,src) (reg 0) ,reg)) 
					 %% otherwise normal code.
	      (depositinstruction `(or (lo ,src) ,reg ,reg)))))

(define-instruction  move-relocatable-object-and-load 8 (*apply* movex-load))

(de movex-load (name reg src )
   (cond (*writingfaslfile
	   (store-word-expression  src)  %% generate
	   (relocatable-movex-load  reg)) %% special code for relocation.
	 (t (depositinstruction 
		`(ldah ,reg (displacement (reg symvalcenter) (adjhi ,src))))
				 %% otherwise normal code.
	    (depositinstruction
		 `(ldq ,reg (displacement ,reg (adjlo ,src)) )))))

(define-instruction move-relocatable-object-and-store  
			8 (*apply* movex-store))

(de movex-store (name reg src cont)
  (cond (*writingfaslfile
	   (store-word-expression src) %% generate
	   (relocatable-movex-store reg cont)) %% code for relocation.
	(t (depositinstruction 
		`(ldah ,reg (displacement (reg symvalcenter) (adjhi ,src))))
					  %% otherwise normal code.
	   (depositinstruction
		 `(stq ,cont (displacement ,reg (ajdlo ,src)))))))

(de relocatable-movex-gen (reg)
  (let ((regcode (get-register-code reg))
	(mov-instr 2#111001))
	(store-relocatable-movex            % or instr 
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21) (wshift regcode 16) )))))

(de relocatable-movex-store (reg cont)
  (let ((regcode (get-register-code reg))
	(regcode1 (get-register-code cont))
	(mov-instr 8#000111))  %st!.l
       (store-relocatable-movex 
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21)
			 (wor (wshift regcode1 16) 1))))))

(de relocatable-movex-load (reg)
  (let ((regcode (get-register-code reg))
	(mov-instr 16#101))
	(store-relocatable-movex                      % ld.!l  immediate
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21)
			 (wor (wshift regcode 16) 1))))))

(loadtime
(dm define-instr (u)
   (prog (name opcode fmt extra)
	(setq name (cadr u))
	(setq opcode (caddr u))
	(setq fmt '(*memory_format* ra based))
	(when (cdddr u)
		(if (pairp (cadddr u))
			(cond ((eqcar (cadddr u) '*operate*)
				(setq extra 
				 `(put ',name '*operate* ,(cadr (cadddr u))))
				(setq fmt '(*operate_format* ra rb rc)))
			      ((eqcar (cadddr u) '*branch-code*)
				(setq extra 
				`(put ',name '*branch-code* ,(cadr (cadddr u))))
				(setq fmt '(*memory_format_branch* ra d_rb))))
			(when (eq (cadddr u) '*branch*)
				(setq fmt '(*branch_format* ra brdisp)))))
	(return `(progn ,extra (put  ',name '*opcode* ,opcode)
		 (put ',name '*format* ',fmt)
		 (put ',name '*op-length* 4)))))
)

% Memory Format Instructions

% op ra,disp(rb)

(define-instr   ldl    16#28)
(define-instr   ldq    16#29)
(define-instr   ldl_l  16#2a) 
(define-instr   ldq_l  16#2b)
(define-instr   ldq_u  16#0b)
(define-instr   lda    16#08)
(define-instr   stl    16#2c)
(define-instr   stq    16#2d)
(define-instr   stl_c  16#2e)
(define-instr   stq_c  16#2f)
(define-instr   stq_u  16#0f)
(define-instr   ldah   16#09)
(define-instr   ldf    16#20)
(define-instr   ldg    16#21)
(define-instr   lds    16#22)
(define-instr   ldt    16#23)
(define-instr   stf    16#24)
(define-instr   stg    16#25)
(define-instr   sts    16#26)
(define-instr   stt    16#27)

% Memory Format Instructions with Function Code

(define-instr   fetch   16#18 (*function-code* 16#8000))
(define-instr   rc      16#18 (*function-code* 16#e000))         
(define-instr   trapb   16#18 (*function-code* 16#0000))         
(define-instr   fetch_m 16#18 (*function-code* 16#a000))         
(define-instr   rpcc    16#18 (*function-code* 16#c000))         
(define-instr   mb      16#18 (*function-code* 16#4000))         
(define-instr   rs      16#18 (*function-code* 16#f000))         

% Memory Format Branch Instructions

% op ra,(rb)

(define-instr   jmp     16#1a (*branch-code* 0))       
(define-instr   jsr     16#1a (*branch-code* 1))       
(define-instr   ret     16#1a (*branch-code* 2))       
(define-instr   jsr_coroutine  16#1a (*branch-code* 3))       

% Branch Format Instructions

% op ra,disp

(define-instr   br      16#30 *branch*) 
(define-instr   bsr     16#34 *branch*)
(define-instr   blbc    16#38 *branch*) 
(define-instr   blbs    16#3c *branch*)
(define-instr   fbeq    16#31 *branch*) 
(define-instr   fbne    16#35 *branch*)
(define-instr   beq     16#39 *branch*) 
(define-instr   bne     16#3d *branch*) 
(define-instr   fblt    16#32 *branch*) 
(define-instr   fbge    16#36 *branch*) 
(define-instr   blt     16#3a *branch*) 
(define-instr   bge     16#3e *branch*) 
(define-instr   fble    16#33 *branch*) 
(define-instr   fbgt    16#37 *branch*)
(define-instr   ble     16#3b *branch*) 
(define-instr   bgt     16#3f *branch*) 

% Operate Format Instructions

%  op ra,rb,rc  or
%  op ra,#i,rc

(define-instr   addl    16#10 (*operate* 16#00))
(define-instr   addl!/v 16#10 (*operate* 16#40))
(define-instr   addq    16#10 (*operate* 16#20))
(define-instr   addq!/v 16#10 (*operate* 16#60))
(define-instr   cmpule  16#10 (*operate* 16#3d))
(define-instr   cmpbge  16#10 (*operate* 16#0f))
(define-instr   s4addl  16#10 (*operate* 16#02))
(define-instr   s4addq  16#10 (*operate* 16#22))
(define-instr   subl    16#10 (*operate* 16#09))
(define-instr   subl!/v 16#10 (*operate* 16#49))
(define-instr   subq    16#10 (*operate* 16#29))
(define-instr   subq!/v 16#10 (*operate* 16#69))
(define-instr   s4subl  16#10 (*operate* 16#0b))
(define-instr   s4subq  16#10 (*operate* 16#2b))
(define-instr   cmpeq   16#10 (*operate* 16#2d))
(define-instr   cmplt   16#10 (*operate* 16#4d))
(define-instr   cmple   16#10 (*operate* 16#6d))
(define-instr   cmpult  16#10 (*operate* 16#1d))
(define-instr   s8addl  16#10 (*operate* 16#12))
(define-instr   s8addq  16#10 (*operate* 16#32))
(define-instr   s8subl  16#10 (*operate* 16#1b))
(define-instr   s8subq  16#10 (*operate* 16#3b))

(define-instr   and     16#11 (*operate* 16#00))
(define-instr   bic     16#11 (*operate* 16#08))
(define-instr   cmoveq  16#11 (*operate* 16#24))
(define-instr   cmovne  16#11 (*operate* 16#26))
(define-instr   cmovlbs 16#11 (*operate* 16#14))
(define-instr   bis     16#11 (*operate* 16#20))
(define-instr   ornot   16#11 (*operate* 16#28))
(define-instr   cmovlt  16#11 (*operate* 16#44))
(define-instr   cmovge  16#11 (*operate* 16#46))
(define-instr   cmovlbc 16#11 (*operate* 16#16))
(define-instr   xor     16#11 (*operate* 16#40))
(define-instr   eqv     16#11 (*operate* 16#48))
(define-instr   cmovle  16#11 (*operate* 16#64))
(define-instr   cmovgt  16#11 (*operate* 16#66))

(define-instr   sll     16#12 (*operate* 16#39))
(define-instr   extbl   16#12 (*operate* 16#06))
(define-instr   extwl   16#12 (*operate* 16#16))
(define-instr   extll   16#12 (*operate* 16#26))
(define-instr   extql   16#12 (*operate* 16#36))
(define-instr   extwh   16#12 (*operate* 16#5a))
(define-instr   extlh   16#12 (*operate* 16#6a))
(define-instr   extqh   16#12 (*operate* 16#7a))
(define-instr   sra     16#12 (*operate* 16#3c))
(define-instr   insbl   16#12 (*operate* 16#0b))
(define-instr   inswl   16#12 (*operate* 16#1b))
(define-instr   insll   16#12 (*operate* 16#2b))
(define-instr   insql   16#12 (*operate* 16#3b))
(define-instr   inswh   16#12 (*operate* 16#57))
(define-instr   inslh   16#12 (*operate* 16#67))
(define-instr   insqh   16#12 (*operate* 16#77))
(define-instr   srl     16#12 (*operate* 16#34))
(define-instr   mskbl   16#12 (*operate* 16#02))
(define-instr   mskwl   16#12 (*operate* 16#12))
(define-instr   mskll   16#12 (*operate* 16#22))
(define-instr   mskql   16#12 (*operate* 16#32))
(define-instr   mskwh   16#12 (*operate* 16#52))
(define-instr   msklh   16#12 (*operate* 16#62))
(define-instr   mskqh   16#12 (*operate* 16#72))
(define-instr   zap     16#12 (*operate* 16#30))
(define-instr   zapnot  16#12 (*operate* 16#31))

(define-instr   mull    16#13 (*operate* 16#00))
(define-instr   mulq!/v 16#13 (*operate* 16#60))
(define-instr   mull!/v 16#13 (*operate* 16#40))
(define-instr   umulh   16#13 (*operate* 16#30))
(define-instr   mulq    16#13 (*operate* 16#20))

%
% Basic Predicates
%

(de alpha-immediate-p(r)
  (cond ((fixp r) t)
	((stringp r) t)
	((idp r)  (cond ((get r '*register-code) nil)
			((get r '*f-register-code) nil)
			 (t     t)))  %% label
	(t nil)))

(de alpha-relocatable-immediate-p (r)  (eqcar r 'IMMEDIATE))

(de alpha-short-immediate-p (r)
	  (and (posintp r) (lessp r 16#100)))

(de alpha-disp-p(r)
          (and (fixp r) (geq r (minus 16#8000)) (lessp r 16#8000)))

(de alpha-huge-displacement-p (r)
	   (and (eqcar r 'displacement)
		(numberp (caddr r))
		(not (alpha-short-immediate-p (caddr r)))))

(de alpha-deferred-p (r)    (eqcar r 'deferred))

(de alpha-indirect-p (r)
  (and (pairp r)
       (or (alpha-register-p (car r))
	   (eq (car r) 'deferred)
	   (eq (car r) 'displacement))))

(de alpha-double-indirect-p (r)
   (and (eqcar r 'deferred) (alpha-indirect-p (cadr r))))

(de alpha-fluid-p (r)
	  (and (pairp r) (member (car r)  '(!$fluid !$global fluid global))))

(de alpha-absolute-p (r)
	  (or (eqcar r 'quote) (stringp r)))

(de alpha-foreignentry-p (r)
	  (eqcar r 'foreignentry))

