%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  File:        $pxc/lapopt.sl
%
%  Description: Code Optimization
%               Reformation of jump sequences
%
%  Author:      H. Melenk, ZIB Berlin
%  Date:        22 Dec 1986
%
% Copyright (c) 1987,1992  Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (progn
	(on fast-integers)
	(flag '(operands LapOptRemoveToteHosen lookForLabelForwards
	 lookForLabelBackwards label-from  read-label goodjump?)
		 'internalfunction)))
 
(fluid '(*unconditionalJumps* *goodjumps* % *badjumps*
         *nonMovables* *nonMovables1* *alljumps* *lapopt))
 
(setq *unconditionalJumps* '(BV BB BLE BL))

(setq *goodjumps* '(BV BB BLE BL COMB COMIB))

(de goodjump? (in)
    (and (pairp in) (or (memq (car in) *unconditionalJumps*)
			(and (memq (car in) '(COMB COMIB))
			     (equal (cadr in) '(COMPL TR))))))
		
(setq *nonMovables* (append *goodjumps!* '(
 MOVE-RELOCATABLE-OBJECT MOVE-RELOCATABLE-OBJECT-and-load vshd
 MOVE-RELOCATABLE-OBJECT-and-store nop *entry call)))

(setq *nonMovables1* (append *goodjumps!* '(vshd nop *entry call)))

(setq *alljumps* (append *goodjumps!* *unconditionalJumps*))

(compiletime (ds NOPp (x) ((lambda(h) (and (pairp h)(eq (car h) 'NOP))) x)))

(de LapOptReformJumps (code)
  (let (rcode rrcode inst inst2 x y z cond jump inv lab lastinst delayslot)
       (when (NOT *LAPOPT) (return code)) % Main switch

% first phase
% change 		(any (without reg))
%                       (ldw .. (reg))
%                       (any (using reg))
% into (ldw ..) (any ..) (any ..)

      (while code
         (setq inst (pop code))
         (when (and (pairp inst) rcode (pairp (setq lastinst (car rcode)))
                (or (eqcar lastinst 'ldw) (eqcar lastinst 'ldwx))
		(member (caddr lastinst) (operands inst))
                (cdr rcode) (pairp (setq inst2 (cadr rcode)))
                (not (member (lastcar lastinst) (operands inst2)))
                (or (atom (cddr rcode)) (atom (caddr rcode))
                    (not (member (caaddr rcode) *alljumps*)))
                % dont "steal" from delay slot
                (or (atom (cdr inst2)) 
                    (not (member (lastcar inst2) (operands lastinst))))
                % target of first instr must not be needed for second
                (not (and (or (eqcar inst2 'stwm) (eqcar inst2 'stw))
                          (equal (lastcar inst2) (cadr lastinst))))
                % no intermix of load and store of same object
                (not (member (car inst2) *nonMovables1*)))
                (setq x (pop rcode)) (setq y (pop rcode))
%	(print (list "tausche" inst x y))
                (push x rcode) (push y rcode))
        (push inst rcode))

      (setq code (reversip rcode)) (setq rcode nil)

        % second phase: look for jumps to jumps
     (setq delayslot 0)
     (while code
          (setq inst (pop code))
          (while(and (pairp inst)    % jump to jump ?
                (memq (car inst) *goodJumps*)
                (setq lab (label-from inst))  
                code (equal (car code) '(NOP))
                (setq x (or (lookForLabelForwards code lab)
                            (lookForLabelBackwards rcode lab)))
                (eqcar (car x) 'BB)
                (NOPp (cadr x) )
              (setq inst (reverse (cons (lastcar (car x))
					 (cdr (reverse inst)))))))
             %(setq inst (cons (car inst) (cdar x)))    
          (push inst rcode))
     (setq code (reversip rcode)) (setq rcode (setq rrcode (cons nil nil))) 
          % third phase: fill in delayed instructions
     (while code
         (setq inst (pop code)) 
         (when (pairp inst)
		(setq delayslot (cond  	((eq delayslot 3) 2)
					((eq delayslot 2) 1)
					((eq delayslot 1) 0)
					(t 0))))
	 (while(and (pairp inst)
                (goodjump? inst)
                (setq lab (label-from inst))  
                code (NOPp  (car code) ) 
                (setq x (or (lookForLabelForwards code lab)
                            (lookForLabelForwards rcode lab)))
                (setq inst2 (car x))
                (pairp inst2)  
                (not (memq (car inst2) *nonMovables*)))
                                     % we insert a label
		(when (pairp (setq lab (cadr x)))
			(setq lab (gensym)) (rplacd x (cons lab (cdr x))))
		(setq z (copy inst))
 		(rplaca (lastpair z) lab)
            (rplacd rrcode  (setq z (list z)))
     % used to be 3
	    (setq delayslot 2) %(setq lastinst (car rrcode))
            (setq rrcode z)
            (setq inst (car x))
            (setq code (cdr code)))
        (when % last case
              % (something) (BR lbl)(NOP)
              % -> (BR lbl) (something)

              % if (something) is no label and is movable
            (and (pairp inst) (memq (car inst) *unconditionalJumps*)
                 code 
                 (NOPp (car code))
                      % if not 2nd successor of a jump
                 (eq 0 delayslot)
                 (pairp (car rrcode)) 
		 (not (member '(reg RETURNADDR) (operands (car rrcode))))
                 (not(memq (caar rrcode) *nonMovables*)))
            (setq code (cdr code))
            (setq delayslot 3) % (setq lastinst inst)
            (setq y (car rrcode)) (rplaca rrcode inst)(setq inst y)
            )
	(when (and (wgreaterp (length inst) 1) (eqcar (cadr inst) 'compl))
		 (setq delayslot 3)) 
        (rplacd rrcode  (setq z (cons inst nil)))
        (when (and (pairp inst) (memq (car inst) *alljumps*))
               (setq delayslot 3)) % this instruction has a delay slot
        (setq rrcode z))
           % done
        (LapOptRemoveToteHosen (cdr rcode))
  ))

% since the jumps are (BA hugo) and not (ba (label hugo)
% and it does not handle the jumps backward correctly (see tak example
(de LapOptRemoveToteHosen (code)
   (prog (l labels rcode inst prevjump1 prevjump2 alllabs)
       (setq alllabs (mapcan code
                   (function (lambda(inst) (when (atom inst) (list inst))))))
       (mapc code (function (lambda(inst)
            (when (pairp inst)
             (setq l (read-label inst alllabs))
             (when (not (member l labels)) (setq labels (cons l labels)))))))
       (while code
              (setq inst (pop code))
              (while (and (atom inst)                   % is a label
                          (not (member inst labels))    % no reference
                          prevjump2)                    % previous is a jump
                     (while (and code (pairp (setq inst (pop code))))))
              (setq prevjump2 
		(and prevjump1 (not (member '(reg r2) (operands inst)))
		       (not (member '(reg returnaddr) (operands inst)))))
              (setq prevjump1
	       (when (pairp inst)
                (or (and (eqcar inst 'BB) (equal (cadr inst) '(COND TR)))
		(equal inst '(BV (DISPLACEMENT2 (REG RETURNADDR) (REG R0))))
                (eqcar inst 'BLE))))
              (push inst rcode))
       (return(reversip rcode)) ))

(de lookForLabelForwards(code lab)
       (cond ((null code) nil)
             ((equal (car code) lab)(cdr code))
             (T (lookForLabelForwards (cdr code) lab))))
 
(de lookForLabelBackwards(code lab)
       (cond ((null code) nil) 
             ((null (cdr code)) nil)
             ((null (cddr code)) nil)
             ((equal (caddr code) lab)(list (cadr code) (car code)))
             (T (lookForLabelBackwards (cdr code) lab))))
         
(de label-from (inst) 
     (setq inst (lastcar inst))
     (if (atom inst) inst (cadr inst)))
     
(de read-label (u alllabs)
    (cond ((atom u) (if (member u alllabs) u nil))
          ((read-label (car u) alllabs))
          (T (read-label (cdr u) alllabs))))

(de operands (x)         %simplified form , only looking for regs
   % extract the operands from x
   (cond ((atom x) nil)
         ((member x '((reg zero))) nil)
         ((or (memq (car x) '(reg))) (list x))
         ((memq (car x) '(displacement2 displacement label wconst indexed
                            indexed2 indirect)) (operands (cdr x)))
         (t (nconc (operands (car x)) (operands (cdr x))))))

