%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:SUN-CMAC.SL
% Description:  Patterns and predicates for M68K PSL cmacro expansion
% Author:       G. Q. Maguire Jr. and M. L. Griss
% Created:      16 January 1982
% Modified:
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% Based on VAX-CMAC.SL of 14 Jan 1982 by Eric Benson
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%  CONVEX: 32 bit model, tags starting in 2nd bit
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% This is needed to set up real 32 bit arithmetic.
 
(loadtime (progn
        (RemProp 'Wtimes2 'OpenFn)     % So need explicit code
        (RemProp 'Wtimes2 'MemModFn)   % Since no longer a cmacro
        (RemProp '*MpyMem 'UnMemMod)   %   "    "   "    "    "
 ))  %  (RemProp 'WQuotient 'OpenCode)
     %  (RemProp 'WRemainder 'OpenCode)))
 
 
(fluid '(AddressingUnitsPerItem       % Constants defined in data machine.
     *ImmediateQuote
     ))
 
 
% The following terminal operands try to follow the same meanings as
% those outlined in the Motorola manuals.
 
(flag '(Immediate                        % #xxx
        UnImmediate                      % used for (unimmediate (immediate x))
        Indirect                         % (An)
        Indexed                          % d(An,Dn)
        absolute                         % absolute.wong
        reglist)
        'TerminalOperand)
 
 
(setf *ImmediateQuote T)
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                    NECESSARY FUNCTIONS
% These are useful macros for defining Cmarco pattern tables.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% The Second version of this function is need for cross compiliation from the
%  the vax since the vax lshift is wierd. Once the compiler is running on the
%  target machine, this module must be recompiled with the second version
%  commented out. The first version is legal for 68K's
%
%(ds BitMask (StartingBit Length)
% (land (lshift 16#ffffffff (minus StartingBit))
%    (lshift 16#ffffffff (difference BitsPerWord
%                    (plus StartingBit Length)))))
 
(ds BitMask (StartingBit Length)
  (lor (lshift 16#1 (difference Bitsperword (plus startingbit 1)))
       (land (lshift 16#7fffffff (minus StartingBit))
         (lshift 16#ffffffff (difference BitsPerWord
                    (plus StartingBit Length))))))
 
 
(ds NegMask (Length) (lshift 16#ffffffff Length))
 
(ds ShiftAmt (StartingBit Length)
  (difference BitsPerWord
          (plus StartingBit Length) %always positive.
          ))
 
(ds MakeTag (tag) (lshift (land tag 16#ff) 24))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% REGISTER HANDLING PREDICATES%
% The following predicates are for the 68000, they try to catagorize the
% constraints relating to legal operands in 68000 assembly language.
% There are a number of common predicates already available,
% they are defined the the file PC:Common-Predicates.SL
 
% The following predicates test for integers in the ranges outlined.
% All numbers are expressed in base 16.
%
% InumP        7FFFFFFF >= x >= -80000000          ( 32 bits )
% PosInumP     7FFFFFFF >= x >          0          ( 32 bits )
% NegInumP            0 >  x >= -80000000          ( 32 bits )
% Geq16P       7FFFFFFF >= x >=        10          ( 32 bits )
% LeqM16P           -10 >= x >= -80000000          ( 32 bits )
% Gt8P         7FFFFFFF >= x >          8          ( 32 bits )
% LtM8P              -8 >  x >= -80000000          ( 32 bits )
% PosHWordP       10000 >  x >          0          ( 16 bits, unsigned)
% NegHWordP           0 >  x >=    -10000          ( 16 bits, unsigned)
% DispInumP        7FFF >= x >=     -8000          ( 16 bits )
% SmallInumP         7F >= x >=       -80          (  8 bits )
 
(off R2I)            % This function should exist in the kernel!           scs
(de IntP(N) (IntP N))
(on  R2I)
 
(de PosInumP (Expression)  (and (InumP    Expression)
                                (GreaterP Expression 0)))
 
(de NegInumP (Expression)  (and (InumP  Expression)
                                (minusp Expression)))
 
(de Geq16P (Expression)    (and (InumP  Expression)
                                (Geq    Expression 16)))
 
(de LeqM16P (Expression)   (and (InumP  Expression)
                (Leq    Expression -16)))
 
(de Gt8P (Expression)      (and (InumP    Expression)
                                (GreaterP Expression 8)))
 
(de LtM8P (Expression)     (and (InumP    Expression)
                (LessP    Expression -8)))
 
(de PosHWordP (Expression) (and (IntP Expression)
                                          (GreaterP Expression 0)
                                          (lessP    Expression 16#10000)))
 
(de NegHWordP (Expression) (and (IntP Expression)
                                          (minusp Expression)
                                          (geq    Expression (minus 16#10000))))
 
(de DispInumP (Expression) (and (IntP Expression)
                                          (leq Expression 16#7fff)
                                          (geq Expression (minus 16#8000))))
 
(de SmallInumP (Expression)(and (IntP Expression)
                          (leq Expression 16#7f)
                          (geq Expression (minus 16#80))))
 
 

(de QInumP    (Expression) (and (IntP Expression)
                          (leq Expression  8)
                          (geq Expression -8)))

(De WZeroP (x) (or (Zerop x) (equal x '(wconst 0))))
 
(De QuoteP (x)    (EQCAR x 'QUOTE))
 
(De AsmModeP (x) % Legal "escapes" in Move
  (AND (PairP x) (or (Memq (car x)
                       '(fluid $fluid global $global memory extrareg))
                     (and (eqcar x 'REG) (numberp (cadr x))
                                         (greaterp (cadr x) 5)))))


(de AnySregButS7P (regName)
         (AND (RegP RegName)
              (MemQ (cadr RegName) '(S0 S1 S2 S3 S4 S5 S6 1 2 3 4 5 nil))))


%---------------------------------------------------------
% The following set of predicates describes certain classes of
% register classes. RegP tests if the ophe operand is a valid 68000 register.%
%
%  RegP  any 68000 register
% ARegP  68000 Address Register
% SregP  68000 Data Register
 
(de ZeroRegP (RegName) NIL)                                         % scs
 
 
(de FakeRegP (Expression) (and (PosIntP  Expression)
                   (GreaterP Expression LastActualReg&)))  % scs
 
% ------------------------------------------------------------
% Symbolic names for registers                                       scs
% ------------------------------------------------------------
 
(de ASregP (RegName) (or (ARegP RegName) (SregP RegName)))
 
% These two functions are used as predicates in pass-2-and-3 of the compiler
%  to find temporary registers. They need to be seperate from Aregp and Sregp
%  now that we no longer use 14 real regs. See the functions &GetAtemp and
%  &GetDtemp in comp-decls for a further explanation. Eventually we can get
%  rid of this if we decide to use patterns that ask for generic temp regs
%  instead of specific ones. /LBS
(de TempSregP (RegName) (AND (RegP RegName) (MemQ (cadr RegName)    %  scs
   '(1 2 3 4 5 6 7 t3 t4))))
 
(de TempARegP (RegName) (AND (RegP RegName) (MemQ (cadr RegName)    %  scs
   '(8 9 10 11 12 13 14 t0 t1 t2))))
 
(commentoutcode loadtime
(DefList '(
       (s0 0)         (a0 0) (st 0) (sp  0)
       (s1 1)  (t0 1) (a1 1)
       (s2 2)  (t1 2) (a2 2)
       (s3 3)  (t2 3) (a3 3)
       (s4 4)  (t5 4) (a4 4)
       (s5 5)         (a5 5)          
       (s6 6) (t3 6)  (a6 6) (NIL 6) (ap 6)
       (s7 7) (t4 7)  (a7 7) (fp 7) 
     )  'RegisterNumber)
)
 
(de RegisterNumber (RegSymbol)
% registers numbered according to D register model                        scs
  (cond ((NumberP RegSymbol)
     (remainder (ISUB1 RegSymbol) 8))
    ( T (OR (GET REGSYMBOL 'REGISTERNUMBER)
        (StdError (BldMsg "CMACS: Unknown register %r"  RegSymbol))))
    ))
 
 
%-------------
 
(de ImmediateP (x)
  (and  (EqCar x 'Immediate) (Null (fixp (cadr x)))))
 
(de TaggedLabel (X) (EqCar X 'Label))
 
 
% For powers of two we could implement (de poweroftwoP (x) (zerop (land x
%                                                           (sub1 x))))
 
(de Minus1P     (x) (equal x  -1))
(de 6P          (x) (equal x   6))
(de EightP      (x) (equal x   8))
(de 26P         (x) (equal x  26))
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       ANYREG DEFINITIONS
% These are the patterns for anyreg definitions.  They are of the form:
%
% (DefAnyreg ANYREG_NAME RESOLUTION_FUNCTION_NAME
%    ((Register_Predicate1) (Returned value))
%    ((Register_Predicate2) (returned value))
%    ...
% )
 
% The last element in the defintion has no predicate, it is a default
% case.  "Anyregs" are flagged (on their property list) with 'ANYREG.
% They have property indicators ANYREGPATTERNTABLE.  The predicate
% value list, and the function  to call - ANYREGRESOLUTIONFUNCTION.
%
% The associated function of most ANYREG's is called with 2
% arguments, a temporary REGISTER, used in the computation if needed,
% and the single SOURCE.
% The associated function usually uses (OneOperandAnyreg Register  Source
% 'Anyregname) or (ExpandOneArgumentAnyReg Register Source
% 'AnyregName). The 'AnyregName is used to find the associated pattern.
% The MEMORY anyreg has an additional argument, called ARGTWO. The
% (OneOperandAnyreg Register Source Anyregname)%
% calls the lower level ExpandOneArgumentAnyreg, after replacing Source
% by (ResolveOperand Register Source).  This function processes Source
% to see if it is. The MEMORY anyreg means that the location is the
% address of the source or destination.
%It is important that the anyregs DO NOT assign any explicit
% registers because an instruction may have two similar anyregs which
% cause the explicit register to be used twice in the same instruction.
% The future may require a scheme to allocate temporary A and D regs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
%%%%%%%%%%%%%%%%%%%%% scs %%%%%%%%%%% scs %%%%%%%%%%%%%%%%% scs %%%%%%%%%%
 
(DefAnyreg CAR
           AnyregCAR   %Grab the source so caller can displace off it.
           (       (!*Field REGISTER SOURCE InfStartingBit InfBitLength)
                   (indexed REGISTER 0))
)

(DefAnyreg CDR
           AnyregCDR     %Same as CAR, except mov to next word in pair.
           (       (!*Field REGISTER SOURCE InfStartingBit InfBitLength)
                   (indexed REGISTER 4))
)

(DefAnyreg CAR AnyregCAR   %Grab the source so caller can displace off it.
           ((aregp) (*inf SOURCE)
                    (indexed SOURCE 0))
           (        (!*move SOURCE (reg a1))
                    (*inf (reg a1))
                    (indexed (reg a1) 0))
)
 
(DefAnyreg CDR AnyregCDR   %Grab the source so caller can displace off it.
           ((aregp) (*inf SOURCE)
                    (indexed SOURCE 0))
           (        (!*move SOURCE (reg a1))
                    (*inf (reg a1))
                    (indexed (reg a1) 4))
)
  
(fluid '(caroffset* cdroffset*))

(put 'caroffset* 'constant!? t)
(put 'cdroffset* 'constant!? t)
 
(compiletime 
  (progn
    (put 'caroffset* 'constant!? nil)
    (put 'cdroffset* 'constant!? nil)
    (makeunbound 'caroffset*)
    (makeunbound 'cdroffset*)
  )
)

(de set-protected-offsets()
      (setq caroffset* (minus (times pair-tag (expt 2 26))))
      (setq cdroffset* (plus (minus (times pair-tag (expt 2 26))) 4))
)

(compiletime 
  (progn
    (put 'caroffset* 'constant!? t) 
    (put 'cdroffset* 'constant!? t) 
    (setq caroffset* (minus (times pair-tag (expt 2 26))))
    (setq cdroffset* (plus (minus (times pair-tag (expt 2 26))) 4))
  )
) 
 
(loadtime (set-protected-offsets))

(DefAnyreg CAR AnyregCAR   %Grab the source so caller can displace off it.
           ((aregp) (indexed SOURCE (wconst caroffset*)))
           (        (!*move SOURCE (reg a1))
                    (indexed (reg a1) (wconst caroffset*)))
)

(DefAnyreg CDR AnyregCDR   %Grab the source so caller can displace off it. 
           ((aregp) (indexed SOURCE (wconst cdroffset*))) 
           (        (!*move SOURCE (reg a1)) 
                    (indexed (reg a1) (wconst cdroffset*))) 
) 

% This new version is based on the old Sun 3.2 cmacro. It is designed to act
%  without the help of the memory pattern under most circumstances. Previously
%  the pattern, was doing most of work using the 14 real reg model.
%
(DefAnyreg MEMORY
           AnyregMEMORY
           ((ARegP ZeroP)         (indexed SOURCE 0))
           ((Anyp  ZeroP)      (*move SOURCE REGISTER)
                                  (indexed REGISTER 0))
           ((ARegP DispInumP)  (indexed SOURCE ARGTWO))
           ((AnyP  DispInumP)  (*move SOURCE REGISTER)
                                  (indexed REGISTER ARGTWO))
           ((RegP  InumP)      (!*move SOURCE REGISTER)
                               (!*WPLUS2 REGISTER ARGTWO)
                                  (indexed REGISTER 0))
           ((ARegP RegP)       (!*WPLUS2 SOURCE ARGTWO)
                                  (indexed SOURCE 0))
           ((SregP AnyP)       (!*move ARGTWO REGISTER)
                               (!*WPLUS2 REGISTER SOURCE)
                                  (indexed REGISTER 0))
           (( RegP AnyP)       (!*move ARGTWO REGISTER)
                               (!*WPLUS2 REGISTER SOURCE) 
                                  (indexed REGISTER 0))
           (                   (!*move SOURCE REGISTER)
                               (!*WPLUS2 REGISTER ARGTWO)
                               (indexed REGISTER 0)))
 
 
(DefAnyreg FRAME
  AnyregFRAME
       %   (         (indexed (REG FP) 
       %                (minus (plus SOURCE addressingunitsperitem) 
       %                       )))
    )
 
(de AnyregFrame (reg source)
       `(indexed (reg fp) ,(times -4 source)))

% The compiler will tag expressions immediate in the procedure ResolveWConst.
% Only expressions are tagged immediate, not numbers.
 
 
(De AnyRegImmediate(REGISTER SOURCE)
   (cond ((InumP  SOURCE)               SOURCE)
         ((Eqcar SOURCE 'Unimmediate)   SOURCE)
         (T                            (list 'immediate SOURCE) )
     ))
 
(DefAnyReg IMMEDIATE
           AnyRegImmediate)
 
 
(Defanyreg QUOTE
           AnyregQUOTE
           ((Null) (REG NIL))  %If NIL, return the NIL register (D6?)
       %   ((EqTP) (FLUID T))  %??? Is this defined?
           ((IntP)  SOURCE)    %?
           (       (QUOTE SOURCE)))   %? Recursivly expand??
 
 
% Took out the indexed from the Zero reg from the fakereg case. The
%  D reg model uses a6 as the nil reg, not the zero reg. /LBS
(commentoutcode DefAnyreg REG
           AnyregREG
           ((FakeRegP)          (*move (fluid argumentblock) REGISTER)
                  (indexed REGISTER
                           (times addressingunitsperitem SOURCE)))
           (                  (REG SOURCE)))
 
(fluid '(fluidregs*))
 
(setq fluidregs* [ &&reg6 &&reg7 &&reg8 &&reg9 &&reg10
                   &&reg11 &&reg12 &&reg13 &&reg15 &&reg 15])

(commentoutcode DefAnyreg REG
           AnyregREG
           ((FakeRegP)        (fluid (getv fluidregs* SOURCE)))
           (                  (REG SOURCE)))
 
(DefAnyreg REG AnyregREG)
 
(de AnyregREG(aux r)
     (if (and (numberp r)(greaterp r 5))
         `(fluid ,(getv fluidregs* (difference r 6)))
         `(reg ,r)))

 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                        CMACRO DEFINITIONS
% Remember that it IS legal to use Other CMACROS in expansions%
% or to define "fake" cmacros, ie procedures, for%
% Common cases. If a new C-macro is not standard,%
% e.g (*WFOO argone argtwo), then define its table and procedure.%
% You may have to define some more "anyregs" or 'TerminalOperands%
% Or define predicates to be CAREFUL.%
% (def Cmacro *WFOO%
%      ...... )%
% (de *WFOO (Arg1 Arg2)%
%   (Expand2OperandCmacro Arg1 Arg2 '*Wfoo))%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%---------------------------------------------------------------------
%   A new Cmacro to strip off the tag and place it non sign extended
%    into the argument

(de *Tag (ArgOne ArgTwo)  (Expand2OperandCmacro ArgOne ArgTwo '*Tag))

%   FORMAT: LocationToPlaceResult ArgumentToStripTag

(DefCMacro *Tag
         ((SRegP anyP)(*Move ArgTwo ArgOne)
                      (shf -26 ArgOne)
                      (and 31 ArgOne))
         ((AnyP AnySRegButS7P)(*tag (Reg S7) ArgTwo)
                      (*Move (Reg S7) ArgOne))
         (            (*Tag  (Reg S6) ArgTwo)
                      (*Move (Reg S6) ArgOne)))


(DefCMacro *Loc                           % This needs more study    scs
          ((AregP ImmediateP)(ldea (unimmediate ARGTWO) ARGONE))
          ((ARegP AnyP)      (ldea              ARGTWO  ARGONE))
          ((AnyP  ImmediateP)(*move           ARGTWO  ARGONE))
          (                  (ldea              ARGTWO (Reg T2))
                             (*move          (Reg T2) ARGONE)))
 
 
 
% Changed to use regs t0 and t1. Originally used regs t1 and t2, which caused
%  a register conflict in fastapply since it uses reg t2 to store the arg
%  list. The conflict occured when the number of args > maxnargs, since the
%  mov from the argumentblock needed reg t2, which destroyed the argument
%  list.
(de *move (Source Destination)            % redefined from COMMON-CMACROS
  (prog (resultingcode*)
    (return
    (if (and (not (equal Source '(quote NIL)))   % this is a reg too
             (not (regp Source))
             (not (regp Destination)))
        `((*move ,Source (reg s7)) (*move (reg s7) ,Destination))
     (CMacroPatternExpand
      (list (ResolveOperand '(REG t0) Source)
        (ResolveOperand '(REG t1) Destination))
      (get '*move 'CMacroPatternTable))))))
 
%   FORMAT: Source Destination
(DefCMacro *Move
         (Equal)
         ((SRegP  SRegP)       (mov!.w ArgOne ArgTwo))
         ((ASRegP  ASRegP)     (mov ArgOne ArgTwo))
         ((wzerop ASregP)      (sub!.w ArgTwo ArgTwo))
         ((InumP  ASRegP)      (ld!.w ArgOne argTwo))
         ((FixP   SRegP)       (ld!.w (lit (fullword ArgOne)) argtwo))
         ((FloatP SRegP)       (ld!.w (quote ArgOne) argtwo))
         ((ImmediateP Sregp)   (ld!.w ArgOne ArgTwo))
         ((ImmediateP Aregp)   (ld!.w ArgOne argTwo))
         ((labelp Aregp)       (ldea argone argtwo))
         ((labelp Anyp)        (ldea argone (reg A1))
                               (*move (reg A1) argtwo))
         ((IndexedP ASRegP)    (ld!.w ArgOne argtwo))
         ((ASRegP IndexedP)    (st!.w ArgOne ArgTwo))
         ((IndexedP AnyP)      (*move ArgOne (reg s6))
                               (*move (Reg s6) ArgTwo))
         ((AnyP IndexedP)      (*move ArgOne (reg s7))
                               (*move (Reg s7) ArgTwo))
         ((QuoteP ASRegP)      (ld!.w ArgOne Argtwo))
         ((QuoteP AnyP)        (*MOVE ArgOne (Reg S7))
                               (*MOVE (Reg S7) ArgTwo))
         ((AsmModeP ASRegP)    (ld!.w ArgOne ArgTwo))
         ((ASRegP AsmModeP)    (st!.w ArgOne ArgTwo))
         ((AsmModeP AnyP)      (*Move ArgOne (reg s6))
                               (*Move (Reg s6) ArgTwo))
         ((AnyP AsmModeP)      (*MOVE ArgOne (Reg s6))
                               (*move (Reg s6) ArgTwo))
         (        (*Cerror "Weird Move" ArgOne ArgTwo)))
 
(de *moveUNSBYTE (ARGONE ARGTWO)
 (Expand2OperandCMacro ARGONE ARGTWO '*moveUNSBYTE))
 
(DefCMacro *moveUNSBYTE              %  mov BYTE  ARGONE --> ARGTWO     scs
 ((regp   SregP)     (*move ARGONE ARGTWO)
                     (*wand ARGTWO 255))
 ((anyp   Sregp)     (ld!.b ArgOne ArgTwo)
                     (*wand ARGTWO 255))
 (                   (*moveUNSBYTE argone (reg S7))
                     (*move (reg S7) argtwo))
 )
 
(de *moveSGNBYTE  (ARGONE ARGTWO)
 (Expand2OperandCMacro ARGONE ARGTWO '*moveSGNBYTE))
 
(DefCMacro *moveSGNBYTE                %  mov BYTE  ARGONE --> ARGTWO     scs
 ((AREGP  SregP)     (*move   ARGONE ARGTWO)
                     (cvtb!.w  ARGTWO))
 (                   (*moveSGNBYTE (reg S7))
                     (*move (reg S7) argtwo))
 )
 
(de *moveSGNHWORD  (ARGONE ARGTWO)
 (Expand2OperandCMacro ARGONE ARGTWO '*moveSGNHWORD))
 
(DefCMacro *moveSGNHWORD        %  mov HALFWORD  ARGONE --> ARGTWO     scs
 ((INUMP  REGP)      (*move (SGNFIELD ARGONE 16 16) ARGTWO))
 ((ANYP   AREGP)     ( mov!.w ARGONE ARGTWO))
 ((ANYP   SregP)     ( mov!.w  ARGONE ARGTWO)
                     ( ext!.w ARGTWO))
 (                   ( mov!.w  ARGONE ARGTWO))
 )
 
 

(DefCMacro *push
         ((Sregp)  (psh!.w argone))
         ((Aregp)  (psh!.w argone))
         (         (*move argone (reg a1))
                   (*push (reg a1))))

(DefCMacro *pop
         ((Sregp)  (pop!.w argone))
         ((Aregp)  (pop!.w argone))
         (         (*Cerror "dont pop:" argone )))


(DefCMacro *WPLUS2
         ((Sregp Sregp)    (add!.w argtwo Argone))
         ((Aregp Aregp)    (add!.w argtwo Argone))
         ((Aregp Sregp)    (add!.w argtwo Argone))
         ((Sregp posInump) (add!.w argtwo Argone))
         ((Inump Anyp)     (*wplus2 argtwo Argone))
         ((Sregp negInump) (sub!.w (minus argtwo) argone))
         ((Aregp Inump)    (add!.w argtwo Argone))
         ((AnySregButS7P Anyp)
                           (*move ArgTwo (reg s7))
                           (*wplus2 Argone (reg s7)))
         (               (*move ArgOne (reg s6))
                         (*move ArgTwo (reg s7))
                         (*wplus2 (Reg S6) (reg s7))
                         (*move (Reg s6) Argone)))


(DefCMacro *WDIFFERENCE
         ((anyp wzerop)       )
         ((Sregp Sregp)    (Sub!.w argtwo Argone))
         ((Aregp Aregp)    (Sub!.w argtwo Argone))
         ((Sregp POSInump) (Sub!.w argtwo Argone))
         ((Sregp NEGInump) (Add!.w (minus argtwo) Argone))
         ((Aregp Inump)    (Sub!.w argtwo Argone))
         ((AnySregButS7P Anyp)
                           (*move ArgTwo (reg s7))
                           (*wdifference Argone (reg s7)))
         (              (*move ArgOne (reg s6))
                        (*move ArgTwo (reg s7))
                        (*wdifference (Reg S6) (reg s7))
                        (*move (Reg s6) Argone)))

 
%---------------------------------------------------------------------
%   Arg1 <- Arg1 * Arg2    Multiplication for integers
%   FORMAT: Destination|-Operand1 Operand2 
(DefCMacro *wTimes2
         ((AnyP OneP)        )
         ((AnyP WZeroP) (*Move 0 ArgOne))
         ((Sregp Sregp) (mul!.w argtwo Argone))
         ((Aregp Aregp) (mul!.w argtwo Argone))
         ((Sregp Inump) (mul!.w argtwo Argone))
         ((Aregp Inump) (mul!.w argtwo Argone))
         ((AnySregButS7P Anyp)
                        (*move ArgTwo (reg s7))
                        (*wtimes2 Argone (reg s7)))
         (              (*move ArgOne (reg s6))
                        (*move ArgTwo (reg s7))
                        (*wtimes2 (Reg S6) (reg s7))
                        (*move (Reg s6) Argone)))

(put 'wtimes2 'opencode '((*wtimes2 (reg 1)(reg 2))))
%---------------------------------------------------------------------
%   Arg1 <- Arg1 / Arg2
%   FORMAT: Destination|-Operand1 Operand2

(defcmacro *WQuotient
         ((AnyP OneP)        )
         ((Sregp Sregp) (div!.w argtwo Argone))
         ((Aregp Aregp) (div!.w argtwo Argone))
         ((Sregp Inump) (div!.w argtwo Argone))
         ((Aregp Inump) (div!.w argtwo Argone))
         ((AnySregButS7P Anyp)
                        (*move ArgTwo (reg s7))
                        (*wquotient Argone (reg s7)))
         (              (*move ArgOne (reg s6))
                        (*move ArgTwo (reg s7))
                        (*wquotient (Reg S6) (reg s7))
                        (*move (Reg s6) Argone)))
 
(de *WQuotient (Arg1 Arg2)
   (Expand2OperandCMacro Arg1 Arg2 '*WQuotient))


(put 'wquotient 'opencode '((*WQuotient (reg 1)(reg 2))))

(put 'wremainder 'opencode '((*move (reg 1)(reg 3))
                            (*wquotient (reg 3)(reg 2))
                            (*wtimes2 (reg 3)(reg 2))
                            (*wdifference (reg 1)(reg 3))))

(de *WNegate(ARG1)
 (Expand1OperandCMacro ARG1 '*WNegate))
 
(DefCMacro *WNegate
           ( AregP            (*WMinus ARGONE (Reg T4))
                              (*move   (Reg T4) ARGONE))
           (                  (neg!.w ARGONE ARGONE))
 )
 
(DefCMacro *WMinus                                               %  scs
           ((AnyP  InumP)     (*move (MINUS ARGTWO) ARGONE))
           ( Equal            (*WNegate ARGONE))
           ((SregP AnyP)      (*move ARGTWO ARGONE)
                              (neg!.w ARGONE ARGONE))
           ((AregP AnyP)      (*move 0 ARGONE)
                              (sub!.w ARGTWO ARGONE))
           (                  (*WMinus ARGTWO (Reg T4))
                              (*move (reg t4) ARGONE))
)
 
(de *WComplement(ARG1)
 (Expand1OperandCMacro ARG1 '*WComplement))
 
(DefCMacro *WComplement
           ( AregP            (*WNot ARGONE (Reg T4))
                              (*move (Reg T4) ARGONE))
           (                  (not ARGONE ARGONE))
 )
 
(DefCMacro *WNot
           ((AnyP  InumP)     (*move (LNOT ARGTWO) ARGONE))
           ( Equal            (*WComplement ARGONE))
           ((SregP AnyP)      (*move ARGTWO ARGONE)
                              (not ARGONE ARGONE))
           (                  (*WNot ARGTWO (Reg T4))
                              (*move (reg t4) ARGONE))
 )
 
(DefCMacro *WAnd                                               %    scs
           ( equal                )
           ((AnyP  Minus1P)       )
           ((AnyP  ZeroP)       (*move  0 ARGONE))
           ((ASregp FixP)       (and   ARGTWO ARGONE))
           ((AnyP  AregP)       (*move  ARGTWO (Reg T4))
                                (*WAND  ARGONE (Reg T4)))
           ((AregP AnyP)        (*move  ARGONE (Reg T3))
                                (*WAND  (Reg T3) ARGTWO)
                                (*move  (Reg T3) ARGONE))
           ((AnyP  NegHWordP)   (and   ARGTWO ARGONE))
           ((AnyP  SmallInumP)  (*move   ARGTWO (Reg T4))
                                (and  (Reg T4) ARGONE))
           ((SregP AnyP)        (and  ARGTWO ARGONE))
           ((AnyP  SregP)       (and  ARGTWO ARGONE))
           ((AnyP  InumP)       (and  ARGTWO ARGONE))
           (                    (*move  ARGTWO (Reg t4))
                                (and  (Reg t4) ARGONE))
)
 
(DefCMacro *WOr                                               %  scs
           ( equal                )
           ((AnyP  ZeroP)         )
           ((AnyP  Minus1P)     (*move -1 ARGONE))
           ((ASregp Inump)      (or    ARGTWO ARGONE))
           ((AnyP  AregP)       (*move ARGTWO (Reg T4))
                                (*WOR ARGONE (Reg T4)))
           ((AregP AnyP)        (*move ARGONE (Reg T3))
                                (*WOR (Reg T3) ARGTWO)
                                (*move (Reg T3) ARGONE))
           ((AnyP  PosHWordP)   (or    ARGTWO ARGONE))
           ((AnyP  SmallInumP)  (*move  ARGTWO (Reg T4))
                                (or    (Reg T4) ARGONE))
           ((SregP AnyP)        (or     ARGTWO ARGONE))
           ((AnyP  SregP)       (or     ARGTWO ARGONE))
           ((AnyP  InumP)       (or     ARGTWO ARGONE))
           (                    (*move  ARGTWO (Reg t4))
                                (or    (Reg t4) ARGONE))
)
 
(DefCMacro *WXOr                                                %  scs
           ((AnyP  ZeroP)         )
           ( equal              (*move  0 ARGONE))
           ((AnyP  Minus1P)     (*WNOT    ARGONE))
           ((ASregp Inump)      (xor   ARGTWO ARGONE))
           ((AnyP  AregP)       (*move ARGTWO (Reg T4))
                                (*WXOR ARGONE (Reg T4)))
           ((AregP AnyP)        (*move ARGONE (Reg T3))
                                (*WXOR (Reg T3) ARGTWO)
                                (*move (Reg T3) ARGONE))
           ((AnyP  PosHWordP)   (xor    ARGTWO ARGONE))
           ((AnyP  SmallInumP)  (*move  ARGTWO (Reg T4))
                                (xor    (Reg T4) ARGONE))
           ((AnyP  InumP)       (xor    ARGTWO ARGONE))
           ((AnyP  SregP)       (xor     ARGTWO ARGONE))
           (                    (*move  ARGTWO (Reg t4))
                                (xor    (Reg t4) ARGONE))
)
 
 
% ----------
% Ashift
% +index shifts left and shifts in zeroes.
% -index shifts right and sign extends.
%  ARGONE <- ARGONE shifted by ARGTWO
%------------------------------------------
 
(DefCMacro *AShift                                            %    scs
           ((anyp wzerop)           )
           ((SRegP posinump)(*wshift ARGONE ARGTWO))
           ((SRegP inump)  (*wquotient ARGONE 
                                      (wconst(expt 2 (minus ARGTWO)))))
           (     (*cerror "no variable ASHIFT available" ARGONE ARGTWO)))
 
(DefCMacro *WShift                     %Logical shift. +index=left.
           ((anyp wzerop)           )
           ((ARegP AnyP)        (shf  ARGTWO ARGONE))
           ((Sregp posInump)    (shf  ARGTWO ARGONE))
           ((SregP AnyP)        (ld!.lu (wconst 0) ARGONE)   % clear upper half first
                                (shf  ARGTWO ARGONE))
           (                    (*move ARGONE (Reg t3))
                                (*WSHIFT (Reg t3) ARGTWO)
                                (*move (Reg t3) ARGONE))
)
 
 
%---------------------------------------------------------------------
%   Jump to the argument's address
%   FORMAT: Address
(DefCMacro *Jump
         ((TaggedLabel)      (jmp  ArgOne))
         (          (*Move (PickOffIndexedReg ArgOne) (Reg a1))
                    (Jmp (Reg a1))))

%---------------------------------------------------------------------
%   Jump to the argument's address if the operands are eq.
%   FORMAT: Address Operand1 Operand2
(DefCmacro *JumpEq
         ((ARegP ARegP)      (eq!.w ArgOne ArgTwo)
                             (jmpa!.t ArgThree))
         ((ARegP smallinump) (eq!.w  ArgTwo ArgOne)
                             (jmpa!.t ArgThree))
         ((smallinump aregp) (eq!.w ArgOne Argtwo)
                             (jmpa!.t ArgThree))
         ((SRegP SRegP)      (eq!.w ArgOne ArgTwo)
                             (jmps!.t ArgThree))
         ((SRegP smallinump) (eq!.w ArgTwo ArgOne)
                             (jmps!.t ArgThree))
         ((smallinump Sregp) (eq!.w ArgOne Argtwo)
                             (jmps!.t ArgThree))
         ((smallinump anyp)  (*move ArgTwo (reg s7))
                             (eq!.w ArgOne (reg s7))
                             (jmps!.t ArgThree))
         ((anyp smallinump)  (*move ArgOne (reg s7)) 
                             (eq!.w ArgTwo (reg s7))
                             (jmps!.t ArgThree))
         ((anyp AnySregButS7P) (*move ArgOne (reg s7))
                             (eq!.w ArgTwo (reg s7))
                             (jmps!.t ArgThree))
         ((AnySregButS7P anyp) (*move ArgTwo (reg s7))
                             (eq!.w ArgOne (reg s7))
                             (jmps!.t ArgThree))
         (                   (*move ArgOne (reg s6))
                             (*move ArgTwo (reg s7))
                             (*JumpEq Argthree (reg s6) (reg s7) )))

%---------------------------------------------------------------------
%   Jump to the argument's address if the Tag(Operand) is in the range,
%    from 0 to TypeTag or is a Negative Integer Tag (-1).
%   Be careful, since the operand should not be destroyed|||
%   FORMAT: Operand TypeTag Address

(DefCmacro *jumpintype
    ((anyp wzerop)                            % intp case
                 (*Move ArgOne (Reg S6))
                 (*TAG (Reg S6)(Reg S6))
                 (*JUMPEQ ArgThree (Reg S6) 0)
                 (*JUMPEQ ArgThree (Reg S6) 31))
    (            (*Move ArgTwo (Reg S7))
                 (*Move ArgOne (Reg S6))
                 (*TAG (Reg S6)(Reg S6))
                 (*JUMPWLEQ ArgThree (Reg S6) (Reg S7))
                 (*JUMPEQ ArgThree (Reg S6) 31)) )

%---------------------------------------------------------------------
%   Jump to the argument's address if the operands are not eq.
%   FORMAT: Address Operand1 Operand2

(DefCmacro *JumpNotEQ
         ((ARegP ARegP)      (eq!.w ArgOne ArgTwo)
                             (jmpa!.f ArgThree))
         ((ARegP smallinump) (eq!.w ArgTwo ArgOne)
                             (jmpa!.f ArgThree))
         ((smallinump aregp) (eq!.w ArgOne Argtwo)
                             (jmpa!.f ArgThree))
         ((SRegP SRegP)      (eq!.w ArgOne ArgTwo)
                             (jmps!.f ArgThree))
         ((SRegP smallinump) (eq!.w ArgTwo ArgOne)
                             (jmps!.f ArgThree))
         ((smallinump Sregp) (eq!.w ArgOne Argtwo)
                             (jmps!.f ArgThree))
         ((anyp smallinump)  (*move ArgOne (reg s6))
                             (eq!.w ArgTwo (reg s6))
                             (jmps!.f ArgThree))
         ((smallinump anyp)  (*move ArgTwo (reg s6)) 
                             (eq!.w ArgOne (reg s6))
                             (jmps!.f ArgThree))
         ((anyp AnySregButS7P) (*move ArgOne (reg s7)) 
                             (eq!.w ArgTwo (reg s7)) 
                             (jmps!.f ArgThree)) 
         ((AnySregButS7P anyp) (*move ArgTwo (reg s7)) 
                             (eq!.w ArgOne (reg s7)) 
                             (jmps!.f ArgThree)) 
         (                   (*move ArgOne (reg s6))
                             (*move ArgTwo (reg s7))
                             (*JumpNotEq Argthree (reg s6) (reg s7) )))
 
%---------------------------------------------------------------------
%    from 0 to TypeTag or is not a Negative Integer Tag (-1).
%   Be careful, since the operand should not be destroyed !
%   FORMAT: Operand TypeTag Address

(DefCmacro *jumpNOTintype           
    ((anyp Wzerop)                       %intp test
                 (*Move ArgOne (Reg S7))
                 (*TAG (Reg S7)(Reg S7))
                 (*JumpEq TempLabel (Reg S7) 0)
                 (*JumpnotEq ArgThree (Reg S7) 31)
                 (*Lbl (Label TempLabel)))
    (            (*Move ArgTwo (Reg S6))
                 (*Move ArgOne (Reg S7))
                 (*TAG (Reg S7)(Reg S7))
                 (*JumpWLEQ TempLabel (Reg S7) (Reg S6))
                 (*JumpnotEQ ArgThree (Reg S7) 31)
                 (*Lbl (Label TempLabel))) )

%---------------------------------------------------------------------
%   Jump to the argument's address if the Tag(Operand) is not eq to
%    the TypeTag.
%   Be careful, since the operand should not be destroyed|||
%   FORMAT: Operand TypeTag Address

(DefCMacro *JumpNotType             
       ((TempSregp) (*Tag (Reg S7) ArgOne)
                    (*JumpNotEQ ArgThree (Reg S7) ArgTwo))
       ((Anyp Wzerop) (*Tag (Reg S6) ArgOne)
                    (*JumpNotEQ ArgThree (Reg S6) 0))
       (            (*move ArgTwo (reg S7))
                    (*Tag (Reg S6) ArgOne)
                    (*JumpNotEQ ArgThree (Reg S6) (reg s7))))

%---------------------------------------------------------------------
%   Jump to the argument's address if the Tag(Operand) is eq to
%    the TypeTag.
%   Be careful, since the operand should not be destroyed|||
%   FORMAT: Operand TypeTag Address

(DefCMacro *JumpType                
          ((TempSregp)  (*Tag (Reg S7) ArgOne)
                        (*JumpEQ ArgThree (Reg S7) ArgTwo))
          ((Anyp Wzerop)(*Tag (Reg S6) ArgOne)
                        (*JumpEQ ArgThree (Reg S6) 0))
          (             (*move Argtwo (reg S7))
                        (*Tag (Reg S6) ArgOne)
                        (*JumpEQ ArgThree (Reg S6) (reg S7))))

%---------------------------------------------------------------------
%   Jump to the argument's address if Operand1 >= Operand2
%   FORMAT: Address Operand1 Operand2

(De *JumpWGEQ (Lbl ArgOne ArgTwo)     
             (expand3operandcmacro ArgOne ArgTwo Lbl '*jumpwgeq))

(DefCmacro *JumpWGEQ                                              
         ((ANYP smallinump)  (*jumpwleq Argthree ArgTwo ArgOne))
         ((ARegP ARegP)      (lt!.w ArgOne ArgTwo)
                             (jmpa!.f ArgThree))
         ((smallinump ARegP) (lt!.w ArgOne ArgTwo)
                             (jmpa!.f ArgThree))
         ((SRegP SRegP)      (lt!.w ArgOne ArgTwo)
                             (jmps!.f ArgThree))
         ((smallinump SRegP) (lt!.w ArgOne ArgTwo) 
                             (jmps!.f ArgThree))
         (                   (*move ArgOne (reg s6))
                             (*move ArgTwo (reg s7))
                             (*JumpWgeq Argthree (reg s6) (reg s7))))

%---------------------------------------------------------------------
%   Jump to the argument's address if Operand1 > Operand2
%   FORMAT: Address Operand1 Operand2

(De *JumpWGreaterP (Lbl ArgOne ArgTwo)
      (expand3operandcmacro Argone Argtwo lbl '*jumpWgreaterp))

(DefCmacro *JumpWGreaterP                                       
         ((AnyP  smallinump) (*jumpwlessp Argthree ArgTwo ArgOne))
         ((ARegP ARegP)      (le!.w ArgOne ArgTwo)
                             (jmpa!.f ArgThree))
         ((SRegP SRegP)      (le!.w ArgOne ArgTwo)
                             (jmps!.f ArgThree))
         (                   (*move ArgOne (reg s6))
                             (*move ArgTwo (reg s7))
                          (*JumpWgreaterp Argthree (reg s6) (reg s7))))

%---------------------------------------------------------------------  
%   Jump to the argument's address if Operand1 <= Operand2
%   FORMAT: Address Operand1 Operand2

(De *JumpWLEQ (Lbl ArgOne ArgTwo)     
             (expand3operandcmacro ArgOne ArgTwo Lbl '*jumpwleq))

(DefCmacro *JumpWleq                                              
         ((ARegP ARegP)      (lt!.w ArgTwo ArgOne)
                             (jmpa!.f ArgThree))
         ((ARegP smallinump) (lt!.w ArgTwo Argone)
                             (jmpa!.f ArgThree))
         ((smallinump aregp) (le!.w Argone ArgTwo)
                             (jmpa!.t ArgThree))
         ((SRegP SRegP)      (le!.w ArgOne ArgTwo)
                             (jmps!.t ArgThree))
         ((SRegP posinump)   (lt!.w ArgTwo ArgOne)
                             (jmps!.f ArgThree))
         ((posinump Sregp)   (le!.w ArgOne ArgTwo)
                             (jmps!.t ArgThree))
         (                   (*move ArgOne (reg s6))
                             (*move ArgTwo (reg s7))
                             (*JumpWleq Argthree (reg s6) (reg s7))))

%---------------------------------------------------------------------
%   Jump to the argument's address if Operand1 < Operand2
%   FORMAT: Address Operand1 Operand2

(De *JumpWLessP (Lbl ArgOne ArgTwo)   
             (expand3operandcmacro ArgOne ArgTwo Lbl '*jumpwlessp))

(DefCmacro *JumpWLessp                                              
         ((ARegP ARegP)      (lt!.w ArgOne ArgTwo)
                             (jmpa!.t ArgThree))
         ((ARegP smallinump) (le!.w ArgTwo Argone)
                             (jmpa!.f ArgThree))
         ((smallinump aregp) (lt!.w Argone ArgTwo)
                             (jmpa!.t ArgThree))
         ((SRegP SRegP)      (lt!.w ArgOne ArgTwo)
                             (jmps!.t ArgThree))
         ((SRegP posinump)   (le!.w ArgTwo ArgOne)
                             (jmps!.f ArgThree))
         ((posinump Sregp)   (lt!.w ArgOne ArgTwo)
                             (jmps!.t ArgThree))
         (                   (*move ArgOne (reg s6))
                             (*move ArgTwo (reg s7))
                           (*JumpWlessp Argthree (reg s6) (reg s7))))

 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MkItem may be used when evaluating WConst expressions.%
% An item is formed on the CONVEX 
% ---------------------------------%
% |1| Tag |       Info field      |%
% ---------------------------------%
% 0 1      6                     31%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(commentoutcode DefCMacro *MkItem
     (              (*PUTFIELD ARGTWO ARGONE 1 6))
 )

% ASSUMING that the leading part ot the item is stil clean and the
% item is within 5 bits range:

(DefCMacro *MkItem
     ((AnyP InumP)  (*wand ARGONE (lnot (lshift 31 26)))
                    (*wor ARGONE (LSHIFT ARGTWO 26)))
     (              (*move ARGTWO (reg t3))
                    (*wshift (reg t3) 26)
                    (*wand ARGONE (lnot (lshift 31 26)))
                    (*wor ARGONE (reg t3)) ))
 
 
% Note that the arguments to *PutField are source, destination, start, size.
% This is not what the CMACRO document indicates. MkItem above has been
% modified to reflect this ordering.  This is entirely backwards to the
% majority of the CMACRO's....
 
(DefCMacro *PutField
  ((InumP ASregP AnyP  AnyP)
                     (*WAND   ARGTWO (LNOT (BITMASK  ARGTHREE ARGFOUR)))
                     (*WOR    ARGTWO (LAND (BITMASK  ARGTHREE ARGFOUR)
                             (LSHIFT ARGONE (SHIFTAMT ARGTHREE ARGFOUR)))))
 
  ((SregP SregP ZeroP AnyP)
                     (*WSHIFT ARGONE    (SHIFTAMT ARGTHREE ARGFOUR))
                     ( and  (LNOT (BITMASK  ARGTHREE ARGFOUR)) ARGTWO)
                     ( or   ARGONE ARGTWO))
  ((SregP SregP AnyP AnyP)
                     (*WSHIFT ARGONE       (SHIFTAMT ARGTHREE ARGFOUR))
                     ( and  (BITMASK  ARGTHREE ARGFOUR) ARGONE)
                     (*WAND   ARGTWO (LNOT (BITMASK  ARGTHREE ARGFOUR)))
                     (*WOR    ARGTWO ARGONE))
  ((AnyP  SregP AnyP AnyP)
                     (*move ARGONE (reg t3))
                     (*PUTFIELD    (Reg T3) ARGTWO ARGTHREE ARGFOUR))
  (                  (*move ARGTWO     (reg t4))
                     (*PUTFIELD ARGONE (Reg T4) ARGTHREE ARGFOUR)
                     (*move            (reg t4) ARGTWO))
 )
 
(DefCMacro *SignedField
  ((Sregp anyp anyp anyp)(*move   ARGTWO ARGONE)
                         (*ASHIFT ARGONE ARGTHREE)
                         (*ASHIFT ARGONE (DIFFERENCE ARGFOUR 32)))
  (                      (*SignedField (reg t3) ARGTWO ARGTHREE ARGFOUR)
                         (*move        (reg t3) ARGONE))
)
 
% *Field and *SignedField could be improved by using ROL/ROR.
 
(DefCMacro *Field
  ((ASregp anyp zerop anyp) (*move   ARGTWO ARGONE)
                           (*WSHIFT ARGONE (DIFFERENCE ARGFOUR 32)))
  ((ASregp anyp 6p 26p)
                           (*move ARGTWO ARGONE)
                           (*inf ARGONE))
  ((Sregp anyp anyp anyp)(*move   ARGTWO ARGONE)
                         ( and  (bitmask ARGTHREE ARGFOUR) ARGONE)
                         (*WSHIFT ARGONE (minus (shiftamt  ARGTHREE ARGFOUR))))
  (                      (*Field (reg t3) ARGTWO ARGTHREE ARGFOUR)
                         (*move  (reg t3) ARGONE))
)
 

(DefCmacro *inf)

(de *inf (reg)
    `((*wand ,reg (wconst 16#83ffffff))))


% ----------
% Alloc
% Allocates stack space upon procedure entry.
% ----------
 
% Redefine the *ALLOC Cmacro
% so that it fills in each stack location with NIL.
 
 
(de *ALLOC (framesize)
  (progn
    (setq NAlloc!* framesize)
    (cond
      ((ZeroP framesize)
        NIL)
      (T `((sub!.w ,(times2 framesize addressingunitsperitem) (reg st)))))))

  % a special pass in compiler will do the job

        % Otherwise, we could allocate the space and then the code to clear
        % it out.  Instead, we just push the appropriate number of NILs.

%        (for (from i 1 framesize)
%          (collect
%            '(*PUSH (REG NIL))))))))
 
% Declare *ALLOC to be a "cmacro".
% *ALLOC function handles its expansion.
 
(defcmacro *ALLOC)
 
(commentoutcode DefCMacro *Call
   ((InternallyCallableP) (calls (InternalEntry ARGONE)))
           (              (*move (entry ARGONE) (reg a1))
                          (*move (idloc ARGONE) (reg S7))
                          (calls (indexed (reg a1) 0) )))
 
(DefCMacro *Call
   ((InternallyCallableP) (calls (InternalEntry ARGONE)))
           (              (*move (idloc ARGONE) (reg S7))
                          (calls! !@ (entry ARGONE) )))


(DefCMacro *DeAlloc
   ((ZeroP))
   (                    (add!.w ARGONE (REG st))
    ))
 
(DefCMacro *Exit
   ((ZeroP)             (rtn))
                 %% st will be restaured from rtn
   (             %%%%   (add!.w ARGONE (REG st))
                        (rtn))
     )
 
(commentoutcode DefCMacro *JCall
   ((InternallyCallableP) (jmp (InternalEntry ARGONE)))
           (              (*move (entry ARGONE) (reg a1))
                          (*move (idloc ARGONE) (reg S7)) 
                          (jmp (indexed (reg a1) 0) )))
 
(DefCMacro *JCall 
   ((InternallyCallableP) (jmp (InternalEntry ARGONE))) 
           (              (*move (idloc ARGONE) (reg S7))
                          (jmp! !@ (entry ARGONE) )))
 
(DefCMacro *Jump
   ((Atom)        (jmp ARGONE))% internal labels before compile
   ((TaggedLabel) (jmp ARGONE))% compiler generated labels
   ((ImmediateP)  (jmp  (unimmediate ARGONE)))
   (              (jmp ARGONE)))
 
 
(DefCMacro *Lbl
   (              ARGONE))
 
 
%---------------------------------------------------------------------
% Linkage conventions
 
(commentoutcode de MakeLinkRegs(Fn Nargs)
     `((*Move (IdLoc ,Fn) (reg T2))))
 
(de MakeLinkRegs(Fn Nargs) nil)

%---------------------------------------------------------------------
% FOREIGN Calls: dummy for the moment

(de *ForeignLink (FunctionName FunctionType NumberofArgs)
   (append '((*push (reg nil)))
     (append(ForeignLink1 NumberofArgs)
         `((*move (reg st) (reg a6))  % argument pointer
           (*push ,NumberofArgs)
           (calls (ForeignEntry ,FunctionName))
           (*wplus2 (reg st) ,(times (add1 NumberofArgs) 4))
           (*pop (reg nil))
            ))))
           
(de *ForeignLink (FunctionName FunctionType NumberofArgs)
     (append(ForeignLink1 NumberofArgs)
         `((*move (reg st) (reg a6))  % argument pointer
           (*push ,NumberofArgs)
           (calls (ForeignEntry ,FunctionName))
           (*wplus2 (reg st) ,(times (add1 NumberofArgs) 4))
           (*move ($fluid nil) (reg nil))
            )))
           
(de ForeignLink1 (n)
   (if (eq n 0) nil
       (cons `(*push (reg ,n)) (ForeignLink1 (sub1 n)))))

(DefCMacro *ForeignLink)

%---------------------------------------------------------------------
%   FORMAT: FunctionName FunctionType NumberofArgs

(de *Link (FunctionName FunctionType NumberofArgs) 
  (cond ((FlagP FunctionName 'ForeignFunction)
        `((*ForeignLink ,FunctionName ,FunctionType ,NumberofArgs)))
        ((or (get FunctionName 'opencode)
             (get FunctionName 'exitopencode)
             (InternallyCallableP FunctionName))
                         `((*Call ,FunctionName)))
        (t  (append (MakeLinkRegs FunctionName NumberofArgs)
                         `((*Call ,FunctionName))))))
 
(DefCMacro *Link)
 
%---------------------------------------------------------------------
%   FORMAT: DeAllocCount FunctionName FunctionType NumberofArgs
 
(de *LinkE (DeAllocCount FunctionName FunctionType NumberofArgs)
   (cond ((FlagP FunctionName 'ForeignFunction)
          `((*ForeignLink ,FunctionName ,FunctionType ,NumberofArgs)
            (*Exit ,DeAllocCount)))
         ((or (get FunctionName 'exitopencode)
              (InternallyCallableP FunctionName))
                    `((*DeAlloc ,DeAllocCount)
                      (*JCall ,FunctionName)))
         ((get FunctionName 'opencode)
                    `((*Call ,FunctionName)
                      (*exit ,DeAllocCount)))
         (t  (append (MakeLinkRegs FunctionName NumberofArgs)
                     `((*DeAlloc ,DeAllocCount)
                       (*JCall ,FunctionName))))))
 
(DefCMacro *LinkE)
 
%  lisp procedure *LamBind(Regs, FLst);
%  begin scalar X, Y;
%      FLst := reverse cdr FLst;
%      Regs := reverse cdr Regs;
%      while FLst do
%      <<  if null Regs then
%      X := 0
%  else
%  <<  X := cadr car Regs;
%      Regs := cdr Regs >>;
%  Y := list('byte, 16#FF, X) . list('halfword, list('IDLoc, cadar FLst)) . Y;
%  FLst := cdr FLst >>;
%      return '(*CALL FastBind) . Y;
%  end;
 
(de *lambind (regs flst)
  (let (x y)
    (setq flst (reverse (cdr flst)))
    (setq regs (reverse (cdr regs)))
    (while flst
      (cond ((null regs) (setq x 0))
        (t (setq x (cadr (car regs)))
           (setq regs (cdr regs))))
      (setq y (cons (list (quote byte) 255 x)
            (cons (list (quote halfword)
                (list (quote idloc) (cadar flst))) y)))
      (setq flst (cdr flst)))
    (cons (quote (*call fastbind)) y)))
 
(DefCMacro *Lambind)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Since all code is movd into data space, the relative jmp used by the
% JUMPON cmacro no longer works.  It is changed here to first do an
% add.w #l0028,a0
%jmp 2(a0)
% Note this add.w must be in-line coded because the immediate won't get
% relocated otherwise (i.e., must be and address const)
% Ultimatly this belongs in HP9836-CMAC, but since it's in RLISP>???
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% FOR NOW, THIS ASSUMES 'REGISTER' IS A D REGISTER
 
(DEFCMACRO *Jumpon)
(De *Jumpon (Register Lowerbound Upperbound Labellist)
    (PROG (X LL LL2)
       (setq ll  (gensym))
       (setq ll2 (gensym))
       (SETQ X
        (if (and (weq lowerbound 0) (weq upperbound 31) *syslisp)
                  % jumpon on tags (most probably)
          `((*move ,register (reg a5))
            (mul!.w 3 (reg a5))
            (jmp (indexed (reg a5) (label ,ll2)))
            (*lbl (Label ,ll2)))
          `((lt!.w ,upperbound ,Register)
            (jmps!.t (LAbel ,ll))
            (eq!.w ,lowerbound ,Register)
            (jmps!.t ,(car Labellist))
            (lt!.w ,lowerbound ,Register)
            (jmps!.f (LAbel ,ll))
            (*move ,register (reg a5))
            (*wdifference (reg a5) ,lowerbound)
            (mul!.w 6 (reg a5))
            (jmp (indexed (reg a5) (label ,ll2)))
            (*Lbl (Label ,LL2))
       )) )
      Loop  (Setq x (nconc X `((Jmp ,(car Labellist))) ))
            (setq Labellist (cdr Labellist))
            (cond (Labellist (go loop)))

            (setq x (nconc x `((*lbl (Label ,ll))) ))
            (return x)
)     )

%-------------------------------------------------------------------------

(defcmacro *fast-apply-load
   (       (*move argone (reg t2)))       % Really reg 14. But now is fake./LBS
   )
 
(put 'fast-idapply
     'opencode
      '((*move (reg t2) (reg a1))
        (*move ($fluid SYMFNC) (reg a3))
        (*wand (reg a1) (wconst 16#ffffff))   % hard inf
        (*move (reg a1) (reg s7))             % function number
        (*wshift (reg a1) 2)                  % ID * 4
        (*wplus2 (reg a1)(reg a3))
        (*move (indexed (reg a1) 0) (reg a1)) % target address
        (calls (indexed (reg a1) 0))          % call indirect.
       ))
 
(put 'fast-idapply
     'exitopencode
      '((*move (reg t2) (reg a1)) 
        (*move ($fluid SYMFNC) (reg a3)) 
        (*wand (reg a1) (wconst 16#ffffff))   % hard inf 
        (*move (reg a1) (reg s7))             % function number 
        (*wshift (reg a1) 2)                  % ID * 4
        (*wplus2 (reg a1)(reg a3)) 
        (*move (indexed (reg a1) 0) (reg a1)) % target address
        (jmp  (indexed (reg a1) 0))          % call indirect.
       ))
 
 
% Repaired 2.Jan 1993  /HM
%

(put 'fast-codeapply
     'opencode
     '((*field (reg t2) (reg t2) (wconst infstartingbit)
           (wconst infbitlength))
       (*move (reg t2)(reg a1))
       (calls (indexed (reg a1) 0)))
     )
 
(put 'fast-codeapply
     'exitopencode
     '((*field (reg t2) (reg t2) (wconst infstartingbit)
           (wconst infbitlength))
       (*move (reg t2)(reg a1))
       (jmp (indexed (reg a1) 0)))
 
     )
 
%------------------------------------------------------------------------
%
%  Lambdind, Progbind etc.
 
(setq *unsafebinder t)   % has to save Registers across calls


(De *LamBind (Regs Fluids)
 
 % be careful, code is somewhat tricky
 
 (prog (n lng list cregs cfluids cadrcfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Regs (rest Regs))   % Remove "Registers" from the front
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Regs) 8)) % two words per BndStk entry *4
 
   %looking for free register to preload values of fluid s
 
      (setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
	               (function (lambda (x)
	                            (cond ((member x regs) nil)
	                                   (t (cons x nil))
      )              ) )         )  )
          % dont use reg t4, because that will be needed for *move (reg s7)
      (setq freeregs (append (list '(reg T3)) freeregs))
      (setq cfluids fluids) % copy of fluids
 
preload  (setq initload
	   (progn (setq cadrcfluids
	            (nconc cadrcfluids (cons (car freeregs) nil)))
	    (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))
 
       (cond ((and freeregs cfluids) (go preload)))
 
       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist
 
      (setq list `((*move ($fluid BndStkPtr) (Reg t1))
                   (*move (Reg t1) (Reg t2))
	           (*wplus2 (Reg t2) ,lng)
	           (*move ($fluid BndstkUpperBound) (Reg t5))
                   (*jumpwlessp (label ,lab) (Reg T2)(Reg t5))
                   (*call Bstackoverflow)
                 ,lab
	           (*move (Reg t2) ($fluid BndstkPtr))  )) %start of code
 
      (setq list (append initload list))
 
  loop
      (setq cregs (car Regs))
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	       (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
	 `((*move ,(car freeregs) (indexed (Reg t1) ,n))
	   (*move (quote ,Cadrcfluids)
	               (indexed (reg t1) ,(wplus2 n -4)))
	   (*move ,cregs ,cfluids)
      )          ))
      (setq fluids (cdr Fluids))
      (setq freeregs (cdr freeregs))
      (cond ((setq regs (cdr Regs)) (go loop)))
      (return list)
)    )

(defcmacro !*lambind)
 
(De *ProgBind (Fluids)
 
 % be careful, code is somewhat tricky
 
 (prog (n lng list cfluids cadrcfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 8)) % two words per BndStk entry
	                                     % * 4 addressingunits
      (setq freeregs '((reg t3)(reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
 
preload  (setq initload
	   (progn (setq cadrcfluids
	            (nconc cadrcfluids (cons (car freeregs) nil)))
	    (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))
 
       (cond ((and freeregs cfluids) (go preload)))
 
       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist
 
      (setq list `((*move ($fluid BndStkPtr) (Reg t1))
                   (*move (Reg t1) (Reg t2))
                   (*wplus2 (Reg t2) ,lng)
                   (*move ($fluid BndstkUpperBound) (Reg t5))
                   (*jumpwlessp (label ,lab) (Reg T2)(Reg t5))
                   (*call Bstackoverflow)
                  ,lab
                   (*move (Reg t2) ($fluid BndstkPtr))  )) %start of code

     (setq list (append initload list))
 
 loop
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	       (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
	         `((*move ,(car freeregs) (indexed (Reg t1) ,n))
	           (*move (quote ,Cadrcfluids)
	                       (indexed (reg t1)  ,(wplus2 n -4)))
	           (*move (reg nil) ,cfluids)
      )          ))
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *progbind)
 
(De *FreeRstr (Fluids)
 
 (prog (n lng list cfluids listfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
      (setq freeregs '((reg t3)(reg t4)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
      (setq n (wtimes2 4 (wdifference 2 lng)))
      (setq lng (wtimes2 lng 4)) % * addressingunitperitem
      (setq initload (list '(*move ($fluid Bndstkptr) (reg t1))))
 
preload  (setq initload
	   (progn (setq listfluids
	           (if freeregs
	            (nconc listfluids (cons (car freeregs) nil))
	            (nconc listfluids (cons nil nil))) )
	    (nconc initload
	     (if freeregs
%%%      `((*move (indexed (reg t1) ,(wtimes2 4 n)) ,(car freeregs))) nil)
	 `((*move (indexed (reg t1) ,n) ,(car freeregs))) nil)
       )   ))
       (setq n (wplus2 n 8))
       (when freeregs (setq freeregs (cdr freeregs)))
       (setq cfluids (cdr cfluids))
 
       (cond (cfluids (go preload)))
 
       (setq freeregs listfluids ) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and nil if not enough regs available
 
      (setq list `((*move (Reg t1) (reg t2))
                   (*wdifference (reg t2) ,lng)
	           (*move ($fluid BndstkLowerBound) (Reg t5))
                   (*jumpwgeq (label ,lab) (reg t2) (Reg t5))
                   (*call Bstackunderflow)
                  ,lab
	           (*move (Reg t2) ($fluid BndstkPtr))  )) %start of code
 
     (setq list (append initload list))
     (setq n 0)
 
 loop
      (setq cfluids (car Fluids))
      (setq n (wplus2 n 8))
 
  % insert reloaded register or memory reference
 
      (setq list (append list
	       (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
%                    `((*move (indexed (Reg t2) ,(wtimes2 4 n)) ,cfluids )))
	            `((*move (indexed (Reg t2) ,n) ,cfluids )))

      )          )
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *freerstr)

