%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  File: CRAY-LAPOPT-3
%
%  Description: Cray Code Optimization
%               Full instruction scheduling
%
%  Author:      H. Melenk, ZIB Berlin
%  Date:        16 Dec 1986
%
% Copyright (c) 1987,  Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%% instruction scheduler for Cray PSL
%%
%% machine model:
%%
%%  an instruction has n input and m output resources
%%                 (registers, memory)
%%  and it may occupy a functional unit for an individual fixed time
%%  (number of clock periods).
%%
%%  instructions which need different functional units and which have
%%  no conflicts regarding input and output resources can be performed
%%  in parallel. At any clock period one instruction be issued.
%%
%%  Conflicts regarding resources arise in the following sense:
%%  Input:   An instruction can start only if all its inputs are
%%           available. All instructions, which produce parts of the
%%           inputs must be ready.
%%  Output:  An instruction can start only it all other instructions
%%           needing the old contents of the ouput resources have
%%           been started (not necessarily finished) because they read
%%           the contents in their first clock period.
%%
%%  Basis for the scheduling algorithm is a directed network of a basic
%%  block of instructions which is established using the input and
%%  output relations.
 
%%  the structure INST describes the nodes of the network
 
 
(compiletime
  (load vector-instructions f-evectors) (on fast-integers)
 )
 
(compiletime
   (flag '(sched-build-node sched-build-network
           sched-build-Conflictset sched-build-weights
           print-code-sequence print-network
           sched-clear-dependencies set-diff-ip) 'InternalFunction))
 
(compiletime (progn
 
%(defstruct (inst :print-function print-node)
%
% % static description:
%    nbr                  % original number in the code sequence
%    instruction          % original code of the instruction
%    inputs
%    outputs
%    unit                 % unit occupied by the instruction
%    time                 % time needed
%
% % network slots:
%    (input-predecessors nil)
%    (output-predecessors nil)
%    (input-successors nil)
%    (output-successors nil)
%
% % working slots:
%    (limit  0)           % earliest time to start
%    (weight 0)           % "importance" of the instruction
%)
  (ds inst-nbr ($$x) (egetv $$x 1))
 
  (ds setf-inst-nbr ($$x $$v) (eputv $$x 1 $$v))
 
  (put 'inst-nbr 'assign-op 'setf-inst-nbr)
 
 
  (ds inst-instruction ($$x) (egetv $$x 2))
  (ds setf-inst-instruction ($$x $$v) (eputv $$x 2 $$v))
  (put 'inst-instruction 'assign-op 'setf-inst-instruction)
 
 
  (ds inst-inputs ($$x) (egetv $$x 3))
  (ds setf-inst-inputs ($$x $$v) (eputv $$x 3 $$v))
  (put 'inst-inputs 'assign-op 'setf-inst-inputs)
 
  (ds inst-outputs ($$x) (egetv $$x 4))
  (ds setf-inst-outputs ($$x $$v) (eputv $$x 4 $$v))
  (put 'inst-outputs 'assign-op 'setf-inst-outputs)
 
  (ds inst-unit ($$x) (egetv $$x 5))
  (ds setf-inst-unit ($$x $$v) (eputv $$x 5 $$v))
  (put 'inst-unit 'assign-op 'setf-inst-unit)
 
  (ds inst-time ($$x) (egetv $$x 6))
  (ds setf-inst-time ($$x $$v) (eputv $$x 6 $$v))
  (put 'inst-time 'assign-op 'setf-inst-time)
 
 
  (ds inst-input-predecessors ($$x) (egetv $$x 7))
  (ds setf-inst-input-predecessors ($$x $$v) (eputv $$x 7 $$v))
  (put 'inst-input-predecessors 'assign-op 'setf-inst-input-predecessors)
 
  (ds inst-output-predecessors ($$x) (egetv $$x 8))
  (ds setf-inst-output-predecessors ($$x $$v) (eputv $$x 8 $$v))
  (put 'inst-output-predecessors 'assign-op 'setf-inst-output-predecessors)
 
 
  (ds inst-input-successors ($$x) (egetv $$x 9))
  (ds setf-inst-input-successors ($$x $$v) (eputv $$x 9 $$v))
  (put 'inst-input-successors 'assign-op 'setf-inst-input-successors)
 
  (ds inst-output-successors ($$x) (egetv $$x 10))
  (ds setf-inst-output-successors ($$x $$v) (eputv $$x 10 $$v))
  (put 'inst-output-successors 'assign-op 'setf-inst-output-successors)
 
  (ds inst-limit ($$x) (egetv $$x 11))
  (ds setf-inst-limit ($$x $$v) (eputv $$x 11 $$v))
  (put 'inst-limit 'assign-op 'setf-inst-limit)
 
 
  (ds inst-weight ($$x) (egetv $$x 12))
  (ds setf-inst-weight ($$x $$v) (eputv $$x 12 $$v))
  (put 'inst-weight 'assign-op 'setf-inst-weight)
 
 
  (ds make-inst ($$nbr $$instr $$unit $$time)
    (let ((stru
             (if *squirrel*
                 (prog1
                    (car *squirrel*)
                    (setq *squirrel* (cdr *squirrel*)))
                 (mkevector 12 'inst))))
          (vsetvl 12)
          (Vset (Vreg 0) nil)
          (Vsetelement (Vreg 0) 0 $$nbr)
          (Vsetelement (Vreg 0) 1 $$instr)
          (Vsetelement (Vreg 0) 4 $$unit)
          (Vsetelement (Vreg 0) 5 $$time)
          (Vsetelement (Vreg 0) 10 0)   % limit
          (Vsetelement (Vreg 0) 11 0)   % weight
          (Vstore (Vreg 0) (aplus2 stru 2) 1)
          stru ))
 
 
  (ds commonlispmapc (a b) (mapc b a))
  (ds princ (x) (prin2 x))
  (ds consp (x) (pairp x))
 
 
  (defmacro  defun (u . v) `(de ,u . ,v))
  (defmacro  funcall (u . v) `(idapply ,u (list . ,v)))
  (defmacro  loop (u . v) `(while t ,u . ,v))
 
 
  (ds commonlisppush($a $b)
   (setf $b (cons $a $b)))
 
  (fluid '(*debugSched* *lapopttrace *squirrel*))
))
 
           % set difference in place
(defun set-diff-ip(b a)
   (prog()
 loop
    (when (null a) (return b))
    (when (member (car a) b) (setq b (deletip (car a) b)) )
    (setq a (cdr a))
    (go loop) ))
 
%%%%%%%%%%%%%%%%% end of common lisp interface %%%%%%%%%%%%%%%%%%%%%%%%%
 
(defun LapOptSched (code)
    (prog (rcode block1 block2 block3 found x y z time
           lcode instr fn cs debug trace *squirrel*)
 
     (setq debug (eq *debugsched* t))
     (setq trace (eq *lapopttrace t))
     (when debug (setq trace t))        % any case
           %% 1st phase% isolate basic block of optimization
 
start
    (setq found t)
    (setq block1 nil)
    (setq x 1)                   % counter
    (setq y t)                   % indicator of triviality
 
    (loop  (when (not (and code found)) (return nil))
           (setq found nil)
           (setq instr (car code))
           (setq fn (sched-build-node instr x))
           (when fn
                 (setq found t)
                 (setq x (+ 1 x))
                 (setq block1 (cons fn block1))
                 (when y
                       (setq y (= (inst-time fn) 1)))
                 (setq code (cdr code)) )
        )
 
                                   % empty block?
     (when (null block1) (go empty-block))
 
                                   % trivial block (1 instr only)?
     (when (= x 2) (setq rcode
                          (cons
                            (inst-instruction (car block1))
                            rcode))
                   (setq *squirrel* (cons (car block1) *squirrel*))
                   (go empty-block))
 
     (when (and code (pairp (car code)))
            (cond ((memq (caar code) '(JAZ JAM JAN JAP))
                     (setq y nil)
                     (setq block1 (cons
                            (sched-build-node '(dummy (Reg A0) (Reg A0))
                                         (+ x 1))
                                  block1))
                     (setq block1 (cons
                            (sched-build-node '(dummy (Reg A0) (Reg A0))
                                         (+ x 2))
                                  block1)))
                  ((memq (caar code) '(JSZ JSM JSN JSP))
                     (setq y nil)
                     (setq block1 (cons
                            (sched-build-node '(dummy (Reg S0) (Reg S0))
                                         (+ x 1))
                                   block1))
                     (setq block1 (cons
                            (sched-build-node '(dummy (Reg S0) (Reg S0))
                                         (+ x 2))
                                   block1)))))
 
     (when y
           (commonlispmapc
               #'(lambda (p) (setq rcode (cons (inst-instruction p) rcode)))
               (reversip block1))
           (setq *squirrel* (nconc block1 *squirrel*))
           (go empty-block))
 
        (when trace
           (princ "Scheduler:") (terpri)
           (princ "basic block:") (terpri)
           (print-code-sequence (reverse block1))  )
 
           %%         phase 2: build network
 
 
       (setq block2 (Sched-build-Network block1))
 
       (Sched-build-Weights block2)
 
 
       %%  phase 3: select instructions in optimal sequence
 
 
       (setq time 0)
       (setq lcode nil)
lp
 
       (setq time (+ 1 time))
 
       %%  establish the conflict set
       (setq cs (sched-build-conflict-set block2 time))
 
       (when debug
             (terpri)
             (princ "time ")
             (print time)
             (print-network "network" block2)
             (print-network "conflict set" cs) )
 
       (when (null cs)                 % should no happen
             (princ "conflict set empty") (terpri)
             (print-code-sequence block2)
             (return nil) )
 
       (when (null (cdr cs))           % only one instruction selectable
             (setq instr (car cs))
             (go selected))
 
        %% phase 4: selection among the candidates
       (setq x   -1
             y   cs)
 
                     % select list of longest instructions
       (loop (when (not y) (return nil))
             (setq z (inst-time (car y)))
             (when (< z 2) (setq z 2))
             (cond ((> z x)
                       (setq block3 (list (car y)) x z))
                    ((= z x)
                       (setq block3 (cons (car y) block3))) )
             (setq y (cdr y)) )
 
       (setq x   -1
             y   block3)
 
                     % select the heaviest now
       (loop (when (not y) (return nil))
             (when (> (setq z (inst-weight (car y))) x)
                   (setq instr (car y)  x z))
             (setq y (cdr y)) )
 
        %% phase 5: put instr to result and resolve dependencies
 
 
  selected
 
        (when debug
             (print-network "selected:" (list instr)))
 
        (when (not (eq (car (inst-instruction instr)) 'DUMMY))
           (setq rcode (cons (inst-instruction instr) rcode))
           (setq lcode (cons instr lcode)))
        (setq time (inst-limit instr))
        (setq block2 (delqip instr block2) )
        (when (null block2) (go block-done))
        (setq x (+ time (inst-time instr)))
        (sched-clear-dependencies instr block2 x)
                    % save the defstruct for later usage
        (setq *squirrel* (cons instr *squirrel*))
 
 
        (go lp)
 
block-done
        (when trace
            (princ "block done") (terpri)
            (print-code-sequence (reverse lcode)) )
 
empty-block
 
        (when (null code) (return (reverse rcode)))
 
        (when trace (print (car code)))
        (setq rcode (cons (car code) rcode))
        (setq code (cdr code))
        (go start)
))
 
 
(defun sched-build-node (instr x)
     (prog (fn object)
         (when (not (pairp instr)) (return nil))
         (setq fn (get (car instr) 'LapSchedFormFn))
         (when (not fn) (return nil))
         (setq object
               %        (make-inst
               %           :nbr x
               %           :instruction instr
               %           :unit (get (car instr) 'LapSchedUnit)
               %           :time (get (car instr) 'LapSchedTime)
               %        )
                (make-inst
                     x
                     instr
                     (get (car instr) 'LapSchedUnit)
                     (get (car instr) 'LapSchedTime)
         )      )
         (when (funcall fn instr object) (return object))
         (return nil)          % fn can deny form
)    )
 
(defun Sched-build-network (block1)
     (prog (block2 inputs outputs instr instr1 block3 d)
           (setq block2 nil)
           (loop (when (null block1) (return nil))
                 (setq instr (car block1))
                                % append does copying for later modify in place
                 (setq inputs (append (inst-inputs instr) nil))
                 (setq outputs (inst-outputs instr))
                 (setq block3 (cdr block1))
 
                     %% resolve find dependencies from predeceeding instrs
                 (loop
                    (when (not (and block3 (or inputs outputs)))
                          (return nil))
                        (setq instr1 (car block3))
 
                              % an input of instr make it dependent of
                              % last instruction outputting this input
                        (when (setq d (intersection
                                          inputs
                                          (inst-outputs instr1)
                                         %:test #'equal
                                                       ))
                              (commonlisppush instr1
                                    (inst-input-predecessors instr))
                              (commonlisppush instr
                                    (inst-input-successors instr1))
                                    % reduce "inputs" because the absolute
                                    % dependency has been resolved
                              (setq inputs (set-diff-ip inputs d)))
 
 
                              % maintain original sequence among instructions
                              % with respect to each output device
                        (when (setq d (intersection
                                          outputs
                                          (inst-outputs instr1)
                                         %:test #'equal
                                                       ))
                              (commonlisppush instr1
                                    (inst-output-predecessors instr))
                              (commonlisppush instr
                                    (inst-output-successors instr1))
                                    % reduce "outputs" because this is handled
                                    % now completely
                              (setq outputs (set-diff-ip outputs d)))
 
 
                              % produce an output not before all
                              % instructions using the previous
                              % value of that device are ready
                        (when         (intersection
                                          outputs
                                          (inst-inputs instr1)
                                         %:test #'equal
                                                       )
                              (commonlisppush instr1
                                    (inst-output-predecessors instr))
                              (commonlisppush instr
                                    (inst-output-successors instr1))
                           % outputs is not reduced because the predeceeding
                           % instructions of this class may be intermixed and
                           % so the depedency of all of them has to be recorded
                           % in the graph
                        )
 
 
                        (setq block3 (cdr block3))
                  )
 
                  (setq block1 (cdr block1))
                  (setq block2 (cons instr block2))
       ) 
       (return block2) ))
 
(defun sched-build-weights (u) % calc in forward order
     (cond((null u))
          (t (calculateWeight (car u))
             (sched-build-weights (cdr u)))))
 
(defun sched-build-conflict-set (block2 time)
    (prog (earliest cs instr x)
       (setq earliest 9999)
 
       (loop (when (not block2) (return nil))
             (setq instr (car block2))
             (when (and (null (inst-input-predecessors instr))
                        (null (inst-output-predecessors instr)))
                   % a candidate
                   (setq x (inst-limit instr))
                   (when (< x time) (setq x time))
                   (cond
                     ((> x earliest))           % not relevant
                     ((< x earliest) (setf cs (list instr)
                                           earliest x
                                           (inst-limit instr) x))
 
                     (t (setf cs (cons instr cs) (inst-limit instr) x))
             )     )
             (setq block2 (cdr block2))   )
       (return cs)
))
 
(defun sched-clear-dependencies(instr block2 x)
   (let ((unit (inst-unit instr)))
        (commonlispmapc
            #'(lambda (p)
                      (setf (inst-output-predecessors p)
                            (delqip instr
                                    (inst-output-predecessors p))))
            (inst-output-successors instr))
 
        (commonlispmapc
            #'(lambda (p)
                      (setf (inst-input-predecessors p)
                            (delqip instr
                                    (inst-input-predecessors p)))
                      (when (> x (inst-limit p))(setf (inst-limit p) x))  )
            (inst-input-successors instr))
 
                   % functional unit occupied
        (when unit
          (commonlispmapc
            #'(lambda (p)
                      (when (and (eq (inst-unit p) unit)
                                 (> x (inst-limit p)) )
                            (setf (inst-limit p) x)  ))
           block2))
))
 
(defun print-code-sequence (c)
     (commonlispmapc
        #'(lambda(p)
              (prin1 (inst-nbr p))
              (princ "   ")
              (print (inst-instruction p)))
        c))
 
(defun print-network (text c)
    (terpri)
    (print text)
    (commonlispmapc
        #'(lambda(p)
              (prin1 (inst-nbr p))
              (prin1 (inst-instruction p))
              (princ " l:")
              (prin1 (inst-limit p))
              (princ " t:")
              (prin1 (inst-time p))
              (princ " w:")
              (prin1 (inst-weight p))
              (princ " ip:")
              (commonlispmapc
                   #'(lambda(q)
                       (prin1 (inst-nbr q))
                       (princ " "))
                   (inst-input-predecessors p))
              (princ " op:")
              (commonlispmapc
                   #'(lambda(q)
                       (prin1 (inst-nbr q))
                       (princ " "))
                   (inst-output-predecessors p))
              (terpri)
            )
        c)
    (terpri)
)
 
