%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:SUN-ASM.SL
% Description:  SUN specific information for LAP-TO-ASM
% Author:       
% Created:      16-Jan-1982
% Modified:
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 27-May-87 (Harold Carr & Leigh Stoller)
%  Added fluid declarations of *declarebeforeuse, printexpressionformpointer*,
%  and printexpressionform*. Added global declaration of $eol$.
% 22-Apr-87 (Harold Carr)
%  Moved *foreignlink cmacro to xxx-cmac.sl.
% 20-Sep-86 (Leigh Stoller)
%  Copied from Bobcat compiler directory and modified to spit out Sun assembly
%  syntax using the 3.2 version as a guide. The jumpon function needed a jmp
%  instead of a bra so that it would not be shrunk by the assembler. Added the
%  sun-mnemonic-change-table to convert from Bobcat ops to Sun formated ops.
% 09-Aug-86 (Leigh Stoller)
%  Added two new temp A regs called t5,t6. Also modified the ordering so that
%  t0 is a1 and t6 is a5.
% 24-Jul-86 (Leigh Stoller)
%  Added print name of new Temp reg T0 to the reg deflists.
% 13-Dec-84 12:58:33 (Vicki O'Day)
%  Fixed typo in the name of the function OperandPrintIndirect.
% 21-Jun-84 (Vicki O'Day)
%  Duplicated Bob Morain's changes for HP-UX
% 26-Apr-84 08:55:22 (Brian Beach)
%  Removed duplicate definition of ASMOUT.  Updated version in LAP-TO-ASM.
% 30-Mar-84 09:31:50 (Brian Beach)
%  Added *FOREIGNLINK function from HP-CMAC, since it is used only by
%  the cross compiler.  Noted that CHIPMUNK-KLUDGE has been removed somewhere
%  along the line.  Now defined in garbage collector where it belongs.
% 07-Mar-83 09:36:42 (Brian Beach)
%  Made constant usage conform to new model.
% 15-Dec-83 13:15:49 (Tim Tillson)
%   Added header to Sam's D-register model version
% 22-Jul-83  Nancy Kendzierski
%   Changed use of "deferred" (in a comment) to "indirect".
%  1-Jul-83 Steve Lowder
%   Changed names of two procedures to reflect the changes in TLM 
%    addressing modes. PrintAutoIncrement was changed to PrintPostIncrement
%    and PrintAutoDecrement was changed to PrintPreIncrement.
%  10-Mar-83  Nancy Kendzierski
%   Added size specifications to InitFileNameFormat!*, DataFileNameFormat!*,
%    and CodeFileNameFormat!* to allow building on local disks as well as
%    an SRM.
%  25-Jan-83  Nancy Kendzierski
%    In ASMOUT, replaced EVIN with PATHIN to allow building on a "connected
%     directory" other than PHP:.
%  18-Jan-83  Gwyn Osnos
%   mods to ASMOUT for uppercase filenames
%  16-Sep-82
%   mods for OS 2.0
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load if-system))

(global '($eol$))

(Fluid '(CodeFileNameFormat* 
         DataFileNameFormat* 
         InitFileNameFormat* 
         InputSymFile* 
         OutputSymFile* 
         CommentFormat* 
         GlobalDataFileName* 
         LabelFormat* 
         ExternalDeclarationFormat* 
         ExportedDeclarationFormat* 
         FullWordFormat* 
         HalfWordFormat* 
         ReserveDataBlockFormat* 
         ReserveZeroBlockFormat* 
         DefinedFunctionCellFormat* 
         UndefinedFunctionCellInstructions* 
         MainEntryPointName* 
         CodeOut* 
         DataOut* 
         *Lower 
         ASMOpenParen* 
         ASMCloseParen* 
         ModuleName* 
         NumericRegisterNames* 
         ForeignExternList* 
         DataProcState* 
         PathIn*
         printexpressionform*
         printexpressionformpointer*
         *declarebeforeuse))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This assembler produces two files: a code file and a data file. Both  %
% of these files are assembled to reside in the HP dataspace            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%FORMATS%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq *DeclareBeforeUse t)

(setq CodeFileNameFormat* "%w.s")        % The name of the code segment file
(setq DataFileNameFormat* "d%w.s")       % postfix the "d" so that the files
                                         % [20] is a reasonable guess at aU
                                         % size specification

(setq InitFileNameFormat* "%w.init")     % [20] is a reasonable guess at a
                                         % size specification; it may need to
                                         % be changed.

(setq InputSymFile* "sun.sym")           % default for full-kernel-build
(setq OutputSymFile* "sun.sym")

(setq MainEntryPointName* '_!m!a!i!n)     % chose a simple default
                                          % main procedure name

(setq NumericRegisterNames* '[nil "d0" "d1" "d2" "d3" "d4" "d5" "d6" "d7" 
                                "a0" "a1" "a2" "a3" "a4" "a5"]) %  scs/vo

(setq LabelFormat* "%w:%n")             % Labels are in the first column
(setq CommentFormat* "# %p%n")          % Comments begin with an sharp sign
                                        % will group alphabetically

(setq ExportedDeclarationFormat* " .globl %w%n")
(setq ExternalDeclarationFormat* " .globl %w%n") % All in DATA space

(setq FullWordFormat* " .long %e%n")     % FullWord expects %e for parameter
(setq HalfWordFormat* " .word %e%n")     % Will EVAL formatter

(setq ReserveDataBlockFormat* " .skip %e%n nop%n")
% This does *not* make zero blocks, however, the Sun manuals
% promise that a.out memory is init'ed to 0s
% Changed below to be like Vax version, so heap will be in bss. bao
%(setq ReserveZeroBlockFormat* "   .comm %w,%e%n nop%n")
(setq ReserveZeroBlockFormat* "%w:  .skip %e%n nop%n")

%(put 'MkItem 'ASMExpressionFormat "%e<24+%e")   % MkItem(Tag, Inf)
%
(put 'MkItem 'ASMExpressionFormat "((%e*0x8000000)+%e)" )

%(setq DefinedFunctionCellFormat* " jmp %w%n")   %/ Must be LONG
%
(setq DefinedFunctionCellFormat* "  .word 0x4EF9%n  .long %w%n")

% SymFnc + 6*IdLoc(UndefinedFunction)
%(setq UndefinedFunctionCellInstructions*        
%       '((JSR (plus2 SymFnc 
%       (times2 6                        % For relative JMP
%             (IDLoc UndefinedFunction))))))
%
(setq UndefinedFunctionCellInstructions* 
  '((halfword 16#4EB9)    % Explicit JSR in ABS mode
    (fullword (plus2 SymFnc (times2 6 (IDLoc UndefinedFunction))))
   )
)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%LISTS and CONSTANT DEFINITIONS%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq ASMOpenParen* "(")
(setq ASMCloseParen* ")")

(DefList '((LAnd &) 
  (LOr !!)) 'BinaryASMOp)

(DefList '((nil "a6")                     % This reg always holds NIL.     
  (t0 "a1")
  (t1 "a2") 
  (t2 "a3") 
  (t3 "d6") 
  (t4 "d7") 
  (t5 "a4")
  (t6 "a5")
          (db "a5")                       % HP convention
          (sb "a6")                       % HP convention
          (sp "a7")                       % HP convention
          (st "a7")                       % Stack Pointer
  (a0 "a0")                               % Normal names
  (a1 "a1")                               % Normal names
  (a2 "a2")                               % Normal names
  (a3 "a3")                               % Normal names
  (a4 "a4")                               % Normal names
  (a5 "a5")                               % Normal names
  (a6 "a6")                               % Normal names
  (a7 "a7")                               % Normal names
  (d0 "d0")                               % Normal names
  (d1 "d1")                               % Normal names
  (d2 "d2")                               % Normal names
  (d3 "d3")                               % Normal names
  (d4 "d4")                               % Normal names
  (d5 "d5")                               % Normal names
  (d6 "d6")                               % Normal names
  (d7 "d7"))                              % Normal names
  'RegisterName)

(DefList '((nil "a6")                     % (Reg NIL)   
  (t0 "a1")
  (t1 "a2") 
  (t2 "a3") 
  (t5 "a4")
  (t6 "a5")
          (db "a5")                       % HP convention
          (sb "a6")                       % HP convention
          (sp "a7")                       % HP convention
          (st "a7")                       % Stack Pointer
  (a0 "a0")                               % Normal names
  (a1 "a1")                               % Normal names
  (a2 "a2")                               % Normal names
  (a3 "a3")                               % Normal names
  (a4 "a4")                               % Normal names
  (a5 "a5")                               % Normal names
  (a6 "a6")                               % Normal names
  (a7 "a7"))                              % Normal names
  'ARegisterName)


(DefList 
  '((t3 "d6") 
  (t4 "d7") 
  (d0 "d0")                               % Normal names
  (d1 "d1")                               % Normal names
  (d2 "d2")                               % Normal names
  (d3 "d3")                               % Normal names
  (d4 "d4")                               % Normal names
  (d5 "d5")                               % Normal names
  (d6 "d6")                               % Normal names
  (d7 "d7")) 'DRegisterName)


(setq DataProcState* 'data)

% This patch is until init files can be read
%
% (CopyD 'OldSaveUncompiledExpression 'SaveUnCompiledExpression) %Save previous
% (CopyD 'SaveUncompiledExpression 'SaveForCompilation) %/Now clobber

% This is redefined from LAP-TO-ASM
(commentoutcode
(de ASMPseudoPrintFloat (x) 
  (progn (cond
	  ((equal (cadr x) 0.00000E+000) 
	   (printf " dc.l 0%n dc.l 0%n")) 
	  ((equal (cadr x) 1.0) 
	   (printf " dc.l 0xFF00000%n dc.l 0%n")) 
	  (t 
	   (StdError 
	    (BldMsg "Floating point constant %r cannot be used" (cadr x)))))))
)

%%% This put is to associate the above code with (float x) in LAP

(put 'Float 'ASMPseudoOp 'ASMPseudoPrintFloat)

(de ASMPseudoPrintFloat (x)
  (cond ((EqN (cadr x) 0.0)
         (printf "  .long 0%n  .long 0%n")
        )
        ((EqN (cadr x) 1.0)
         (printf "  .long 0x3ff00000%n  .long 0%n")
        )
  )
)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%CODE HEADERS AND TRAILERS%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de CodeFileHeader nil                  % Pure Code Segment
  (CodePrintF "        .text%n")
  (setq DataProcState* 'data) 
  (setq ForeignExternList* nil))
                
(de DataFileHeader nil 
  (DataPrintF "        .data%n")) 

(de DataFileTrailer nil 
  nil)

(de CodeFileTrailer nil 
   nil)

(de CheckForeignExtern (Fn) 
  (cond
   ((not (Memq Fn ForeignExternList*))
    (setq ForeignExternList* (cons Fn ForeignExternList*)))))

(de CodeBlockHeader nil nil)            %/ Chuck this?
    

(de CodeBlockTrailer nil nil)
    

(de DataAlignFullWord nil nil)
  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%`%%
% PRINT PROCEDURES%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de PrintString (S) 
 (prog (n) 
    (cond
     ((neq (Size S) -1) 
      (progn (setq n (Size S)) 
	     (printf " .byte ") 
	     (PrintExpression (Indx S 0)) 
	     (for (from i 1 n 1)
		  (do (PrintByte!, (Indx S i))))
	     (PrintByte!, 0) 
	     (cond
	      ((equal (Remainder n 2) 1)
	       (PrintByte!, 0)))
	     (Terpri)
	     nil))
     (t 
      (progn 
	(printf " .byte ") 
	(PrintExpression 0) 
	(PrintByte!, 0) 
	(Terpri) nil)))))
 

(de PrintByte!, (x) 
  (progn (cond
	  ((greaterp (POSN) 40)
	   (printf "%n .byte "))
	  (t (Prin2 ","))) 
	 (PrintExpression x)))
    
(de TruncateString (S n) 
  (cond
   ((leq (Size S) n)
    S)
   (t
    (SUB S 0 n))))

(de PrintByteList (L) 
    (foreach x in L do (PrintByte x)))

(de PrintByte (x) 
  (progn (printf " .byte ") 
	 (PrintExpression x) 
	 (Prin2 $EOL$)))

(de PrintHalfWordList (L) 
    (foreach x in L do (PrintHalfWord x)))

(de PrintHalfWord (x) 
  (progn (printf " .word ") 
	 (PrintExpression x) 
	 (Prin2 $EOL$)))

(de PrintHalfWords (x) 
  (progn (RplacA PrintExpressionFormPointer* x) 
	 (printf HalfWordFormat* PrintExpressionForm*)))


(put 'HalfWord 'asmpseudoop 'ASMPseudoPrintHalfWords)

(de ASMPseudoPrintHalfWords (x) 
    (foreach Y in (cdr x) do (PrintHalfWord Y)))


(fluid '(*sun-mnemonic-change-table*))

%%% Print out an opcode id (derived from LAP)
%%% The Sun requires a number of transformations to the opcodes
%%% (which are based on the Motorola standard, ignored by Sun)
%%% Must change to lower case, take out the . separator, and
%%% change the names of certain mnemonics...

(de PrintOpcode (opcode)
  (let ((xform (atsoc opcode *sun-mnemonic-change-table*)))
    (if xform (setq opcode (cdr xform)))
    (prin2 (eval `(string 
             ,@(for (in x (delete 46 (string2list (id2string opcode))))
                    (collect (asm-char-downcase x))
               )
                  )
           )
    )
  )
)

%%% Taken from PU:chars.lsp, this routine converts a character (a small
%%% number) from uppercase to lowercase.  This version is considerably
%%% simplified from the PU: one so that it works in a bare PSL (a primitive
%%% if there ever was one)

(de asm-char-downcase (c)
  (if (not (or (lessp c (char A)) (greaterp c (char Z))))
      (plus (char !a) (difference c (char !A)))
      c
  )
)

%%% Mapping table for the mnemonics that Sun decided to "improve"

(setq *sun-mnemonic-change-table*
  '((adda!.b . add!.b)
    (adda!.w . add!.w)
    (adda!.l . add!.l)
    (addi!.b . add!.b)
    (addi!.w . add!.w)
    (addi!.l . add!.l)

    (suba!.b . sub!.b)
    (suba!.w . sub!.w)
    (suba!.l . sub!.l)
    (subi!.b . sub!.b)
    (subi!.w . sub!.w)
    (subi!.l . sub!.l)

    (andi!.b . and!.b)
    (andi!.w . and!.w)
    (andi!.l . and!.l)

    (movea!.b . mov!.b)
    (movea!.w . mov!.w)
    (movea!.l . mov!.l)
    (move!.b . mov!.b)
    (move!.w . mov!.w)
    (move!.l . mov!.l)

    (cmpa!.b . cmp!.b)
    (cmpa!.w . cmp!.w)
    (cmpa!.l . cmp!.l)
    (cmpi!.b . cmp!.b)
    (cmpi!.w . cmp!.w)
    (cmpi!.l . cmp!.l)

    (eori!.b . eor!.b)
    (eori!.w . eor!.w)
    (eori!.l . eor!.l)

    (ori!.b . or!.b)
    (ori!.w . or!.w)
    (ori!.l . or!.l)
   )
)


(de SpecialActionForMainEntryPoint nil 
  (progn (CodePrintLabel MainEntryPointName*) 
	 (CodeDeclareExported MainEntryPointName*)))


%%% Predicate to decide whether the given id or string is a valid
%%% one to the machine's assembler (though what lap-to-asm does if
%%% this predicate fails is beyond me!)

(de ASMSymbolP (x)
  (SunSymbolP (if (idp x) (id2string x) x))
)

(de SunSymbolP (x)
  (let ((n (size x))
        (c (indx x 0))
       )
    (and (geq n 0)  % empty strings not valid
         (or (and (geq c (char A)) (leq c (char Z)))
             (and (geq c (char !a)) (leq c (char !z)))
             (eqn c (char !_))
         )
         (for (from i 1 n)
              (with (rslt t))
              (while rslt)
              (do (setq c (indx x i))
                  (setq rslt (or (and (geq c (char A)) (leq c (char Z)))
                                 (and (geq c (char !a)) (leq c (char !z)))
                                 (eqn c (char !_))
                             )
                  )
              )
              (returns rslt)
         )
    )
  )
)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% You have to be careful with the adressing modes on the HP regs.%
% For example, you cannot used a D register as indirect. The address%
% mode code generators below should handle the tests of the reg types.%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de PrintNumericOperand (x) 
  (printf "#%w" x))


(de OperandPrintIndirect (x)            % (Indirect x)
  (progn (setq x (cadr x)) 
	 (PrintOperand x) 
	 (Prin2 "@")))

(put 'Indirect 'OperandPrintFunction 'OperandPrintIndirect)

(de OperandPrintDisplacement (x)        % (Displacement (reg x) disp)
   (progn (setq x (cdr x)) 
	  (PrintOperand (car x)) 
	  (Prin2 "@(")
	  (PrintExpression (cadr x)) 
	  (Prin2 ")")))

(put 'displacement 'OperandPrintFunction 'OperandPrintDisplacement)

% (Indexed (reg y)(displacement (reg x) disp))
%
(de OperandPrintIndexed (x)  % (indexed (reg y) (displacement (reg x) disp))
  (let ((part2 (third x)))
       (PrintRegister (cadr part2))
       (prin2 "@(")
       (PrintExpression (third part2))
       (prin2 ", ")
       (PrintRegister (cadr x))
       (prin2 ":L)")))

(put 'Indexed 'OperandPrintFunction 'OperandPrintIndexed)

(de OperandPrintImmediate (x)           % (Immediate x) % for ADDRESS
  (progn (prin2 "#")
	 (PrintExpression (cadr x))))

(put 'Immediate 'OperandPrintFunction 'OperandPrintImmediate)


(de OperandPrintPostIncrement (x)       % (PostIncrement x)
  (progn (PrintOperand (cadr x)) 
	 (Prin2 "@+")))

(put 'PostIncrement 'OperandPrintFunction 'OperandPrintPostIncrement)

(de OperandPrintRegList (x)             % (Reglist x)
  (progn (setq x (cdr x)) 
	 (PrintOperand (car x)) 
	 (setq x (cdr x)) 
	 (While x 
		(progn (Prin2 "/") 
		       (PrintOperand (car x)) 
		       (setq x (cdr x)))) nil))


(put 'RegList 'OperandPrintFunction 'OperandPrintRegList)

(de OperandPrintPreDecrement (x)        % (PreDecrement x)
  (progn (PrintOperand (cadr x)) 
	 (Prin2 "@-")))

(put 'PreDecrement 'OperandPrintFunction 'OperandPrintPreDecrement)

(de OperandPrintAbsolute (x)            % (Absolute x)
    (PrintExpression (cadr x)))


(put 'Absolute 'OperandPrintFunction 'OperandPrintAbsolute)

(de OperandPrintForeignEntry (x)        % (ForeignEntry FcnName)
  (let ((*lower t))
       (printf "_%w" (cadr x))))
  
(put 'ForeignEntry 'OperandPrintFunction 'OperandPrintForeignEntry)

(Fluid '(ResultingCode*))

(de MCPrint (x)                         % Echo of MC's
 (CodePrintF "# %p%n" x))

(de InstructionPrint (x) 
 (CodePrintF "#%p%n" x))

(de *cerror (x) 
 (prog (i) 
    (setq i (wrs nil)) 
    (printf "%n *** CERROR: %r %n " x) 
    (wrs i) 
    (return (list (list 'cerror x)))))
 

(put 'cerror 'asmpseudoop 'printcomment)

(DefCMacro *cerror)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% A HORRID patch for BRA/BRA!.W incompatibility between
% KERNEL and LAP
% In Kernel, BRA means word sized branch
% IN LAp and FASL BRA seems to mean bytes sized (?) even though
% must be coerced somewhere
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%@%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This version to be used ONLY in SUN-CROSS

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Since all code is moved into data space, the relative jmp used by the
% JUMPON cmacro no longer works.  It is changed here to first do an
% adda.l #l0028,a0
%jmp 2(a0)
% Note this adda.l must be in-line coded because the immediate won't get
% relocated otherwise (i.e., must be and address const)
% Ultimatly this belongs in HP9836-CMAC, but since it's in RLISP>???
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% FOR NOW, THIS ASSUMES 'REGISTER' IS A D REGISTER

(de *JumpOn (Register LowerBound UpperBound LabelList) 
  (prog (BaseLbl EndLbl Result) 
	(setq BaseLbl (Gensym)) 
	(setq EndLbl (Gensym)) 
	(setq Result (cons nil nil)) 
	(cond 
	 ((ZeroP  LowerBound)
	  (TConc Result (list 'tst!.l Register))) 
	 ((QInumP LowerBound)
	  (TConc Result (list '*WDifference Register LowerBound)))
	 (T
	  (TConc Result (list 'sub!.l LowerBound Register))))
	(TConc Result (list 'blt EndLbl)) 
	(TConc Result (list 'cmp!.l
			    (difference UpperBound LowerBound) Register)) 
	(TConc Result (list 'bgt EndLbl)) 
	(TConc Result (list 'lsl!.l 2 Register)) 
	(TConc Result (list 'HALFWORD 16#4EFB))      % JMP BaseLbl(Dn.L)
	(TConc Result (list 'HALFWORD (iplus2 16#0802 
					      (ilsh (registernumber 
						     (cadr register)) 12))))
	(TConc Result BaseLbl) 
	(foreach x in LabelList do
		 (TConc Result (list 'jmp (cadr x))))		 
	(TConc Result EndLbl) 
	(return (car Result))))

% End of file.





