%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PK:FASLIN.SL 
% Title:        Loading of binary format files. 
% Author:       E. Benson 
% Created:      ??? 
% Modified:     14-Jun-88 (Tsuyoshi Yamamoto)
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  PL:FASL-DECLS.B 
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 4-Feb-1991 (Winfried Neun)
%  alignment for code block on double word boundary
% 14-Jun-88 (Tsuyoshi Yamamoto)
%  Added new relocation types (relocate-movex,relocate-call) and long word
%  alignment for halfword access.
% 10-Jan-84 (Brian Beach)
%  Commented-in checking of fasl-magic-number.
% 13-Nov-84 14:13:11 (Brian Beach)
%  Moved FASLIN-INTERN to INTERN.SL.
% 10-May-84 14:35:05 (Brian Beach)
%  Changed fasl-magic-number to faslin-magic (which is defined now in SYS-CONSTS.)
% 01-Dec-83 14:44:33 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load fasl-decls))

(on fast-integers)

(fluid '(code-base-hack
	 symfnc
	 symval
	 tokenbuffer
	 argumentblock))

(de faslin (file)
  (prog (fid                  % file pointer 
	 local-id-count       % number of ids in the file
	 local-id-table       % table for mapping local ID numbers to global ID numbers.
	 code-size            % number of words of code
	 code-base            % location of the start of the code
	 init-function-address% Offset into the code of the init function
	 bit-table-size       % number of words in the bit table
	 bit-table            % the bit table
         Btop
	 )

    % Open the file
    (setf fid (binaryopenread file))

    % Check that the first word is the correct magic number.
    (let ((first-word (binaryread fid)))
      (unless (weq first-word faslin-magic)
	(binaryclose fid)
	(faslin-bad-file file)
	(return nil)
	))

    % Read in the ID table.
    (setf local-id-table (read-id-table fid))

    % Read the code.
    (setf code-size (wquotient (wplus2 1 (binaryread fid)) 2))
			% Size of code segment in words

 % align 'real' code on double word boundary
    (setq nextbps (wand (wplus2 nextbps 7) (wnot 7)))

    (setf code-base (gtbps code-size)) % Allocate space in BPS
    (setq Btop (GtBPS 0))              % pointer to top of alloc. BPS
    (setf init-function-address (wplus2 code-base (binaryread fid)))
    (binaryreadblock fid (loc (wgetv code-base 0))  code-size)

    % Read the bit table
    (setf bit-table-size (binaryread fid))
    (setq bit-table (mkwrds (gtwrds bit-table-size)))
    (binaryreadblock fid (loc (words-fetch bit-table 0)) bit-table-size)

    % Close the file
    (binaryclose fid)

    % Twiddle the bits.
    (do-relocation code-base code-size bit-table local-id-table)

    % Call the init code
    (let ((temp code-base-hack))  % avoid use of fluid binding
      (setf code-base-hack code-base)
      (cacheflush code-base (wdifference btop code-base) 0)
      (addressapply0 init-function-address)
      (setf code-base-hack temp)
      (DelBPS (wplus2 init-function-address 4) Btop)
      )
    ))
%
%	These 2 constants are added for SPARC. (Tsuyoshi Yamamoto)
%
(define-constant reloc-movex 1)
(define-constant reloc-call  2)

(compiletime
 (put 'half_getmem 'opencode '((ld (displacement (reg 1) 0) (reg 1)))))
(compiletime
 (put 'half_putmem 'opencode '((st (reg 2) (displacement (reg 1) 0)))))

(de do-relocation (code-base code-size bit-table id-table)
  % CODE-AU-SIZE is the size of the code measured in addressing
  % units, rather than words.

  (let ((code-au-size  (wtimes2 code-size addressingunitsperitem)))
    (for (from i 0 (wdifference code-au-size 4))  %%% should have offset 4!!  
	 (do 
	  (let ((bit-table-entry  (bittable (loc (words-fetch bit-table 0)) i))   %%% HACK!
		(code-location    (wplus2 code-base i))
                memo )
       (console-print-string "Bittable: ")
       (unixputn bit-table-entry) (console-newline)
	    (case bit-table-entry
	      ((reloc-word)
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%	Following lines are added for SPARC relocation.
	%                               May 18, 1988 (Tsuyoshi Yamamoto)
	%				
	       (case (bittable (loc (words-fetch bit-table 0)) (wplus2 i 1))
		     ((0)
		      (relocate-word code-location code-base id-table))
	             ((reloc-movex)
		      (progn (putbittable (loc (words-fetch bit-table 0))
					  (wplus2 i 1)
					  0)
			     (relocate-movex code-location code-base id-table)))
		      ((reloc-call)
                        (if (eq 0 (wshift
				   (wshift (setq memo
					 (half_getmem code-location)) 34) -50))
                                      % its a CALL
		              (progn (putbittable (loc (words-fetch bit-table 0))
					   (wplus2 i 1) 0)
				(relocate-word code-location code-base id-table)
				(relocate-call code-location code-base id-table))
                                      % its a BCC , new invention
                              (progn (putbittable (loc (words-fetch bit-table 0))
                                           (wplus2 i 1) 0)

                                (half_putmem code-location
                                      (wand memo (wnot (wshift 16#3fff 16))))
                                (relocate-word code-location code-base id-table)
                                (relocate-bcc code-location code-base memo))))
		      ))
		     %
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	      ((reloc-inf) 
	       (relocate-inf code-location code-base id-table))
	      ((reloc-right-half)
	       (relocate-right-half code-location code-base id-table))
	      ))))))
		    
		     
(de relocate-word (code-location code-base id-table)
  (let ((reloc-tag (reloc-word-tag (half_getmem code-location)))
	(reloc-inf (reloc-word-inf (half_getmem code-location))))
    (half_putmem code-location
      (compute-relocation reloc-tag reloc-inf code-base id-table)
      )))

(de relocate-inf  (code-location code-base id-table)
  (let ((reloc-tag (reloc-inf-tag (getmem code-location)))
	(reloc-inf (reloc-inf-inf (getmem code-location)))
        tmp)
      (setq tmp (compute-relocation reloc-tag reloc-inf code-base id-table))
      (putmem code-location (mkitem (tag (getmem code-location))
					(inf tmp)))))

(de relocate-right-half (code-location code-base id-table)
  (let ((reloc-tag nil)
	(reloc-inf nil)
	(temp1 nil)
	(temp2 nil))
    (cond ((weq (wand code-location 3) 0)	% longword address.
      (setq reloc-tag (reloc-right-half-tag (half_getmem code-location)))
      (setq reloc-inf (reloc-right-half-inf (half_getmem code-location)))
      (half_putmem code-location 
	(wor (wshift (half_getmem code-location) 30)
	(compute-relocation reloc-tag reloc-inf code-base id-table)
		    )))
	     (t					% halfword address.
	      (setq temp1 (half_getmem (wdifference code-location 2)))
	      (setq temp2 (wshift temp1 16))
	      (setq reloc-tag (reloc-right-half-tag temp2))
	      (setq reloc-inf (reloc-right-half-inf temp2))
	      (setq temp2 (compute-relocation reloc-tag
					      reloc-inf
					      code-base
					      id-table))
	      (setq temp1 (wor (wand temp1 16#ffff0000)
			       (wand temp2 16#0000ffff)))

	      (half_putmem (wdifference code-location 2) temp1)))))

(de compute-relocation (reloc-tag reloc-inf code-base id-table)
  (case reloc-tag
    ((reloc-code-offset) 
     (wplus2 code-base reloc-inf))
    ((reloc-value-cell) 
     (cond ((extraargumentp reloc-inf) 
	    (loc (wgetv argumentblock
			(makeextraargument reloc-inf))))
	   ((local-id-number? reloc-inf) 
	    (loc (symval (local-to-global-id reloc-inf id-table))))
	   (t
	    (loc (symval reloc-inf)))))
    ((reloc-function-cell)
     (progn
      (when (local-id-number? reloc-inf)
	(setq reloc-inf (local-to-global-id reloc-inf id-table)))
      (wplus2 symfnc                        %%% Should be (LOC (SYMFNC xxx)) ???
	      (wtimes2 addressingunitsperfunctioncell
		       reloc-inf))))
    ((reloc-id-number)
     (if (local-id-number? reloc-inf)
       (local-to-global-id reloc-inf id-table)
       reloc-inf
       ))
    ))

(de local-to-global-id (local-id-number id-table)
  (words-fetch id-table (wdifference local-id-number 2048))
  )

(de read-id-table (fid)
  % Read in the table of local IDs at the front of the FASL file.
  % Each ID is stored as one word which holds the length, followed 
  % by the appropriate number of words holding the string.

  (let* ((local-id-count  (binaryread fid))
	 (id-table        (mkwrds (gtwrds (wplus2 local-id-count 1)))))

    (for 
     (from i 0 local-id-count)
     (do (setf (wgetv tokenbuffer 0) (binaryread fid))
					 % word is length of ID name
	 (binaryreadblock fid
			  (loc (wgetv tokenbuffer 1)) 
			  (strpack (wgetv tokenbuffer 0)))
	 (setf (words-fetch id-table i) 
	   (idinf (faslin-intern (mkstr (loc (wgetv tokenbuffer 0))))))
	 ))
    id-table
    ))

(de putentry (name type offset)
  % Called by the initcode.
  % CODE-BASE-HACK is set by FASLIN before the initcode is called.
  (putd name type (mkcode (wplus2 code-base-hack offset))))

(de faslin-bad-file (name)    % To be redefined
  (console-print-string "File is not FASL format")
  (console-newline)
  )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%       These 2 relocater are added for SPARC. May 18, 1988
%
%
(de relocate-movex (code-location code-base id-table)
    (let ((val (half_getmem (wplus2 code-location -4)))
	  (add (half_getmem code-location))
	  (sethi nil))
	 (setq sethi (wor 16#01000000  % Sparc sethi instruction.
			  (wor (wand (wshift val -10) 16#003fffff)
                                  % pull the rs1 out of next instr
			       (wshift (wand add 16#7c000) 11))))
	 (setq add (wor add
		       (wand val 16#3ff)))
	 (half_putmem (wplus2 code-location -4) sethi)
	 (half_putmem code-location add)))

(de relocate-call (code-location code-base id-table)
  (let ((entry (half_getmem code-location)))
       (half_putmem code-location
	     (wor 16#40000000          %% Sparc call instruction.
		  (wand 16#3fffffff
		        (wshift (wdifference entry code-location)
				-2))))))

(de relocate-bcc (code-location code-base memo) 
 % memo contains info of bcc operation
    (let ((entry (half_getmem code-location))
          (annul (wand 1 (wshift memo -17)))
          (cond  (wand 15 (wshift memo -24)))
          (opcode (wand 7 (wshift memo -18))))
       (half_putmem code-location
             (wor (wshift annul 29)   %% Build a bcc instruction
               (wor (wshift cond 25)
                 (wor (wshift opcode 22) 
                  (wand 16#3fffff      %% disp 22
                        (wshift (wdifference entry code-location)
                                -2))))))))
%
%	End of hacking.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de delbps (bottom top)
 % returns space to Bps
       (when (weq nextbps top)
	    (cacheflush bottom (wdifference nextbps bottom) 1)
            (setq nextbps bottom)))

(off fast-integers)

