%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXK:ARITH.SL
% Description:  Arithmetic required by the kernel.
% Author:       Brian Beach
% Created:      04-Jun-84
% Modified:     
% Package:      
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 01-Apr-87 (Leigh Stoller & Harold Carr)
%  Optimized wtimes and wquotient to take advantage of 68020 instructions.
% 29-Jul-86 (Leigh Stoller)
%  Changed all references of fake reg 14 to real reg t2
% 06-Jul-84 14:51:26 (RAM)
%  Fixed wremainder.  Do not need to special case negatives as did on
%  the PWS system.
% 27-Jun-84 20:26:05 (Vicki O'Day)
%  Changed calls on ldiv, lrem and lmod to be *link instead of *jsr
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load if-system))
(compiletime (reload sys-consts))

% these routines are written so that in the normal case they use
% only registers (reg 1) (reg 2) and (reg 9)

(compiletime (flag '(lmul ldiv lrem) 'foreignfunction))

(if_system mc68020
(lap '((*entry wtimes2 expr 2)             % (wtimes2 arg1 arg2)
       (muls!.l (reg 2) (reg 1))
       (*exit 0)))

%% else not a 68020

(lap '((*entry wtimes2 expr 2)             % (wtimes2 arg1 arg2)
       (*push (reg t2))
       (movea!.w (reg 1) (reg t2))         % this sign extends 16 to 32 bits
       (cmpa!.l (reg 1) (reg t2))          % check to see that arg1 is 16 bits
       (bne!.s longmult)
       (movea!.w (reg 2) (reg t2))         % check arg2 in the same way
       (cmpa!.l (reg 2) (reg t2))
       (bne!.s longmult)
       (muls (reg 2) (reg 1))              % do the arithmetic
       (*pop (reg t2))
       (*exit 0)
  
LONGMULT                                   % later we'll do a better job 
       (movem!.l (reglist (reg d2) (reg d3) (reg d4) (reg d5) (reg d6)
			  (reg d7) (reg a0) (reg a1) (reg a2) (reg a3)
			  (reg a4) (reg a5))
                 (predecrement (reg st)))  % save all the registers
       (*link lmul expr 2)
       (movem!.l (postincrement (reg st)) % restore all the registers
		 (reglist (reg d2) (reg d3) (reg d4) (reg d5) (reg d6)
			  (reg d7) (reg a0) (reg a1) (reg a2) (reg a3)
			  (reg a4) (reg a5)))
       (*pop (reg t2))
       (*exit 0)))
) % End of if_system.

(if_system mc68020
(lap '((*entry wquotient expr 2)
       (divs!.l (reg 2) (reg 1))
       (*exit 0)))

%% else not a 68020

(lap '((*entry wquotient expr 2)
       (*push (reg t2))
       (movea!.w (reg 2) (reg t2))          % check arg2 for 16 bits
       (cmpa!.l (reg 2) (reg t2))
       (bne!.s longdiv)
       (divs (reg 2) (reg 1))              % do the arithmetic
       (bvs!.s longdiv)                    % do it the hard way if it failed
       (ext!.l (reg 1))                    % sign extend answer to 32 bits
       (*pop (reg t2))
       (*exit 0)
  
LONGDIV                                    % later we'll do a better job
       (movem!.l (reglist (reg d2) (reg d3) (reg d4) (reg d5) (reg d6)
			  (reg d7) (reg a0) (reg a1) (reg a2) (reg a3)
			  (reg a4) (reg a5))
                 (predecrement (reg st)))  % save all the registers
       (*link ldiv expr 2)
       (movem!.l (postincrement (reg st)) % restore all the registers
		 (reglist (reg d2) (reg d3) (reg d4) (reg d5) (reg d6)
			  (reg d7) (reg a0) (reg a1) (reg a2) (reg a3)
			  (reg a4) (reg a5)))
       (*pop (reg t2))
       (*exit 0)))
) % End of if_system.

