%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for 88k
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 5-Mar-90  (Winfried Neun)
%  added and refined (x-)code-address-to-symbol
% 10-Mar-89 (Winfried Neun)
%  Handling for 88k (SUN4). Complete redesign.
%  Also installed fluid *bruch* to allow gc (etc) to turn off
%  cntrl-c interrupt detection, which destroy system.
% 18-Aug-86 (Leigh Stoller)
%  Reworked so that signals can be added more easily. Also added calls to
%  external functions which set up the trap, and reinit the trap after a
%  signal is received.
% 2-Jan-85 10:21:47 (Vicki O'Day)
%  Instead of just calling stderror with a message about the kind of trap,
%  we now find the name of the routine that caused the trap and report
%  that as well.  CODE-ADDRESS-TO-SYMBOL is used to figure out the name.
% 27-Dec-84 10:55:56 (Vicki O'Day)
%  Added sigunwind, to pop the signal handler's stack frame before calling
%  stderror.  (Many thanks to Jim Davis for deciphering the contents of this
%  frame.)
% 17-Dec-84 10:24:41 (Vicki O'Day)
%  Added lots more signals to catch, to try to prevent those occasional
%  dumps out of Nmode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (flag '(build-trap-message) 'internalfunction))

(fluid '(*dump 88k-dumparray*))

% this is the interface for toploop

(de initializeinterrupts () 

	  % all signals are initiated expect for those in the following list
          % they are ignored

    (for (from i 1 31) (do
       (when (not (memq i '(1 18 19 20 21 22 23 26 28))) (sigsetup i)))))

% this installs the handler

(de sigsetup (signumber)
    (psl_sigset signumber (wgetv symfnc (id2int 'herculerpv)) 88k-dumparray*))

% this is the 'handler'

(setq 88k-dumparray* (gtwarray 60))

(de lisprpv () (wputv 88k-dumparray* 0 2) (herculerpv))
                      % entry from io routines
(de herculerpv()
   (prog (mess nu)
    (initializeinterrupts)
    (smalldump)
    (setq mess (atsoc (wgetv 88k-dumparray* 0) '((1 . "Hangup")
                          (-1 . "Interrupt") (2 . "Interrupt")
                          (3 . "Quit")
                          (4 . "Illegal Instruction")
                          (5 . "Trace Trap")
                          (6 . "IOT Instruction")
                          (7 . "EMT Instruction")
                          (8 . "Floating Point Exception")
                          (10 . "Bus Error")
			  (11 . "Segmentation Violation")
                          (12 . "Bad Args to System Call")
                          (13 . "Write on Pipe With None to Read")
                          (14 . "Alarm Clock")
                          (15 . "Software termination signal")  )))
    (setq mess (if mess (cdr mess)
                  (bldmsg " Unknown signal type %d" (wgetv 88k-dumparray* 0 ))))
    (setq mess (build-trap-message mess (wgetv 88k-dumparray* 16)))
    (setq nu (wgetv 88k-dumparray* 0))
    (when (eq nu -1) (setq nu 2))
    (wputv 88k-dumparray* 0 0)
    (return (error (wplus2 10000 nu) mess))))

(de smalldump()
  (when *dump
    (console-print-string "*** UNEXPECTED INTERRUPT ***")
    (console-newline)
    (console-print-string "*** System signal number: 16#")
    (unixputn (wgetv 88k-dumparray* 0))
    (console-print-string " , signal code: 16#")
    (unixputn (wgetv 88k-dumparray* 1))
    (console-newline)
    (console-print-string "*** signal pc: 16#")
    (unixputn (wgetv 88k-dumparray* 16))
    (console-print-string " , signal stack: 16#")
    (unixputn (wgetv 88k-dumparray* 3))
    (console-print-string " , context: 16#")
    (unixputn (wgetv 88k-dumparray* 2))
    (console-newline)
    (console-print-string "*** Registers  0-7 : ")
    (console-print-string " 0  ")
    (unixputn (wgetv 88k-dumparray* 19)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 20)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 21)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 22)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 23)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 24)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 25)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** Registers  8-15: ")
    (unixputn (wgetv 88k-dumparray* 26)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 27)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 28)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 29)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 30)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 31)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 32)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 33)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** Registers 16-23: ")
    (unixputn (wgetv 88k-dumparray* 34)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 35)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 36)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 37)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 38)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 39)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 40)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 41)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** Registers 24-31: ")
    (unixputn (wgetv 88k-dumparray* 42)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 43)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 44)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 45)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 46)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 47)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 48)) (console-print-string " ")
    (unixputn (wgetv 88k-dumparray* 49)) (console-print-string " ")
     (console-newline)
     (console-newline)
))

(compiletime (flag '(x-code-address-to-symbol) 'internalfunction))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (setf extra-info (bldmsg "%w%w" " in "
                                 (x-code-address-to-symbol (inf trap-addr))))
      (bldmsg "%w%w" trap-type extra-info)))

(fluid '(code-address* closest-address* closest-symbol*))

(de x-code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0))
        (mapobl #'(lambda (symbol)
                 (when (fcodep symbol)
                       (let ((address (inf (getfcodepointer symbol))))
                            (when (and (ileq address code-address*)
                                       (igreaterp address closest-address*))
                                  (setq closest-address* address)
                                  (setq closest-symbol* symbol))))))
       closest-symbol*))

(de code-address-to-symbol (ad) (x-code-address-to-symbol ad))
