%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:EVAL-APPLY.SL
% Title:        Function calling mechanism
% Author:       Eric Benson
% Created:      20 August 1981
% Modified:     20-Oct-86
% Status:       Experimental
% Mode:         Lisp
%
% (c) Copyright 1987, Konrad Zuse - Zentrum Berlin, for modifications
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 25 Feb 88 W. Neun , ZIB Berlin
% Cray1 version of handcoded eval without DRI/ERI instruction
% 20 Oct 86 H. Melenk and W. Neun , ZIB Berlin
% handcoded eval in lap
% 10 Oct 86 H. Melenk and W. Neun , ZIB Berlin
% changed eval that fntype is no longer in property list, but in symfnc
% vector,to save many get operations
% 31-May-84 10:43:21 (Brian Beach)
%  Calls on IDAPPLY1 --> IDAPPLY.
% 29-Mar-84 13:50:04 (Mike Creech)
%   Added IDAPPLY for use when compiler does not optimize to opencode or when
%   interpreting.
% 01-Dec-83 14:42:43 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime
(progn (Load vector-instructions if-system)
(flag '(lambdaevalapply lambdaapply lambdaevalapply-spec
        evallambdacase evalconterror evalundefconterror
        fnidapply
        evalmacrocase evalnexprcase) 'internalfunction)
 
(fluid '(lambda-link-Address undefined-function-Address))
 
(ds opengetfntype (u)
  (get u 'type))
 
(ds openfunboundp (u)
  % does U not have a function defn?
  (if (idp u)
    (weq (symfnc (idinf u)) undefined-function-Address)
    (noniderror u 'funboundp)))
 
(ds evalfunboundp (u)
  % does U not have a function defn?
    (weq u undefined-function-Address))
 
(ds openfcodep (u)
  % is U a compiled function?
  (if (idp u)
    (and (wneq (symfnc (idinf u)) undefined-function-Address)
         (wneq (symfnc (idinf u)) lambda-link-Address))
    (noniderror u 'fcodep)))
 
(ds evalfcodep (u)
  % is U a compiled function?
    (and (wneq u undefined-function-Address)
         (wneq u lambda-link-Address)))
 
(ds opengetfcodepointer (u)
  % Get code pointer for U
  (if (idp u)
    (mkcode (symfnc (idinf u)))
    (noniderror u 'getfcodepointer)))
 
)) %end compiletime
 
(de lambdaevalapply (btrmark fn args)
  % Fn is Lambda, Args to be Evaled
  (if (not (eqcar fn 'lambda))
    (continuableerror '1103 '"Ill-formed function expression"
                      (cons fn args))
    (prog (btm n result)
          (setq btm btrmark)
          (setq btrmark nil) % backtrace should appear only one
 
          (setq n (bindeval (cadr fn) args))
          % hand-coded, bind formals to evlis args
          (when (weq n -1)
            (return (continuableerror '1203 '"Argument number mismatch"
                     (cons fn args))))
          (setq result (evprogn (cddr fn)))
          (when (wneq n 0)
            (unbindn n))
          (return result))))
 
(de lambdaapply (fn args)
  % Fn is Lambda, unevaled Args
  (if (not (eqcar fn 'lambda))
    (continuableerror '1104 '"Ill-formed function expression"
                      (cons fn (foreach x in args collect (mkquote x))))
    (prog (formals n result)
          (setq formals (cadr fn))
          (setq n 0)
          (while (and (pairp formals) (pairp args))
            (lbind1 (car formals) (car args))
            (setq formals (cdr formals))
            (setq args (cdr args))
            (setq n (wplus2 n 1)))
          (when (or (pairp formals) (pairp args))
            (return (continuableerror '1204 '"Argument number mismatch"
                     (cons fn (foreach x in args collect (mkquote x))))))
          (setq result (evprogn (cddr fn)))
          (when (wneq n 0)
            (unbindn n))
          (return result))))
 
% Apply differs from the Standard Lisp Report in that functions other
% than EXPRs are allowed to be applied, the effect being the same as
% Apply(cdr GetD Fn, Args)
 
(de apply (fn args)
  % Indirect function call
  (cond ((idp fn)
         (prog (stackmarkforbacktrace result)
               (when (openfunboundp fn)
                 (return (continuableerror 1002
                          (bldmsg "%r is an undefined function" fn)
                          (cons fn (foreach x in args collect (mkquote x))))))
               (setq stackmarkforbacktrace (mkbtr (inf fn)))
               (setq result (if (openfcodep fn)
                       (codeapply (opengetfcodepointer fn) args)
                       (lambdaapply (get fn '*lambdalink) args)))
               (return result)))
        ((codep fn) (codeapply fn args))
        ((eqcar fn 'lambda) (lambdaapply fn args))
        (t (continuableerror 1102 "Ill-formed function expression"
            (cons fn (foreach x in args collect (mkquote x)))))))
 
(de idapply (fn args)
 
  % This is equivalent to APPLY but 'fn' must be an id.
  % Sometimes calls to this function are optimized by the compiler
  % to open code (i.e., this function is not called).
 
  (prog (stackmarkforbacktrace result)
    (unless (idp fn)         % THEN We have an error.
           (continuableerror 1102 "Ill-formed function expression"
                       (cons fn (foreach x in args collect (mkquote x)))))
    (when (openfunboundp fn)
      (return (continuableerror 1002
                        (bldmsg "%r is an undefined function" fn)
                        (cons fn (foreach x in args collect (mkquote x))))))
    (setq stackmarkforbacktrace (mkbtr (inf fn)))
    (setq result (if (openfcodep fn)
               (codeapply (opengetfcodepointer fn) args)
               (lambdaapply (get fn '*lambdalink) args)))
    (return result)))
 
 
(commentoutcode
(de eval (u)
  % Interpret S-Expression as program
    (let (fn symf symbits stackmarkforbacktrace)
     (clear-extraRegs)
                       % make the extraRegs look nice
     (cond
      ((null u) nil)
      ((eq u 't) t)
      ((idp u)
           (let ((val (symval (idinf u))))
             (if (weq (tag val) unbound-tag)
                       (valuecell u)
                       val)))
      ((not (pairp u)) u)
      (t 
         (let ((fn (car u)) symf symbits stackmarkforbacktrace)
          (cond ((eq fn 'quote) (cadr u))
                ((idp fn) (setq symf (wgetv symfnc (idinf fn)))
                          (setq symbits (wand symf 7))
                          (setq symf (wrshift (wlshift symf 6) 40))
                          %symfnc access
 
                          (setq stackmarkforbacktrace
                                                (mkbtr (inf fn)))
                          (cond ((evalfunboundp symf)
                                       (evalconterror 1300
                                  "%r is an undefined function" u))
                                ((weq symbits 0) % must be an EXPR
                                   (if (evalfcodep symf)
                                          (codeevalapply
                                            (mkcode symf) (cdr u) fn)
                                          (evallambdacase fn u)))
                                ((weq symbits 1) % 'fexpr
                                 (fnidapplye (cdr u) fn ))
                                ((weq symbits 2) % 'nexpr
                                 (evalnexprcase fn u))
                                ((weq symbits 3) % 'macro
                                 (evalmacrocase fn u))
                                (t
                                 (evalconterror 1301
                                      "Unknown function type %r"
                                               u))
                       ) )
                ((codep fn) (codeevalapply fn (cdr u) fn))
                ((and (pairp fn) (equal (car fn) 'lambda))
                 (lambdaevalapply stackmarkforbacktrace fn (cdr u)))
                (t
                 (evalconterror 1302
                           "Ill-formed expression in Eval %r" u))))))))
 
)
(compiletime (off lapopt))
 
(if_system Cray1
(lap '((*ENTRY EVAL EXPR 1)
     (*alloc 0)
     (Move (Reg S6) (Reg T00))
     (Xor (Reg S0) (Reg S6) (Reg 1)) % null ?
     (Move (Reg S7) (Reg T21))       % T ?
     (JSZ G0001)
     (Xor (Reg S0) (Reg 1) (Reg S7))
     (Move (Reg S6) (Reg 1))       %idp case ?
     (Rshift (Reg S6) (Reg S6) 59)
     (JSZ G0001)
     (Move (Reg S7) 30)
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (Move (Reg T73) (Reg 1))
     (JSN G0006)                  % not an id
 
     (Move (Reg A6) (Reg SYMVAL))     % 4 move from below
     (Move (Reg A7) (Reg 1))
     (Add (Reg A6) (Reg A6) (Reg A7))
     (Load (Reg 3) (INDEXED (Reg A6) 0)) % accessing symval
     (Move (Reg S7) 29)  % unbound ?
 
     (Move (Reg 1) (Reg 3))
     (Move (Reg 2) (Reg 1))
     (Rshift (Reg 2) (Reg 2) 59)
     (Xor (Reg S0) (Reg 2) (Reg S7))
     (JSN G0001)
     (Move (Reg 1) (Reg T73))
     (*linke 0 VALUECELL expr 1)
 
 g0001
     (J (Reg b00))
 
 G0006                          % not pairp case
     (Move (Reg S6) (Reg 1))
     (Rshift (Reg S6) (Reg S6) 59)
     (Move (Reg S7) 9)
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (Move (Reg A1) (Reg 1))    % quote case moved
     (JSN G0001)
 
     (Load (Reg 1) (INDEXED (Reg A1) 0)) % quote case
     (Load (Reg 2) (INDEXED (Reg A1) 1))
     (Move (Reg A6) 10)    % clear extraregs
     (VLMoveA (Reg A0) (Reg A6))
     (Move (Reg A0) (Reg ARGUMENTBLOCK))
     (Move (Reg S7) -1)
     (VmmoveS (Reg A0) (Reg S7))   % clear extrarregs
     (Move (Reg S7) 0)
     (VmergeSV (Reg V1) (Reg S7) (Reg V0))
     (Vstore (Reg V1) (Reg A0))
     (Move (Reg S7) (XIDLOC QUOTE 15))
     (Move (Reg A6) 4)
     (DRshift (Reg S7) (Reg S7) (Reg A6))
     (Xor (Reg S0) (Reg 1) (Reg S7))
     (Move (Reg T75) (Reg 2))               %preLoaded (cdr u)
     (JSZ G4711)
 
     (Move (Reg S7) (Reg 1))        %idp case
     (Rshift (Reg S7) (Reg S7) 59)
     (Move (Reg S6) 30)
     (Xor (Reg S0) (Reg S7) (Reg S6))
     (Move (Reg A6) (Reg SYMFNC))
     (Move (Reg A7) (Reg 1))
     (Add  (Reg a2) (Reg a6) (Reg a7))
     (JSN G0014)
 
     (Load (Reg S6) (INDEXED (Reg A2) 0))
     (Move (Reg 3) 7)
     (Move (Reg T71) (Reg S6))     %put symf to t71
     (And (Reg 3) (Reg 3) (Reg S6))
     (Move (Reg T72) (Reg 3))      % put symbits to t72
     (Lshift (Reg S6) (Reg S6) 6)
     (Rshift (Reg S6) (Reg S6) 40)
     (Move (Reg T71) (Reg S6))      % symf
%    (Move (Reg 4) (Reg 1))
%    (Move (Reg A6) 59)
%    (DRshift (Reg S7) (Reg 4) (Reg A6))
%    (Move (Reg 4) 28)
%    (DLshift (Reg 4) (Reg S7) (Reg A6))  %mkbtr for future use
     (Move (Reg A6) (Reg UNDEFINED-FUNCTION-AddRESS))
     (Move (Reg S7) (Reg A6))
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (jsz g4712)
 
     (Move (Reg S0) (Reg T72)) %weq symbits 0
     (JSN G0017)
 
     (Move (Reg A6) (Reg LAMBDA-LINK-AddRESS))
     (Move (Reg S7) (Reg A6))
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (JSZ G0019)
 
     (Move (Reg T75) (Reg 2)) %cdr u
%    (Move (Reg A6) 59)
%    (DRshift (Reg S7) (Reg S6) (Reg A6))  %not used
%    (Move (Reg 1) 15)
%    (DLshift (Reg 1) (Reg S7) (Reg A6))
     (Move (reg 3) (reg 1))
     (Move (Reg 1) (Reg s6))
     (*linke 0 CODEEVALAPPLY expr 3)
 G0019
     (Move (Reg 2) (Reg T73))
     (J (InternalEntry EVALLAMBDACASE))
 G0017
     (Move (Reg S7) 1)
     (Xor (Reg S0) (Reg 3) (Reg S7)) %weq symbits 1?
     (JSN G0022)
 
     (Move (Reg 2) (Reg 1))
     (Move (Reg 1) (Reg t75))
%    (Move (Reg S6) 1)  % this is idapplye
%    (Move (Reg S7) (Reg 2))
%    (Move (Reg A7) (Reg S7))
%    (Lshift (Reg S7) (Reg S7) 4)
%    (Or (Reg S6) (Reg S6) (Reg S7))
%    (Move (Reg NARGReg) (Reg S6))
%    (Move (Reg A6) (Reg SYMFNC))
%    (Add (Reg A6) (Reg A6) (Reg A7))
%    (Add (Reg A6) (Reg A6) (Reg A6))
%    (Add (Reg A6) (Reg A6) (Reg A6))
%    (Move (Reg B07) (Reg A6))
%    (J (Reg B07))
     (J (InternalEntry fnidapply))
 
 G0022                              %weq symbits 2?
     (Move (Reg S7) 2)
     (Xor (Reg S0) (Reg 3) (Reg S7))
     (Move (Reg 2) (Reg T73))
     (JSZ (InternalEntry EVALNEXPRCASE))
 
     (Move (Reg S7) 3)
     (Xor (Reg S0) (Reg 3) (Reg S7))
     (JSZ (InternalEntry EVALMACROCASE))
 
     (Move (Reg 3) (Reg T73))
     (*Move (quote "Unknown function type %r") (Reg 2))
     (Move (Reg 1) 1301)
     (J (InternalEntry EVALCONTERROR))
 
 G0014                            %codep
     (Move (Reg S7) (Reg 1))
     (Rshift (Reg S7) (Reg S7) 59)
     (Move (Reg S6) 15)
     (Xor (Reg S0) (Reg S7) (Reg S6))
     (Move (Reg S6) 9)            %pairp
     (JSZ G4713)
 
     (Xor (Reg S0) (Reg S7) (Reg S6))
     (Move (Reg A1) (Reg 1))
     (Move (Reg S6) (XIDLOC LAMBDA 15))
     (JSN G0027)
 
     (Load (Reg S7) (INDEXED (Reg A1) 0))
     (Move (Reg 2) (Reg t75))
     (Move (Reg A6) 4)
     (DRshift (Reg S6) (Reg S6) (Reg A6))
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (Jsz (InternalEntry LAMBDAEVALAPPLY-spec))
 
 g0027
     (Move (Reg 3) (Reg T73))
     (*Move (quote "Ill-formed expression in Eval %r") (Reg 2))
     (Move (Reg 1) 1302)
     (J (InternalEntry EVALCONTERROR))
 g4711
     (Move (Reg A5) (Reg 2))
     (Load (Reg 1) (INDEXED (Reg A5) 0))
     (j (Reg b00))
 G4712
     (Move (Reg 3) (Reg T73))
     (*Move (quote "%r is an undefined function") (Reg 2))
     (Move (Reg 1) 1300)
     (J (InternalEntry EVALundefCONTERROR))
 G4713
     (Move (Reg 2) (Reg t75))
     (Move (reg 3) 0)
     (*linke 0 CODEEVALAPPLY expr 3)
))
(lap '((*ENTRY EVAL EXPR 1)        %  X-MP version
     (*alloc 0)
     (Move (Reg A6) (Reg SYMVAL))     % 4 move from below
     (Move (Reg A7) (Reg 1))
     (Add (Reg A6) (Reg A6) (Reg A7))
     (DRI)
     (Load (Reg 3) (INDEXED (Reg A6) 0)) % accessing symval
     (Move (Reg S6) (Reg T00))
     (Xor (Reg S0) (Reg S6) (Reg 1)) % null ?
     (Move (Reg T73) (Reg 1))
     (Move (Reg S7) (Reg T21))       % T ?
     (Move (Reg S6) (Reg T73))       %idp case ?
     (JSZ G0001)
     (Xor (Reg S0) (Reg 1) (Reg S7))
     (Rshift (Reg S6) (Reg S6) 59)
     (Move (Reg S7) 30)
     (JSZ G0001)
 
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (Move (Reg S7) 29)  % unbound ?
     (JSN G0006)                  % not an id
 
     (Move (Reg 1) (Reg 3)) % preloaded
     (ERI)
     (Move (Reg 2) (Reg 1))
     (Rshift (Reg 2) (Reg 2) 59)
     (Xor (Reg S0) (Reg 2) (Reg S7))
     (JSN G0001)
     (Move (Reg 1) (Reg T73))
     (*linke 0 VALUECELL expr 1)
 
 g0001
     (move (reg s6) (reg 3))    % synchronize load
     (ERI) %%%%%%%%%%%%%%%%%%%%%%
     (J (Reg b00))
 
 G0006                          % not pairp case
     (Move (Reg S6) (Reg 1))
     (Rshift (Reg S6) (Reg S6) 59)
     (Move (Reg S7) 9)
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (Move (Reg A1) (Reg 1))    % quote case moved
     (JSN G0001)
 
     (ERI)
     (Load (Reg 1) (INDEXED (Reg A1) 0)) % quote case
     (Load (Reg 2) (INDEXED (Reg A1) 1))
     (Move (Reg A6) 10)    % clear extraregs
     (VLMoveA (Reg A0) (Reg A6))
     (Move (Reg A0) (Reg ARGUMENTBLOCK))
     (Move (Reg S7) -1)
     (VmmoveS (Reg A0) (Reg S7))   % clear extrarregs
     (Move (Reg S7) 0)
     (VmergeSV (Reg V1) (Reg S7) (Reg V0))
     (Vstore (Reg V1) (Reg A0))
     (Move (Reg S7) (XIDLOC QUOTE 15))
     (Move (Reg A6) 4)
     (DRshift (Reg S7) (Reg S7) (Reg A6))
     (Xor (Reg S0) (Reg 1) (Reg S7))
     (Move (Reg T75) (Reg 2))               %preLoaded (cdr u)
     (JSZ G4711)
 
     (Move (Reg S7) (Reg 1))        %idp case
     (Rshift (Reg S7) (Reg S7) 59)
     (Move (Reg S6) 30)
     (Xor (Reg S0) (Reg S7) (Reg S6))
     (Move (Reg A6) (Reg SYMFNC))
     (Move (Reg A7) (Reg 1))
     (Add  (Reg a2) (Reg a6) (Reg a7))
     (JSN G0014)
 
     (Load (Reg S6) (INDEXED (Reg A2) 0))
     (Move (Reg 3) 7)
     (Move (Reg T71) (Reg S6))     %put symf to t71
     (And (Reg 3) (Reg 3) (Reg S6))
     (Move (Reg T72) (Reg 3))      % put symbits to t72
     (Lshift (Reg S6) (Reg S6) 6)
     (Rshift (Reg S6) (Reg S6) 40)
     (Move (Reg T71) (Reg S6))      % symf
%    (Move (Reg 4) (Reg 1))
%    (Move (Reg A6) 59)
%    (DRshift (Reg S7) (Reg 4) (Reg A6))
%    (Move (Reg 4) 28)
%    (DLshift (Reg 4) (Reg S7) (Reg A6))  %mkbtr for future use
     (Move (Reg A6) (Reg UNDEFINED-FUNCTION-AddRESS))
     (Move (Reg S7) (Reg A6))
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (jsz g4712)
 
     (Move (Reg S0) (Reg T72)) %weq symbits 0
     (JSN G0017)
 
     (Move (Reg A6) (Reg LAMBDA-LINK-AddRESS))
     (Move (Reg S7) (Reg A6))
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (JSZ G0019)
 
     (Move (Reg T75) (Reg 2)) %cdr u
%    (Move (Reg A6) 59)
%    (DRshift (Reg S7) (Reg S6) (Reg A6))  %not used
%    (Move (Reg 1) 15)
%    (DLshift (Reg 1) (Reg S7) (Reg A6))
     (Move (reg 3) (reg 1))
     (Move (Reg 1) (Reg s6))
     (*linke 0 CODEEVALAPPLY expr 3)
 G0019
     (Move (Reg 2) (Reg T73))
     (J (InternalEntry EVALLAMBDACASE))
 G0017
     (Move (Reg S7) 1)
     (Xor (Reg S0) (Reg 3) (Reg S7)) %weq symbits 1?
     (JSN G0022)
 
     (Move (Reg 2) (Reg 1))
     (Move (Reg 1) (Reg t75))
%    (Move (Reg S6) 1)  % this is idapplye
%    (Move (Reg S7) (Reg 2))
%    (Move (Reg A7) (Reg S7))
%    (Lshift (Reg S7) (Reg S7) 4)
%    (Or (Reg S6) (Reg S6) (Reg S7))
%    (Move (Reg NARGReg) (Reg S6))
%    (Move (Reg A6) (Reg SYMFNC))
%    (Add (Reg A6) (Reg A6) (Reg A7))
%    (Add (Reg A6) (Reg A6) (Reg A6))
%    (Add (Reg A6) (Reg A6) (Reg A6))
%    (Move (Reg B07) (Reg A6))
%    (J (Reg B07))
     (J (InternalEntry fnidapply))
 
 G0022                              %weq symbits 2?
     (Move (Reg S7) 2)
     (Xor (Reg S0) (Reg 3) (Reg S7))
     (Move (Reg 2) (Reg T73))
     (JSZ (InternalEntry EVALNEXPRCASE))
 
     (Move (Reg S7) 3)
     (Xor (Reg S0) (Reg 3) (Reg S7))
     (JSZ (InternalEntry EVALMACROCASE))
 
     (Move (Reg 3) (Reg T73))
     (*Move (quote "Unknown function type %r") (Reg 2))
     (Move (Reg 1) 1301)
     (J (InternalEntry EVALCONTERROR))
 
 G0014                            %codep
     (Move (Reg S7) (Reg 1))
     (Rshift (Reg S7) (Reg S7) 59)
     (Move (Reg S6) 15)
     (Xor (Reg S0) (Reg S7) (Reg S6))
     (Move (Reg S6) 9)            %pairp
     (JSZ G4713)
 
     (Xor (Reg S0) (Reg S7) (Reg S6))
     (Move (Reg A1) (Reg 1))
     (Move (Reg S6) (XIDLOC LAMBDA 15))
     (JSN G0027)
 
     (Load (Reg S7) (INDEXED (Reg A1) 0))
     (Move (Reg 2) (Reg t75))
     (Move (Reg A6) 4)
     (DRshift (Reg S6) (Reg S6) (Reg A6))
     (Xor (Reg S0) (Reg S6) (Reg S7))
     (Jsz (InternalEntry LAMBDAEVALAPPLY-spec))
 
 g0027
     (Move (Reg 3) (Reg T73))
     (*Move (quote "Ill-formed expression in Eval %r") (Reg 2))
     (Move (Reg 1) 1302)
     (J (InternalEntry EVALCONTERROR))
 g4711
     (Move (Reg A5) (Reg 2))
     (Load (Reg 1) (INDEXED (Reg A5) 0))
     (j (Reg b00))
 G4712
     (Move (Reg 3) (Reg T73))
     (*Move (quote "%r is an undefined function") (Reg 2))
     (Move (Reg 1) 1300)
     (J (InternalEntry EVALundefCONTERROR))
 G4713
     (Move (Reg 2) (Reg t75))
     (Move (reg 3) 0)
     (*linke 0 CODEEVALAPPLY expr 3)
)))
(compiletime (on lapopt))
% aux stuff
 
(de evalnexprcase(fn u)
  (prog (btm res)
   (setq btm (mkbtr (inf fn)))
   (setq res nil)
   (setq res (idapply fn (list (evlis (cdr u)))))
   (return res)))
 
(de evalmacrocase(fn u)
  (prog (btm res)
   (setq btm (mkbtr (inf fn)))
   (setq res nil)
   (setq res (eval (idapply fn (list u))))
   (return res)))
 
(de evallambdacase (fn u)
 (lambdaevalapply (mkbtr (inf fn)) (get fn '*lambdalink) (cdr u)))
 
(de lambdaevalapply-spec (fn u)
   (lambdaevalapply (mkbtr (inf fn)) fn  u))
 
(de evalconterror (nr string val)
   (continuableerror nr
                    (bldmsg string val) val))
 
(de evalundefconterror (nr string val)
   (continuableerror nr
                    (bldmsg string (car val)) val))
 
(lap '((*entry fnidapply expr 2)
  (*alloc 1)
  (Move (Reg A7) (Reg 2))
  (Move (Reg A1) (Reg SYMFNC))
  (Add (Reg a1) (Reg a1) (Reg A7))
  (Move (Reg S7) (Reg 2))
  (load (Reg 4) (indexed (reg a1) 0))
  (Lshift (Reg S7) (Reg S7) 4)
  (Move (Reg S6) 1)
  (Or (Reg S6) (Reg S6) (Reg S7))
  (Move (Reg NargReg) (Reg S6))
  (*move (Reg 2) (reg 3))
  (*mkitem (Reg 3) btr-tag)
  (*move (Reg 3) (frame 1))
  (ploc (reg a7) (label back))
  (*move (Reg a7) (reg b00))
  (*wshift (reg 4) -32)
  (*move (Reg 4) (Reg b07))
  (J (Reg B07))
back
  (*dealloc 1)
  (j (reg b00))
  (*link idapply expr 2)))  % to get true allocation
