%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:LOAD-PSL.SL
% Description:  Load in the files comprising BARE-PSL
% Author:       Brian Beach, Hewlett-Packard CRC
% Created:      11-May-84
% Modified:     20-Jun-84 13:29:01
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 10-Jan 89   (Winfried Neun)
%  Adapted for ZIB's SParc (SUN4) PSL
% 05-Feb-1987 (Leigh Stoller & Harold Carr)
%  Removed load of vax-environ from pxnk-files list.
% 11-Dec-1986 (Tim Mueller)
%  Removed call to loader-main at the end of the file since loader-main
%  is called in main-start after loading the kernel.
% Tue Oct  8 05:56:55 1985 (Russ Fish)
%  Removed hardwired path to $pnkl from filenames, since the make-bare-psl
%  script cd's to $pnkl before starting bpsl.
% 16-Dec-84 08:30 (Brian Beach)
%  Added VAX-ENVIRONMENT. Removed CHAR-MACRO.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(on fast-integers)

(fluid '(bootstrap-files pnk-files pxnk-files error-files fluid-reg-vector
         stackupperbound))

(de init-viv ()
 (setq stackupperbound 0 stacklowerbound 0)
 nil)                                     % otherwise not restartable
(de preset-viv () nil ) % same

(de loader-main ()
  (suck-in-files bootstrap-files)
  (suck-in-files pnk-files)
  (suck-in-files pxnk-files)
  (suck-in-files error-files)
  (suck-in "putd-getd.b")
  (console-newline)
  (setf $eof$ '!)             % Must find the correct place to do this.

  (setq bootstrap-files nil
        pnk-files nil
        pxnk-files nil
        error-files nil)
  (pre-main)
  )

(de suck-in-files (file-list)
  (for (in f file-list)
       (do
    (suck-in f)
    ))
  )

(de suck-in (file-name)
  (console-print-string "loading ")
  (console-print-string file-name)
  (console-print-string " to ")
  (unixputn nextbps)
  (console-newline)
  (faslin file-name)
  )

(de trace-message (string)
  (console-print-string "TRACE: ")
  (console-print-string string)
  (console-newline)
  )

(setf bootstrap-files
  '(
    "constant.b"
    "compsupp.b"
    "consmkve.b"
    "proplist.b"
    "symbvals.b"
    "typeconv.b"
    "easysl.b"
    "functpri.b"
    "fluidglo.b"
    ))

(setf pnk-files
  '(
    "allocato.b"
    "applylap.b"
    "arithmet.b"
    "autoload.b"
    "autotr.b"
    "backtrac.b"
    "baseio.b"
    "binding.b"
    "break.b"
    "carcdr.b"
    "catchthr.b"
    "chario.b"
%   "compact-gc.b"
    "consmkve.b"
    "conterr.b"
    "copiers.b"
    "copying.b"
    "defconst.b"
    "defsmac.b"
    "dskin.b"
    "easynons.b"
    "equal.b"
    "evalappl.b"
    "evalwhen.b"
    "explcom.b"
    "ioextens.b"
    "knowntoc.b"
    "lispmacr.b"
    "load.b"
    "loopmacr.b"
    "oblist.b"
    "onoff.b"
    "openclos.b"
    "otherio.b"
    "otherssl.b"
    "printers.b"
    "printf.b"
    "progfri.b"
    "rdswrs.b"
    "read.b"
    "sequence.b"
    "sets.b"
    "stringge.b"
    "tokensca.b"
    "toploop.b"
    "vectors.b"

    ))

(setf pxnk-files
  '(
    "dumplisp.b"
    "iodata.b"
    "lap.b"
    "scantabl.b"
    "sysdate.b"
    "sysextra.b"
    "sysfasli.b"
    "sysgc.b"
    "sysio.b"
    "timc.b"
    "trap.b"
    "wfloat.b"
    "writefl.b"
    ))


(setf error-files
  '(
    "errorset.b"
    "errorhan.b"
    "ioerrors.b"
    "typerror.b"
    ))

(off fast-integers)
