%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:LAP.SL
% Description:  Miscellaneous LAP routines for the D-register model
% Author:       Sam Sands, Tim Tillson
% Created:      12-Dec-83
% Modified:     23-Apr-84 10:21:43
% Mode:         Lisp
% Package:      Kernel
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 01-Sep-88 (Julian Padget)
%  Added global declarations for bnd...
% 02-Aug-86 (Leigh Stoller)
%  Changed all references of fake reg 14 to real reg t2.
% 18-Jul-86 (Robert Kessler and Leigh Stoller)
%  Edited fastapply to use t regs instead of reg 8 and 7.
% 08-Jul-86 (Leigh Stoller)
%  Added an instruction to codeapply the tag strips the RTS pointer before
%   it is puished onto the stack.
% 29-Mar-84 12:50:31 (Jim Ambras/CRC)
%  Killed lap code for wtimes2, wquotient, wremainder.
% 16-Jan-84 14:32:43 (Nancy Kendzierski)
%  (Re)moved fast-binder stuff to php:fast-binder.sl.
% 12-Dec-83 15:23:25 (Tim Tillson)
%  Added header
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(global '(bndstkptr bndstkupperbound))

(lap '(              % This file contains miscellaneous LAP routines    scs


% syslsp procedure CodeApply(CodePtr, ArgList);
%
% if this could be written in Syslisp, it would look something like this:
% The arguments, up to 14, are loaded into the registers and execution jumps
% to the code pointer.
%
% begin scalar N;
%     N := 0;
%     while PairP ArgList do
%     <<  N := N + 1;
%	  ArgumentRegister[N] := car ArgList;
%	  ArgList := cdr ArgList >>;
%     (jump to address of code pointer)
% end;

% Note that this code assumes that (reg t2) is an A register            scs
%    and that an (EXIT 0) will do an RTS
%    and that the tagged code pointer is a valid address

  (!*entry CodeApply expr 2)           %  CodeApply(CodePointer, ArgList)
     %  code pointer, ready for RTS, but first it is tag stripped, then pushed
  (*field (reg 1) (reg 1) (wconst InfStartingBit) (wconst InfBitLength))
  (!*push (reg 1))
  (!*jumpeq (Label L0) (reg  2)  (quote nil))    %  done if no arguments

  (!*move (reg 2) (reg t2))            %  move argument list to safe place

  (!*move (car (reg t2)) (reg  1))     %  get actual argument from list
  (!*move (cdr (reg t2)) (reg t2))     %  advance pointer
  (!*jumpeq (Label L0) (reg t2)  (quote nil))    %  test for end of list  

  (!*move (car (reg t2)) (reg  2))     %  repeat for each argument
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg  3))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg  4))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg  5))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg  6))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg  7))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg  8))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg  9))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg 10))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg 11))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg 12))  
  (!*move (cdr (reg t2)) (reg t2))  
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*move (car (reg t2)) (reg 13))  
  (!*move (cdr (reg t2)) (reg t2))
  (!*jumpeq (Label L0) (reg t2)  (quote nil)) 

  (!*jumpnoteq (Label TooMany) (cdr (reg t2)) (quote nil))  
                                       %  check for more than 14

  (!*move (car (reg t2)) (reg 14))     % the 14th and last argument 

L0
  (!*exit 0)                           % JUMP to code pointer by RTS    scs

TooMany
	(!*MOVE (QUOTE "Too many arguments to function") (reg 1))
	(!*JCALL StdError)


% procedure IDApply0
% procedure IDApply1
% procedure IDApply2
% procedure IDApply3
% procedure IDApply4
%
%  The following assume that (reg t2) is an A register
%    and that the tag bit portions of addresses are don't cares

  (!*entry XIDApply0 expr 1)              %  IDApply0(CodePointer)
  (*field (reg 1) (reg 1)
	  (wconst InfStartingBit) (wconst InfBitLength))
  (!*move   (reg  1) (reg t2))           %  copy ID
  (!*wplus2 (reg t2) (reg t2))           %  x 2
  (!*wplus2 (reg t2) (reg  1))           %  x 3
  (!*wplus2 (reg t2) (reg t2))           %  x 6
  (!*wplus2 (reg t2) (fluid SymFnc))     % index into function cell
  (!*jump (memory (reg t2) 0))           % jump to function

  (!*entry XIDApply1 expr 2)              %  IDApply1(Arg1, CodePointer)
  (*field (reg 2) (reg 2) 
	  (wconst InfStartingBit) (wconst InfBitLength))
  (!*move   (reg 2) (reg t2))           %  copy ID
  (!*wplus2 (reg t2) (reg t2))           %  x 2
  (!*wplus2 (reg t2) (reg  2))           %  x 3
  (!*wplus2 (reg t2) (reg t2))           %  x 6
  (!*wplus2 (reg t2) (fluid SymFnc))     % index into function cell
  (!*jump (memory (reg t2) 0))           % jump to function

  (!*entry XIDApply2 expr 3)              %  IDApply2(Arg1, Arg2, CodePointer)
  (*field (reg 3) (reg 3) (wconst InfStartingBit) (wconst InfBitLength))
  (!*move   (reg  3) (reg t2))           %  copy ID
  (!*wplus2 (reg t2) (reg t2))           %  x 2
  (!*wplus2 (reg t2) (reg  3))           %  x 3
  (!*wplus2 (reg t2) (reg t2))           %  x 6
  (!*wplus2 (reg t2) (fluid SymFnc))     % index into function cell
  (!*jump (memory (reg t2) 0))           % jump to function

  (!*entry XIDApply3 expr 4)    %  IDApply3(Arg1, Arg2, Arg3, CodePointer)
  (*field (reg 4) (reg 4) (wconst InfStartingBit) (wconst InfBitLength))
  (!*move   (reg  4) (reg t2))        
  (!*wplus2 (reg t2) (reg t2))        
  (!*wplus2 (reg t2) (reg  4))        
  (!*wplus2 (reg t2) (reg t2))        
  (!*wplus2 (reg t2) (fluid SymFnc))     % index into function cell
  (!*jump (memory (reg t2) 0))           % jump to function

  (!*entry XIDApply4 expr 5)    %  IDApply4(Arg1, Arg2, Arg3, Arg4, CodePointer)
  (*field (reg 5) (reg 5)
	  (wconst InfStartingBit) (wconst InfBitLength))
  (!*move   (reg  5) (reg t2))   
  (!*wplus2 (reg t2) (reg t2))   
  (!*wplus2 (reg t2) (reg  5))   
  (!*wplus2 (reg t2) (reg t2))   
  (!*wplus2 (reg t2) (fluid SymFnc))      % index into function cell
  (!*jump (memory (reg t2) 0))            % jump to function
      ))


%
% TAG( FastApply )
%

(lap '((!*entry FastApply expr 0)	%. Apply with arguments loaded
%
% Called with arguments in the registers and functional form in reg t2 (14)
% Note: this routine is called from the compiler
%       

        (*MOVE (reg t2) (Fluid Codeform*))     %put lambda in codeform!*
        (!*FIELD (reg t1) (reg t2)
		 (wconst TagStartingBit) (wconst TagBitLength)) %tag in t1
	
        (!*FIELD (reg t2) (reg t2)
		 (wconst InfStartingBit) (wconst InfBitLength)) %inf in t2
	
        (!*JUMPNOTEQ (Label NotAnID) (reg t1) (wconst id-tag))
        
        (*move (fluid symfnc) (reg t1))
        (*wshift (reg t2) 2)
        (*wplus2 (reg t2) (reg t1))
        (*move (Fluid Codeform*) (reg t1))
        (jmp (indirect (reg t2)))
NotAnID
	(!*JUMPNOTEQ (Label NotACodePointer) (reg t1) (wconst code-tag))
        (*move (Fluid Codeform*) (reg t1))
        (jmp (reg t2))

NotACodePointer
	(*JUMPNOTEQ (Label IllegalFunctionalForm) (reg t1) (wconst pair-tag))
        (*MOVE (MEMORY (reg t2) 0) (reg t2))  % CAR with pair already untagged
        
	(*JUMPNOTEQ (Label IllegalFunctionalForm) (reg t2) (QUOTE LAMBDA))

	(*JCALL FastLambdaApply)
IllegalFunctionalForm
	(*MOVE (QUOTE "Illegal functional form in Apply") (reg 1))
	(*MOVE (reg t2) (reg 2))
	(*CALL List5)
	(*JCALL StdError)
)))

%
% TAG( UndefinedFunctionaux )
%

(lap '((!*entry UndefinedFunction-aux expr 0)	%. Error Handler for non code
       %
       % also called by JMP in the Function Cell
       %
       (*move (reg t1) (reg 2))
       (*MKITEM (reg 2) (wconst id-tag))
       (*MOVE (QUOTE "Undefined function %r called from compiled code")
	       (reg 1))
       (*CALL BldMsg)
       (*JCALL StdError)
       ))

