%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXU:disassemble.sl
% Title:          Disassemble of instructions for Mips
% Author:         Winfried Neun
% Created:        October 2, 1989
% Status:         Experimental
%
% Copyright (c) 1989, Konrad-Zuse-Zentrum Berlin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The disassemble takes its information from the compiler patterns
% mips-instrs.sl and inverts them in order to disassemble
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

(compiletime (flag '(dispatch-disassemble safe-cdr special-dispatch-disassemble
	cop1-dispatch-disassemble bcond-dispatch-disassemble
	normal-dispatch-disassemble print-instruction prinbl prinb prinbo
	prinbn prindig) 'internalfunction))

(fluid '(*instrsvect* *specialopcodevect* *bcondopcodevect* *regnames*
	*fregnames* *cop1opcodevect* klingelbeutel))

(setq *regnames*
   '((0 . "$0") (1 . "ff1") (2 . "reg T") ( 3 . "ff2")
     (4 . "reg 1") (5 . "reg 2") (6 . "reg 3") (7 . "reg 4") (8 . "reg 5")
     (9 . "reg t1") (10 . "reg t2") (11 . "reg t3") (12 . "reg t4") 
    (13 . "reg t5") (14 . "reg t6")
    (15 . "ff3") (16 . "reg heaplast") (17 . "reg heaptrapbound")
    (18 . "reg infmask") (19 . "$19") 
    (20 . "reg bndstkptr") (21 . "reg bndstklowerbound")
    (22 . "reg bndstkupperbound") (23 . "$23") (24 . "$24")
    (25 . "ff4") (26 . "$26") (27 . "$27") (28 . "$28") (29 . "reg st") 
    (30 . "reg NIL") (31 . "reg returnaddr")))

(setq *fregnames*
   '((0 . "f0") (1 . "f1") (2 . "f2") ( 3 . "f3") (4 . "f4") (5 . "f5")
     (6 . "f6") (7 . "f7") (8 . "f8") (9 . "f9") (10 . "f10") (11 . "f11")
     (12 . "f12") (13 . "f13") (14 . "f14") (15 . "f15") (16 . "f16") 
     (17 . "f17") (18 . "f18") (19 . "f19") (20 . "f20") (21 . "f21")
     (22 . "f22") (23 . "$23") (24 . "$24") (25 . "$25") (26 . "$26")
     (27 . "f27") (28 . "$28") (29 . "f29") (30 . "f30") (31 . "f31")))

(setq   *instrsvect* (mkvect 64)
	*specialopcodevect* (mkvect 64)
	*cop1opcodevect* (mkvect 64)
	*bcondopcodevect* (mkvect 32))


(compiletime
 (progn
  (de load-instruction-vectors (id)
     (cond ((get id '*opcode*)  (read-normal-instruction id))
	   ((get id '*special*) (read-special-instruction id))
	   ((get id '*bcond*)   (read-bcond-instruction id))
	   ((get id '*cop1*)    (read-cop1-instruction id))
	   (t nil)))  %etc

  (de read-special-instruction (id)
     (prog (opcode format)
       (setq opcode (car (get id '*special*)))
       (setq format (or (get id '*r-format*)
	                (get id '*i-format*)
	                (get id '*j-format*)))  
       (push `(putv *specialopcodevect* ,opcode ',(list id format))
	    klingelbeutel)))

  (de read-cop1-instruction (id)
     (prog (opcode format)
       (setq opcode (car (get id '*cop1*)))
       (setq format (or (get id '*fr-format*)
	                (get id '*fr2-format*)
	                (get id '*fi-format*)))  
       (push `(putv *cop1opcodevect* ,opcode ',(list id format))
	    klingelbeutel)))

  (de read-bcond-instruction (id)
     (prog (opcode format)
       (setq opcode (car (get id '*bcond*)))
       (setq format (or (get id '*r-format*)
	                (get id '*i-format*)
	                (get id '*j-format*)))  
       (push `(putv *bcondopcodevect* ,opcode ',(list id format))
	        klingelbeutel)))

  (de read-normal-instruction (id)
     (prog (opcode format)
       (setq opcode (car (get id '*opcode*)))
       (setq format (or (get id '*r-format*)
	                (get id '*i-format*)
	                (get id '*j-format*)))  
       (push `(putv *instrsvect* ,opcode ',(list id format))
	        klingelbeutel)))
))
% installing the vectors now:
    
(compiletime
 (dm generate-it()
 (setq klingelbeutel nil)
 (mapobl (function load-instruction-vectors))
 (cons 'progn klingelbeutel)))

(generate-it)

(de dispatch-disassemble (where word nextword)
   (prog (op instr)
     (prinbn where 8) (tab 15)
     (prinbn word  8) (tab 30)
     (setq op (wshift word -26))  % the opcode
     (setq instr (getv *instrsvect* op))
     (cond ((atom instr) (print (list " nanu :" instr)) (return nil))
	   ((eqcar instr 'special) (special-dispatch-disassemble word nextword))
	   ((eqcar instr 'bcond) (bcond-dispatch-disassemble word nextword))
	   ((eqcar instr 'cop1)  (cop1-dispatch-disassemble word nextword))
           ((get instr '*cop1*)  (cop1-dispatch-disassemble word nextword))
	   (t (normal-dispatch-disassemble word nextword))) 
))

% words will be split to sanitize the gc problem with unweaponed words
(de safe-cdr (x) (cond ((null x) x) (t (cdr x)))) 

(de special-dispatch-disassemble (word nextword)
   (prog (op instr target rs rt rd shamt funct base offset imm)
     (setq op (wand word 63))  % the opcode
     (setq instr (getv *specialopcodevect* op))
     (when (atom instr) (print  " ???? " ) (return nil))
     (setq target (wshift (wshift word 6) -4))
     (setq rs (cdr (assoc (wshift (wshift word 6)  -27) *regnames*)))
     (setq rt (cdr (assoc (wshift (wshift word 11) -27) *regnames*)))
     (setq rd (cdr (assoc (wshift (wshift word 16) -27) *regnames*)))
     (setq shamt (wshift (wshift word 21) -27))
     (setq funct (wand word 2#111111))
     (setq offset (ashift (wshift word 16) -16)) 
     (setq imm (wshift (wshift word 16) -16)) 
     (setq base rs)
     (print-instruction 
	       (pair '(instr target rs rt rd shamt funct base offset imm)
	         (list instr target rs rt rd shamt funct base offset imm))
	       (safe-cdr (getv *specialopcodevect* op))) %the format 
))

(de cop1-dispatch-disassemble (word nextword)
   (prog (op instr form fmt fmt2 rt fs ft fd funct base offset imm)
     (setq fmt (wshift (wshift word 6) -27))
     (setq fmt2 (wshift (wshift word 6) -22))
     (if (wgeq fmt 16)
	(progn 
	     (setq op (wand word 63))  % the opcode
	     (setq instr 
		 (bldmsg "%w.%w" (car (getv *cop1opcodevect* op)) "d")))
	(cond ((eq fmt 2#00100) (setq instr "MTC1") (setq form '((rt fs))))
	      ((eq fmt 2#00000) (setq instr "MFC1") (setq form '((rt fs))))
	      ((eq fmt 2#00010) (setq instr "CFC1") (setq form '((rt fs))))
              ((eq fmt 2#00110) (setq instr "CTC1") (setq form '((rt fs))))
	      ((eq fmt2 2#0100000000) (setq instr "BC1F") 
						(setq form '((offset))))
	      ((eq fmt2 2#0100000001) (setq instr "BC1T") 
						(setq form '((offset))))))
     (setq instr (list instr))
     (setq rt (cdr (assoc (wshift (wshift word 11) -27) *regnames*)))
     (setq fs (cdr (assoc (wshift (wshift word 16) -27) *fregnames*)))
     (setq ft (cdr (assoc (wshift (wshift word 11) -27) *fregnames*)))
     (setq fd (cdr (assoc (wshift (wshift word 21) -27) *fregnames*)))
     (setq funct (wand word 2#111111))
     (setq offset (ashift (wshift word 16) -16)) 
     (setq imm (wshift (wshift word 16) -16)) 
     (print-instruction 
	       (pair '(instr rt fd fs ft funct base offset imm)
	         (list instr rt fd fs ft funct base offset imm))
	    (or form
	       (list (cdadr (getv *cop1opcodevect* op))))) %the format
))

(de bcond-dispatch-disassemble (word nextword)
   (prog (op instr target rs rt rd shamt funct offset imm base)
     (setq op (wshift (wshift word 11) -27))  % the bcond-code
     (setq instr (getv *bcondopcodevect* op))
     (setq target (ashift (wshift word 6) -4))
     (setq rs (cdr (assoc (wshift (wshift word 6)  -27) *regnames*)))
     (setq rt (cdr (assoc (wshift (wshift word 11) -27) *regnames*)))
     (setq offset (ashift (wshift word 16) -16)) 
     (setq imm (wshift (wshift word 16) -16)) 
     (setq base rs)
     (print-instruction 
	       (pair '(instr target rs rt rd shamt funct base offset imm)
	         (list instr target rs rt rd shamt funct base offset imm))
	       (cdr (getv *bcondopcodevect* op))) %the format 
))

(de normal-dispatch-disassemble (word nextword)
   (prog (op instr target rs rt rd shamt funct base offset imm simm ft)
     (setq op (wshift word -26))  % the opcode
     (setq instr (getv *instrsvect* op))
     (setq target (ashift (wshift word 6) -4))
     (setq rs (cdr (assoc (wshift (wshift word 6)  -27) *regnames*)))
     (setq rt (cdr (assoc (wshift (wshift word 11) -27) *regnames*)))
     (setq rd (cdr (assoc (wshift (wshift word 16) -27) *regnames*)))
     (setq shamt (wshift (wshift word 21) -27))
     (setq funct (wand word 2#111111))
     (setq offset (ashift (wshift word 16) -16)) 
     (setq imm (wshift (wshift word 16) -16)) 
     (setq simm (ashift (wshift word 16) -16)) 
     (setq base rs)
     (setq ft (cdr (assoc (wshift (wshift word 11) -27) *fregnames*)))
     (print-instruction
	    (pair '(instr target rs rt rd shamt funct base offset imm simm ft)
	      (list instr target rs rt rd shamt funct base offset imm simm ft))
	       (cdr (getv *instrsvect* op))) %the format 
)) 

(de print-instruction (alist format)
   (prog (fmtitem institem comma comment)
     (tab 30)
     (prin2 (cadr (atsoc 'instr alist)))
     (tab 40)
     (setq format (car format))
    loop
     (when (null format) (when comment (tab 70) (prin2 comment))
                         (return (terpri)))
     (setq fmtitem (car format))
     (when (idp fmtitem)
	 (when comma (prin2 ","))
	 (setq comma t)
	 (setq institem (atsoc fmtitem alist))
	 (when (and (not (eq fmtitem 'based)) (null institem))
	         (print (list "komisch:" format)))
	 (if (eq fmtitem 'based)
	      (progn (prin2 (cdr (atsoc 'offset alist)))
	             (prin2 "(")
	             (prin2 (cdr (atsoc 'base alist)))
	             (prin2 ")")
                     (when (equal (cdr (atsoc 'base alist)) "reg st")
                           (setq comment (cdr (atsoc 'offset alist)))
                           (setq comment (wshift comment -2))
                           (setq comment (bldmsg "Frame (%w)" comment)))
                     (when (equal (cdr (atsoc 'base alist)) "$19")
                           (setq comment (cdr (atsoc 'offset alist)))
                           (if (wlessp comment -32000)
                                    (setq comment "Extrareg ")
                               (setq comment (wshift (wplus2 comment 32000) -2))                               (setq comment (symnam comment))))
                     (when (equal (cdr (atsoc 'base alist)) "$23")
                           (setq comment (cdr (atsoc 'offset alist)))
                           (setq comment (wshift (wplus2 comment 32000) -2))
                           (setq comment (symnam comment)))) 
	     (prin2 (if (pairp institem) (cdr institem)
	                                 institem))))
     (setq format (cdr format))
     (go loop)))

(de disassemble (x)
 (prog (stop i ll)
  (setq ll (linelength 120))
  (setq i x)
  (when (idp x) (when (funboundp x) (error 0 (list x "not compiled")))
	        (setq i (symfnc (idinf x))))  
  (prin2l (list "**** Function :" x " at " i)) (terpri) (terpri)
  (while (not stop)
    (dispatch-disassemble i (getmem i) (getmem (plus2 i 1)))
    (when (and (eq 0 (getmem i)) (eq 0 (getmem (wplus2 1 i)))) (setq stop t))
    (when (and (wgreaterp (getmem i) 0) (wlessp (getmem i) 16)) (setq stop t))
    (setq i (plus2 i 1)))
  (linelength ll)
))

(de prinbl (l)                  % binary (octal) printing of a list
     (mapc l (function prinb)))

(de prinb (it)                  % binary (octal) printing of an item
     (cond ((numberp it)(prinbo it))
	   ((eq it 't1) (tab 35))
	   ((eq it 't2) (tab 60))
	   (t       (prin2 it))))

(de prinbo (it)
      (cond ((wlessp it 0) (prin2 "-") (prinbo (wminus it)))
	    ((wgeq it 10)  (prin2 "x'") (prinbn it 1))
    (t            (prinbn it 1))))

(de prinbn (it n)                % print an octal number
     (cond ((and (weq it 0) (wleq n 0)) nil)
	    (t (progn
	          (prinbn (wshift it -4) (wplus2 n -1))
	          (prindig (wand it 15))
)    )     )  )

(de prindig (dig)        % print a numeric digit
     (if (wlessp dig 10) (writeChar  (wplus2 dig 48))
	   (writechar (wplus2 65 (wdifference dig 10)))))
 
