%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           Sparc-disassemble.SL
% Title:          Sparc PSL disassembler 
% Author:         Winfried Neun, ZIB Berlin
% Created:        19 November 1988
% Status:         Experimental
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
(fluid '(sun4-instructions sun4-float-instructions)) 

(setq sun4-instructions (mkvect 255))

% the following macros put the patterns in the appropriate vectors

(dm sun4-instr (u)
  (prog (op op2 opcode)
	% opcode is made of two parts (op and op2) and or 'ed together
	% building a 8 bit mask
    (setq op (cadr u))
    (setq op2 (caddr u))
    (when (eq op 01) (return NIL))

    (setq opcode (lor (lshift (cadr u) 6) (caddr u)))
    (return `(putv sun4-instructions ,opcode ',(cdddr u))))) 

(dm sun4-cond-instr (u) 
    (prog (op op2 opcode vect i)
	 (setq op (cadr u))
	 (setq op2 (caddr u))
	 (setq opcode (lor (lshift (cadr u) 6) (caddr u)))
	 (setq vect (mkvect 32))
	 (setq u (cdddr u))
	 (setq i 0)
	 (while u
	   (putv vect i (car u))
	   (setq u (cdr u))
	   (setq i (add1 i)))

	 (return 
	`(putv sun4-instructions ,opcode (cons 'sun4-cond-instr-handler ',vect))
)))

(dm sun4-float-instr (u) 
   (Prog (opf) 
     (setq opf (cadr u))
     (return 
     `(setq sun4-float-instructions (cons (cons ,opf ',(caddr u))
		                         sun4-float-instructions))
 )))


%load integer

(sun4-instr 2#11 2#001001 ("ldsb" t1 "[" rs1 "+" simm13 "]," rd t2 comment)
		          ("ldsb" t1 "[" rs1 "+" rs2 "]," rd t2 comment))
(sun4-instr 2#11 2#011001
		        ("ldsba" t1 "[" rs1 "+" rs2 "]" asi "," rd t2 comment))

(sun4-instr 2#11 2#001010 ("ldsh" t1 "[" rs1 "+" simm13 "]," rd t2 comment)
		          ("ldsh" t1 "[" rs1 "+" rs2 "]," rd t2 comment))
(sun4-instr 2#11 2#011010
		        ("ldsha" t1 "[" rs1 "+" rs2 "]" asi "," rd t2 comment))

(sun4-instr 2#11 2#000001 ("ldub" t1 "[" rs1 "+" simm13 "]," rd t2 comment)
		          ("ldub" t1 "[" rs1 "+" rs2 "]," rd t2 comment))
(sun4-instr 2#11 2#010001 
		        ("lduba" t1 "[" rs1 "+" rs2 "]" asi "," rd t2 comment))

(sun4-instr 2#11 2#000000 ("ld" t1 "[" rs1 "+" simm13 "]," rd t2 comment)
		          ("ld" t1 "[" rs1 "+" rs2 "]," rd t2 comment))
(sun4-instr 2#11 2#010000 ("lda" t1 "[" rs1 "+" rs2 "]" asi "," rd t2 comment))

(sun4-instr 2#11 2#000011 ("ldd" t1 "[" rs1 "+" simm13 "]," rd t2 comment)
		          ("ldd" t1 "[" rs1 "+" rs2 "]," rd t2 comment))
(sun4-instr 2#11 2#010011 
		        ("ldda" t1 "[" rs1 "+" rs2 "]" asi "," rd t2 comment))

% load floating point

(sun4-instr 2#11 2#100000 ("ld" t1 "[" rs1 "+" simm13 "]," f_rd t2 comment)
		          ("ld" t1 "[" rs1 "+" rs2 "]," f_rd t2 comment))

(sun4-instr 2#11 2#100011 ("ldd" t1 "[" rs1 "+" simm13 "]," f_rd t2 comment)
		          ("ldd" t1 "[" rs1 "+" rs2 "]," f_rd t2 comment))

(sun4-instr 2#11 2#100001 ("ld" t1 "[" rs1 "+" simm13 ",!%fsr")
		          ("ld" t1 "[" rs1 "+" rs2 ",!%fsr"))


% load coprocessor

(sun4-instr 2#11 2#110000 ("ld" t1 "[" rs1 "+" simm13 "]," c_rd t2 comment)
		          ("ld" t1 "[" rs1 "+" rs2 "]," c_rd t2 comment))

(sun4-instr 2#11 2#110011 ("ldd" t1 "[" rs1 "+" simm13 "]," c_rd t2 comment)
		          ("ldd" t1 "[" rs1 "+" rs2 "]," c_rd t2 comment))

(sun4-instr 2#11 2#110001 ("ld" t1 "[" rs1 "+" simm13 ",!%csr")
		          ("ld" t1 "[" rs1 "+" rs2 ",!%csr"))


%store integer

(sun4-instr 2#11 2#000101 ("stb" t1 rd ",[" rs1 "+" simm13 "]" t2 comment)
		          ("stb" t1 rd ",[" rs1 "+" rs2 "]" t2 comment))
(sun4-instr 2#11 2#010101 ("stba" t1 rd ",[" rs1 "+" rs2 "]" asi t2 comment))

(sun4-instr 2#11 2#000110 ("sth" t1 rd ",[" rs1 "+" simm13 "]" t2 comment)
		          ("sth" t1 rd ",[" rs1 "+" rs2 "]" t2 comment))
(sun4-instr 2#11 2#010110 ("stha" t1 rd ",[" rs1 "+" rs2 "]" asi t2 comment))

(sun4-instr 2#11 2#000100 ("st" t1 rd ",[" rs1 "+" simm13 "]" t2 comment)
		          ("st" t1 rd ",[" rs1 "+" rs2 "]" t2 comment))
(sun4-instr 2#11 2#010100 ("sta" t1 rd ",[" rs1 "+" rs2 "]" asi t2 comment))

(sun4-instr 2#11 2#000111 ("std" t1 rd ",[" rs1 "+" simm13 "]" t2 comment)
		          ("std" t1 rd ",[" rs1 "+" rs2 "]" t2 comment))
(sun4-instr 2#11 2#010111 ("stda" t1 rd ",[" rs1 "+" rs2 "]" asi t2 comment))


% store floating point

(sun4-instr 2#11 2#100100 ("st" t1 f_rd ",[" rs1 "+" simm13  "]" t2 comment)
		          ("st" t1 f_rd ",[" rs1 "+" rs2  "]" t2 comment))

(sun4-instr 2#11 2#100111 ("std" t1 f_rd ",[" rs1 "+" simm13  "]" t2 comment)
		          ("std" t1 f_rd ",[" rs1 "+" rs2  "]" t2 comment))

(sun4-instr 2#11 2#100101 ("st" t1 "!%fsr,[" rs1 "+" simm13 "]")
		          ("st" t1 "!%fsr,[" rs1 "+" rs2 "]"))

(sun4-instr 2#11 2#100110 ("std" t1 "!%fq,[" rs1 "+" simm13 "]")
		          ("std" t1 "!%fq,[" rs1 "+" rs2 "]"))


% store coprocessor

(sun4-instr 2#11 2#110100 ("st" t1 c_rd ",[" rs1 "+" simm13  "]" t2 comment)
		          ("st" t1 c_rd ",[" rs1 "+" rs2  "]" t2 comment))

(sun4-instr 2#11 2#110111 ("std" t1 c_rd ",[" rs1 "+" simm13  "]" t2 comment)
		          ("std" t1 c_rd ",[" rs1 "+" rs2  "]" t2 comment))

(sun4-instr 2#11 2#110101 ("st" t1 "!%csr,[" rs1 "+" simm13 "]")
		          ("st" t1 "!%csr,[" rs1 "+" rs2 "]"))

(sun4-instr 2#11 2#110110 ("std" t1 "!%cq,[" rs1 "+" simm13 "]")
		          ("std" t1 "!%cq,[" rs1 "+" rs2 "]"))


% atomic load-store unsigned byte

(sun4-instr 2#11 2#001101 ("ldstub" t1 "[" rs1 "+" simm13 "]," rd t2 comment)
		          ("ldstub" t1 "[" rs1 "+" rs2 "]," rd t2 comment))

(sun4-instr 2#11 2#011101
	 ("ldstuba" t1 "[" rs1 "+" rs2 asi "]," rd t2 comment))


% swap register with memory 

(sun4-instr 2#11 2#001111 ("swap" t1 "[" rs1 "+" simm13 "]," rd t2 comment)
		          ("swap" t1 "[" rs1 "+" rs2 "]," rd t2 comment))

(sun4-instr 2#11 2#011111 
		      ("swapa" t1 "[" rs1 "+" rs2 "]" asi "," rd t2 comment))


% add

(sun4-instr 2#10 2#000000 ("add" t1 rs1 "," simm13 "," rd t2 )
		          ("add" t1 rs1 "," rs2 "," rd t2 ))

(sun4-instr 2#10 2#010000 ("addcc" t1 rs1 "," simm13 "," rd t2 )
		          ("addcc" t1 rs1 "," rs2 "," rd t2 ))

(sun4-instr 2#10 2#001000 ("addx" t1 rs1 "," simm13 "," rd t2 )
		          ("addx" t1 rs1 "," rs2 "," rd t2 ))

(sun4-instr 2#10 2#011000 ("addxcc" t1 rs1 "," simm13 "," rd t2 )
		          ("addxcc" t1 rs1 "," rs2 "," rd t2 ))

% tagged add

(sun4-instr 2#10 2#100000 ("taddcc" t1 rs1 "," simm13 "," rd t2 )
		          ("taddcc" t1 rs1 "," rs2 "," rd t2 ))

(sun4-instr 2#10 2#100010 ("taddcctv" t1 rs1 "," simm13 "," rd t2 )
		          ("taddcctv" t1 rs1 "," rs2 "," rd t2 ))

% sub

(sun4-instr 2#10 2#000100 ("sub" t1 rs1 "," simm13 "," rd t2 )
		          ("sub" t1 rs1 "," rs2 "," rd t2 ))

(sun4-instr 2#10 2#010100 ("subcc" t1 rs1 "," simm13 "," rd t2 )
		          ("subcc" t1 rs1 "," rs2 "," rd t2 ))

(sun4-instr 2#10 2#001100 ("subx" t1 rs1 "," simm13 "," rd t2 )
		          ("subx" t1 rs1 "," rs2 "," rd t2 ))

(sun4-instr 2#10 2#011100 ("subxcc" t1 rs1 "," simm13 "," rd t2 )
		          ("subxcc" t1 rs1 "," rs2 "," rd t2 ))

% tagged sub

(sun4-instr 2#10 2#100001 ("tsubcc" t1 rs1 "," simm13 "," rd t2 )
		          ("tsubcc" t1 rs1 "," rs2 "," rd t2 ))

(sun4-instr 2#10 2#100011 ("tsubcctv" t1 rs1 "," simm13 "," rd t2 )
		          ("tsubcctv" t1 rs1 "," rs2 "," rd t2 ))

% multiply step

(sun4-instr 2#10 2#100100 ("mulscc" t1 rs1 "," simm13 "," rd t2 comment)
		          ("mulscc" t1 rs1 "," rs2 "," rd t2 comment ))

% multiply and divide

(sun4-instr 2#10 2#001010 ("umul" t1 rs1 "," simm13 "," rd t2 comment)
                          ("umul" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#001011 ("smul" t1 rs1 "," simm13 "," rd t2 comment)
                          ("smul" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#011010 ("umulcc" t1 rs1 "," simm13 "," rd t2 comment)
                          ("umulcc" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#011011 ("smulcc" t1 rs1 "," simm13 "," rd t2 comment)
                          ("smulcc" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#001110 ("udiv" t1 rs1 "," simm13 "," rd t2 comment)
                          ("udiv" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#001111 ("sdiv" t1 rs1 "," simm13 "," rd t2 comment)
                          ("sdiv" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#011110 ("udivcc" t1 rs1 "," simm13 "," rd t2 comment)
                          ("udivcc" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#011111 ("sdivcc" t1 rs1 "," simm13 "," rd t2 comment)
                          ("sdivcc" t1 rs1 "," rs2 "," rd t2 comment ))

% logical

(sun4-instr 2#10 2#000001 ("and" t1 rs1 "," simm13 "," rd t2 comment)
		          ("and" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#010001 ("andcc" t1 rs1 "," simm13 "," rd t2 comment)
		          ("andcc" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#000101 ("andn" t1 rs1 "," simm13 "," rd t2 comment)
		          ("andn" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#010101 ("andncc" t1 rs1 "," simm13 "," rd t2 comment)
		          ("andncc" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#000010 ("or" t1 rs1 "," simm13 "," rd t2 comment)
		          ("or" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#010010 ("orcc" t1 rs1 "," simm13 "," rd t2 comment)
		          ("orcc" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#000110 ("orn" t1 rs1 "," simm13 "," rd t2 comment)
		          ("orn" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#010110 ("orncc" t1 rs1 "," simm13 "," rd t2 comment)
		          ("orncc" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#000011 ("xor" t1 rs1 "," simm13 "," rd t2 comment)
		          ("xor" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#010011 ("xorcc" t1 rs1 "," simm13 "," rd t2 comment)
		          ("xorcc" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#000111 ("xnor" t1 rs1 "," simm13 "," rd t2 comment)
		          ("xnor" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#010111 ("xnorcc" t1 rs1 "," simm13 "," rd t2 comment)
		          ("xnorcc" t1 rs1 "," rs2 "," rd t2 comment ))

% shift

(sun4-instr 2#10 2#100101 ("sll" t1 rs1 "," simm13 "," rd t2 comment)
		          ("sll" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#100110 ("srl" t1 rs1 "," simm13 "," rd t2 comment)
		          ("srl" t1 rs1 "," rs2 "," rd t2 comment ))

(sun4-instr 2#10 2#100111 ("sra" t1 rs1 "," simm13 "," rd t2 comment)
		          ("sra" t1 rs1 "," rs2 "," rd t2 comment ))

% sethi

(sun4-instr 2#00 2#100 ("sethi" t1 imm22 "," rd t2 comment))


% save and restore

(sun4-instr 2#10 2#111100 ("save" t1 rs1 "," simm13 "," rd t2 comment)
		          ("save" t1 rs1 "," rs2 ","    rd t2 comment))

(sun4-instr 2#10 2#111101 ("restore" t1 rs1 "," simm13 "," rd t2 comment)
		          ("restore" t1 rs1 "," rs2 "," rd t2 comment ))


% branch on integer condition

(sun4-cond-instr 2#00 2#010
		 ("bn"   t1 disp22 t2 comment)
		 ("be"   t1 disp22 t2 comment)
		 ("ble"  t1 disp22 t2 comment)
		 ("bl"   t1 disp22 t2 comment)
		 ("bleu" t1 disp22 t2 comment)
		 ("bcs"  t1 disp22 t2 comment)
		 ("bneg" t1 disp22 t2 comment)
		 ("bvs"  t1 disp22 t2 comment)
		 ("ba"   t1 disp22 t2 comment)
		 ("bne"  t1 disp22 t2 comment)
		 ("bg"   t1 disp22 t2 comment)
		 ("bge"  t1 disp22 t2 comment)
		 ("bgu"  t1 disp22 t2 comment)
		 ("bcc"  t1 disp22 t2 comment)
		 ("bpos" t1 disp22 t2 comment)
		 ("bvc"  t1 disp22 t2 comment)

		 ("bn,a"   t1 disp22 t2 comment)
		 ("be,a"   t1 disp22 t2 comment)
		 ("ble,a"  t1 disp22 t2 comment)
		 ("bl,a"   t1 disp22 t2 comment)
		 ("bleu,a" t1 disp22 t2 comment)
		 ("bcs,a"  t1 disp22 t2 comment)
		 ("bneg,a" t1 disp22 t2 comment)
		 ("bvs,a"  t1 disp22 t2 comment)
		 ("ba,a"   t1 disp22 t2 comment)
		 ("bne,a"  t1 disp22 t2 comment)
		 ("bg,a"   t1 disp22 t2 comment)
		 ("bge,a"  t1 disp22 t2 comment)
		 ("bgu,a"  t1 disp22 t2 comment)
		 ("bcc,a"  t1 disp22 t2 comment)
		 ("bpos,a" t1 disp22 t2 comment)
		 ("bvc,a"  t1 disp22 t2 comment))

% branch on floating point condition

(sun4-cond-instr 2#00 2#110
		 ("fbn"   t1 disp22 t2 comment)
		 ("fbne"  t1 disp22 t2 comment)
		 ("fblg"  t1 disp22 t2 comment)
		 ("fbul"  t1 disp22 t2 comment)
		 ("fbl"   t1 disp22 t2 comment)
		 ("fbug"  t1 disp22 t2 comment)
		 ("fbg"   t1 disp22 t2 comment)
		 ("fbu"   t1 disp22 t2 comment)
		 ("fba"   t1 disp22 t2 comment)
		 ("fbe"   t1 disp22 t2 comment)
		 ("fbue"  t1 disp22 t2 comment)
		 ("fbge"  t1 disp22 t2 comment)
		 ("fbuge" t1 disp22 t2 comment)
		 ("fble"  t1 disp22 t2 comment)
		 ("fbule" t1 disp22 t2 comment)
		 ("fbo"   t1 disp22 t2 comment)

		 ("fbn,a"   t1 disp22 t2 comment)
		 ("fbne,a"  t1 disp22 t2 comment)
		 ("fblg,a"  t1 disp22 t2 comment)
		 ("fbul,a"  t1 disp22 t2 comment)
		 ("fbl,a"   t1 disp22 t2 comment)
		 ("fbug,a"  t1 disp22 t2 comment)
		 ("fbg,a"   t1 disp22 t2 comment)
		 ("fbu,a"   t1 disp22 t2 comment)
		 ("fba,a"   t1 disp22 t2 comment)
		 ("fbe,a"   t1 disp22 t2 comment)
		 ("fbue,a"  t1 disp22 t2 comment)
		 ("fbge,a"  t1 disp22 t2 comment)
		 ("fbuge,a" t1 disp22 t2 comment)
		 ("fble,a"  t1 disp22 t2 comment)
		 ("fbule,a" t1 disp22 t2 comment)
		 ("fbo,a"   t1 disp22 t2 comment))


% coprocessor branch on condition

(sun4-cond-instr 2#00 2#111
		 ("cbn"   t1 disp22 t2 comment)
		 ("cb123" t1 disp22 t2 comment)
		 ("cb12"  t1 disp22 t2 comment)
		 ("cb13"  t1 disp22 t2 comment)
		 ("cb1"   t1 disp22 t2 comment)
		 ("cb23"  t1 disp22 t2 comment)
		 ("cb2"   t1 disp22 t2 comment)
		 ("cb3"   t1 disp22 t2 comment)
		 ("cba"   t1 disp22 t2 comment)
		 ("cb0"   t1 disp22 t2 comment)
		 ("cb03"  t1 disp22 t2 comment)
		 ("cb02"  t1 disp22 t2 comment)
		 ("cb023" t1 disp22 t2 comment)
		 ("cb01"  t1 disp22 t2 comment)
		 ("cb013" t1 disp22 t2 comment)
		 ("cb012" t1 disp22 t2 comment)

		 ("cbn,a"   t1 disp22 t2 comment)
		 ("cb123,a" t1 disp22 t2 comment)
		 ("cb12,a"  t1 disp22 t2 comment)
		 ("cb13,a"  t1 disp22 t2 comment)
		 ("cb1,a"   t1 disp22 t2 comment)
		 ("cb23,a"  t1 disp22 t2 comment)
		 ("cb2,a"   t1 disp22 t2 comment)
		 ("cb3,a"   t1 disp22 t2 comment)
		 ("cba,a"   t1 disp22 t2 comment)
		 ("cb0,a"   t1 disp22 t2 comment)
		 ("cb03,a"  t1 disp22 t2 comment)
		 ("cb02,a"  t1 disp22 t2 comment)
		 ("cb023,a" t1 disp22 t2 comment)
		 ("cb01,a"  t1 disp22 t2 comment)
		 ("cb013,a" t1 disp22 t2 comment)
		 ("cb012,a" t1 disp22 t2 comment))

% call

(sun4-instr 2#01 ("call" t1 disp30 t2 comment))


% jump and link

(sun4-instr 2#10 2#111000 ("jmpl" t1 rs1 "+" simm13 "," rd t2 comment)
		          ("jmpl" t1 rs1 "+" rs2 "," rd t2 comment))

% return from trap

(sun4-instr 2#10 2#111001 ("rett" t1 rs1 "+" simm13 t2 comment)
		          ("rett" t1 rs1 "+" rs2    t2 comment))

% trap on integer condition

(sun4-instr 2#10 2#111010 ("trap"  t1 disp22 t2 comment))


% read state register

(sun4-instr 2#10 2#101000 ("rd" t1 "%y," rd t2 comment))

(sun4-instr 2#10 2#101001 ("rd" t1 "%psr," rd t2 comment))

(sun4-instr 2#10 2#101010 ("rd" t1 "%wim," rd t2 comment))

(sun4-instr 2#10 2#101011 ("rd" t1 "%tbr," rd t2 comment))


% write state register

(sun4-instr 2#10 2#110000 ("wr" t1 rs1 "," simm13 ",%y" t2 comment)
		          ("wr" t1 rs1 "," rs2 ",%y" t2 comment))

(sun4-instr 2#10 2#110001 ("wr" t1 rs1 "," simm13 ",%psr" t2 comment)
		          ("wr" t1 rs1 "," rs2 ",%psr" t2 comment))

(sun4-instr 2#10 2#110010 ("wr" t1 rs1 "," simm13 ",%wim" t2 comment)
		          ("wr" t1 rs1 "," rs2 ",%wim" t2 comment))

(sun4-instr 2#10 2#110011 ("wr" t1 rs1 "," simm13 ",%tbr" t2 comment)
		          ("wr" t1 rs1 "," rs2 ",%tbr" t2 comment))


% unimp

(sun4-instr 2#00 2#000  ("unimp" t1 ))


% instruction cache flush

(sun4-instr 2#10 2#111011 ("iflush" t1 rs1 "+" simm13 t2 comment)
		          ("iflush" t1 rs1 "+" rs2 t2 comment))



% floating point operate

(sun4-instr 2#10 2#110100 sun4-float-instructionhandler)

(sun4-instr 2#10 2#110101 sun4-float-instructionhandler) 

(sun4-float-instr 2#011000100  ("fitos" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011000100  ("fitos" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011001000  ("fitod" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011001100  ("fitox" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011000001  ("fstoir" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011000010  ("fdtoir" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011000011  ("fxtoir" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011010001  ("fstoi" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011010010  ("fdtoi" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011010011  ("fxtoi" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011001001  ("fstod" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011001101  ("fstox" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011000110  ("fdtos" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011001110  ("fdtos" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011000111  ("fxtos" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#011001011  ("fxtod" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#000000001  ("fmovs" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#000000101  ("fnegs" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#000001001  ("fabss" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#000101001  ("fsqrts" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#000101010  ("fsqrtd" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#000101011  ("fsqrtx" t1 f_rs2 "," f_rd))

(sun4-float-instr 2#001000001  ("fadds" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001000010  ("faddd" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001000011  ("faddx" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001000101  ("fsubs" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001000110  ("fsubd" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001000111  ("fsubx" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001001001  ("fmuls" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001001010  ("fmuld" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001001011  ("fmulx" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001001101  ("fdivs" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001001110  ("fdivd" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001001111  ("fdivx" t1 f_rs1 "," f_rs2 "," f_rd))

(sun4-float-instr 2#001010001  ("fcmps" t1 f_rs1 "," f_rs2 ))

(sun4-float-instr 2#001010010  ("fcmpd" t1 f_rs1 "," f_rs2 ))

(sun4-float-instr 2#001010011  ("fcmpx" t1 f_rs1 "," f_rs2 ))

(sun4-float-instr 2#001010101  ("fcmpes" t1 f_rs1 "," f_rs2 ))

(sun4-float-instr 2#001010110  ("fcmped" t1 f_rs1 "," f_rs2 ))

(sun4-float-instr 2#001010111  ("fcmpex" t1 f_rs1 "," f_rs2 ))


% coprocessor, NOT YET implemented

(sun4-instr 2#10 2#110110 ("coprocessor"))

(fluid '(labellist* *nice-frame-NAMES* *VIV-NAMES*))

(compiletime 
 (ds withinsymfnc (x)
    (and (wgeq x symfnc) 
	 (wleq x (wplus2 symfnc (times2 maxsymbols 
		                        addressingunitsperfunctioncell))))))

% main driver
%            either  (disassemble 'hugo)      or
%                    (disassemble '16#4711)

(de disassemble (fkt)
   (prog (name word op op2 i ll )  
      
     (setq name fkt  i 0  labellist* nil)
     
     (when (and (idp fkt) (wlessp (idinf fkt) maxsymbols))
		 (when (not(fcodep fkt)) (return nil))
		 (setq fkt (codeinf (getfcodepointer fkt))))

     (setq outputbase* 16)
     (prin2t (list " ** Function " name " at" fkt)) (terpri)
     (setq outputbase* 10)
  loop1                      % first pass picking labels only
     (setq word (wgetv fkt i))
     (setq op (wshift word -30))
     (when (and (eq 0 op) (eq 0 (wshift (wshift word 7) -29)))
		               (go loopex)) % finish at UNIMP
     (setq word NIL)        % for gc
		
     (cond ((eq op 1) (label-format-1 fkt i word)) % to fool compiler
	   ((eq op 0) (label-format-2 fkt i)))
    
     (setq i (wplus2 1 i))
     (go loop1)
   loopex
     (setq labellist* (labelsort labellist*))  % order labels
     (mapc labellist* (function (lambda (c) (rplacd c (stringgensym))))) 
     (setq i 0)
   loop 
     (setq word (wgetv fkt i))
     (when (setq ll
	      (assoc (wplus2 i (wquotient fkt 4)) labellist*))
	   (tab 20)
	   (prin2 (cdr ll))
	   (prin2t ":"))
     (setq op (wshift word -30))
     (when (and (eq 0 op) (eq 0 (wshift (wshift word 7) -29)))
		               (return nil)) % finish at UNIMP 
     (setq word NIL)        % for gc
		
     (cond ((eq op 1) (print-format-1 fkt i word)) % to fool compiler
	   ((eq op 0) (print-format-2 fkt i))
	   (t         (print-format-3 op fkt i)))
    
     (setq i (wplus2 1 i))
     (go loop)
)   ) 

(de print-format-1 (fkt i dummy)
   (prog (word word1)
     (setq word (wgetv fkt i))
     (prinbn (wplus2 fkt (wtimes2 4 i)) 8)
     (tab 12)
     (prinbn word 8)
     (tab 24)
     (setq word (ashift (wshift word 2) -2))  % for gc
     (prin2 "call")
     (tab 35)
     (setq word1 (wshift word 2)) % byte address
     (setq word1 (wplus2 (wplus2 fkt (wtimes2 i 4)) word1)) % absolute
     (if (withinsymfnc word1)
	 (setq word (int2id (wquotient (wdifference word1 symfnc) 8)))
	 (progn
	   (setq word (wquotient word1 4))
	   (setq word (assoc word labellist*))
	   (when word (setq word (cdr word)))))

     (prinb word)
     (terpri)) )          % some comments later

(de print-format-2 (fkt i)
   (prog (opcode pattern op2 rd a cond imm22 comment)
     (setq comment "")
     (setq imm22 (wgetv fkt i))
     (setq op2   (wshift (wshift imm22 7) -29))
     (setq rd    (regname (wshift (wshift imm22 2) -27)))
     (setq cond  (wshift (wshift imm22 3) -28))
     (setq a     (wshift (wshift imm22 2) -31))
     (prinbn (wplus2 fkt (wtimes2 4 i)) 8) 
     (tab 12) 
     (prinbn imm22 8) 
     (tab 24)
     (setq imm22 (ashift (wshift imm22 10) -10))
     (setq opcode  op2)
     (setq pattern (getv sun4-instructions opcode))
     (cond ((null pattern) (setq pattern "?????????"))
	   ((atom pattern) NIL)
	   ((and (pairp pattern) (idp (car pattern)))
		(return (apply (car pattern)
		   (list (cdr pattern) rd a cond imm22 i fkt))))
	   ((pairp pattern) (if (and (pairp (cdr pattern)) (eq 0 i))
		               (setq pattern (cadr pattern))
		               (setq pattern (car pattern))))
     )
     (when (pairp pattern)     % some special cases for sethi 
       (when (and (equal (car pattern) "sethi")
		  (eq imm22 0)   % NOP case
		  (eq rd '!r0))
		(return (prin2t "nop")))
       (if (and (equal (car pattern) "sethi") (eq rd '!r16))
	      % Alloc 'marker'
	     (setq comment (bldmsg "Alloc %d" (wand imm22 127)))
        (when (equal (car pattern) "sethi")          % SYMVAL access ?
	     (setq comment (wgetv fkt (wplus2 i 1)))
	     (setq comment (wor (wshift imm22 10) (wand comment 16#3ff)))
		     % the address
	     (if (and (wgreaterp comment SYMVAL)
		      (wlessp comment (wplus2 SYMVAL (wtimes2 4 maxsymbols))))
		(setq comment (symnam (wquotient
		                             (wdifference comment symval) 4)))
		(setq comment ""))))
       (prinbl
	(subla
	 (pair '(opcode op2 rd a cond imm22 comment)
		(list opcode op2 rd a cond imm22 comment))
	 pattern)))

     (terpri)))  % bis auf weiteres
 
(de label-format-1 (fkt i dummy)
   (prog (word word1)
     (setq word (wgetv fkt i))
     (setq word (ashift (wshift word 2) -2))  % for gc
     (setq word1 (wshift word 2)) % byte address
     (setq word1 (wplus2 (wplus2 fkt (wtimes2 i 4)) word1))
     (return
     (if (withinsymfnc word1)     nil
       (when (and (posintp word1) (wleq word1 nextbps))
	(push (cons (wquotient word1 4) "nowhere") labellist*))))))
 
(de label-format-2 (fkt i)
   (prog (imm22 word word1)
     (setq imm22 (wgetv fkt i))
     (setq imm22 (ashift (wshift imm22 10) -10))
     (setq word1 (wshift imm22 2)) % byte address
     (setq word1 (wplus2 (wplus2 fkt (wtimes2 i 4))
		         word1))
(return
     (if (withinsymfnc word1)     nil
      (if (eq 0 imm22) nil  % a NOP instr 
	  (push (cons (wquotient word1 4) "anywhere") labellist*))))))

(de sun4-cond-instr-handler (vect rd a cond imm22 i fkt)
   (prog (comment pattern)
     (setq comment "")
     (setq cond (wor (wshift a 4) cond))
     (setq pattern (getv vect cond))
     (cond ((null pattern) (setq pattern "?????????"))
	   ((atom pattern) NIL)
     )
     (setq imm22 (wtimes2 imm22 4))
     (setq vect (wplus2 imm22 (wplus2 (wtimes2 i 4) fkt)))
     (if (withinsymfnc vect)
	 (setq imm22
	     (int2id (wquotient (wdifference vect symfnc) 8)))
	 (progn
	   (setq vect (wquotient vect 4))
	   (setq vect (assoc vect labellist*))
	   (when vect (setq imm22 (cdr vect)))))

     (when (pairp pattern)
       (prinbl
	(subla
	 (pair '(rd a cond disp22 comment)
		(list rd a cond imm22 comment))
	 pattern)))
 
     (terpri))) 

(de print-format-3 (op fkt ii)
   (prog (pattern opcode i frd crd op3 rd asi rs1 rs2 opf simm13 comment)
     (setq comment "")
     (setq simm13 (wgetv fkt ii))
     (setq rd    (regname (wshift (wshift simm13 2) -27)))
     (setq frd   (assoc (wshift (wshift simm13 2) -27) *nice-frame-NAMES*))
       (when frd   (setq comment (cdr frd)))
     (setq frd   (fregname rd))
     (setq crd   (cregname rd))
     (setq op3   (wshift (wshift simm13 7) -26))
     (setq rs1   (regname (wshift (wshift simm13 13) -27)))
     (setq i     (assoc (wshift (wshift simm13 13) -27) *nice-frame-NAMES*))
       (when i   (if (equal comment "") (setq comment (cdr i))
		     (setq comment (bldmsg "%w/%w" comment (cdr i)))))
     (setq i     (wshift (wshift simm13 18) -31))
     (setq asi   (wshift (wshift simm13 19) -24))
     (setq rs2   (regname (wand simm13 16#1f)))
     (setq opf   (wshift (wshift simm13 18) -23))
     (prinbn (wplus2 fkt (wtimes2 4 ii)) 8)
     (tab 12)
     (prinbn simm13 8) 
     (setq simm13 (ashift (wshift simm13 19) -19))
     (tab 24)
     (setq opcode (wor (wshift op 6) op3))
     (setq pattern (getv sun4-instructions opcode))
     (cond ((null pattern) NIL)
	   ((idp (car pattern))
		 (return (apply (car pattern)
		            (list rd rs1 rs2 opf ))))
	   ((atom pattern) NIL)
	   ((pairp pattern) (if (and (pairp (cdr pattern)) (eq 0 i))
		               (setq pattern (cadr pattern))
		               (setq pattern (car pattern))))
     )
	   % special case : the VIV variables access
     (when (and (eq i 1) (eq rs1 '!r7) (assoc simm13 *viv-names*))
	   (setq comment (cdr (assoc simm13 *viv-names*))))
    
     (when (pairp pattern)
       (prinbl
	(subla
	 (pair '(opcode op3 rd i asi rs1 rs2 opf simm13 f_rd c_rd comment)
		(list opcode op3 rd i asi rs1 rs2 opf simm13 frd crd comment))
	 pattern))
       (terpri))

     ))  % bis auf weiteres

(de sun4-float-instructionhandler (rd rs1 rs2 opf)
   (prog (f_rd f_rs1 f_rs2 pattern)
      (setq f_rd (fregname rd))
      (setq f_rs1 (fregname rs1))
      (setq f_rs2 (fregname rs2)) 
      (setq pattern (assoc opf sun4-float-instructions))
      (when (not pattern) (return nil))
      (setq pattern (cdr pattern))
      (when (pairp pattern)
       (prinbl
	(subla
	 (pair '(f_rd f_rs1 f_rs2 ) (list f_rd f_rs1 f_rs2 ))
	 pattern))
       (terpri))
))

(fluid '(*nice-frame-NAMES*    *nice-F-REGISTER-NAMES* *viv-names*
	 *nice-REGISTER-NAMES* *nice-C-REGISTER-NAMES*))

(setq *nice-F-REGISTER-NAMES*
  '((!r0 !%!f0) (!r1 !%!f1) (!r2 !%!f2) (!r3 !%!f3) (!r4 !%!f4) (!r5 !%!f5)
    (!r6 !%!f6) (!r7 !%!f7) (!r8 !%!f8) (!r9 !%!f9) (!r10 !%!f10) (!r11 !%!f11)
    (!r12 !%!f12) (!r13 !%!f13) (!r14 !%!f14) (!r15 !%!f15) (!r16 !%!f16)
    (!r17 !%!f17) (!r18 !%!f18) (!r19 !%!f19) (!r20 !%!f20) (!r21 !%!f21)
    (!r22 !%!f22) (!r23 !%!f23) (!r24 !%!f24) (!r25 !%!f25) (!r26 !%!f26)
    (!r27 !%!f27) (!r28 !%!f28) (!r29 !%!f29) (!r30 !%!f30) (!r31 !%!f31)))
 
(setq *nice-REGISTER-NAMES*
   '((0 !r0) (1 !r1) (2 !r2) (3 !r3) (4 !r4) (5 !r5) (6 !r6) (7 !r7) (8 !r8)
     (9 !r9)  (10 !r10) (11 !r11) (12 !r12) (13 !r13) (14 !r14) (15 !r15)
    (16 !r16) (17 !r17) (18 !r18) (19 !r19) (20 !r20) (21 !r21) (22 !r22)
    (23 !r23) (24 !r24) (25 !r25) (26 !r26) (27 !r27) (28 !r28) (29 !r29)
    (30 !r30) (31 !r31)))
 
(setq *nice-C-REGISTER-NAMES*
 '((!r0 !%!c0) (!r1 !%!c1) (!r2 !%!c2) (!r3 !%!c3) (!r4 !%!c4) (!r5 !%!c5)
   (!r6 !%!c6) (!r7 !%!c7) (!r8 !%!c8) (!r9 !%!c9) (!r10 !%!c10) (!r11 !%!c11)
   (!r12 !%!c12) (!r13 !%!c13) (!r14 !%!c14) (!r15 !%!c15) (!r16 !%!c16)
   (!r17 !%!c17) (!r18 !%!c18) (!r19 !%!c19) (!r20 !%!c20) (!r21 !%!c21)
   (!r22 !%!c22) (!r23 !%!c23) (!r24 !%!c24) (!r25 !%!c25) (!r26 !%!c26)
   (!r27 !%!c27) (!r28 !%!c28) (!r29 !%!c29) (!r30 !%!c30) (!r31 !%!c31)))

(setq *nice-frame-NAMES*
 '((17 "frame 1") (18 "frame 2") (19 "frame 3") (20 "frame 4") (21 "frame 5")
   (22 "frame 6") (23 "frame 7") (24 "frame 8") (25 "frame 9") (26 "frame 10")
   (27 "frame 11") (28 "frame 12")))

(compiletime  % produce inverted list of viv-numbers from compiler
  (dm make-viv-names (u)
    (mapobl 
     (function
	 (lambda (x)
	     (cond ((get x 'viv-number)
		      (setq *viv-names*
		            (cons (cons (get x 'viv-number) x) *viv-names*)))
		   (t nil)))))
    (list 'setq '*viv-names* (list 'quote *viv-names*)))
)
(make-viv-names) % make it

(de regname (num)
  (prog (ass frm)
     (setq ass (assoc num *nice-REGISTER-NAMES*))
     (when (pairp ass)  (return (cadr ass)))))
 
(de fregname (num)
  (prog (ass)
     (setq ass (assoc num *nice-F-REGISTER-NAMES*))
     (when (pairp ass) (return (cadr ass)))))
 
 
(de cregname (num)
  (prog (ass)
     (setq ass (assoc num *nice-C-REGISTER-NAMES*))
     (when (pairp ass) (return (cadr ass)))))
 
 
(de prinbl (l)                  % binary (octal) printing of a list
     (mapc l (function prinb)))
 
(de prinb (it)                  % binary (octal) printing of an item
     (cond ((numberp it)(prinbo it))
	   ((eq it 't1) (tab 35))
	   ((eq it 't2) (tab 60))
	   (t       (prin2 it))))
 
(de prinbo (it)
      (cond ((wlessp it 0) (prin2 "-") (prinbo (wminus it)))
	    ((wgeq it 10)  (prin2 "x'") (prinbn it 1))
    (t            (prinbn it 1))))
 
(de prinbn (it n)                % print an octal number
     (cond ((and (weq it 0) (wleq n 0)) nil)
	    (t (progn
		  (prinbn (wshift it -4) (wplus2 n -1))
		  (prindig (wand it 15))
)    )     )  )
 
(de prindig (dig)        % print a numeric digit
     (if (wlessp dig 10) (writeChar  (wplus2 dig 48))
	   (writechar (wplus2 65 (wdifference dig 10)))))
 
(de labelsort (l)        % sort labels to ascending sequence
   (prog (so)
     (setq so (labelsort1 l nil))
     (return so)))
 
(de labelsort1 (rest sorted)
    (cond ((null rest) sorted)
	   (t (labelsort1(cdr rest) (labelsortin (car rest) sorted))) ))
 
(de labelsortin (object l)
    (cond ((null l) (list object))
	  ((wgreaterp (car object)(caar l))
		         (cons (car l) (labelsortin object (cdr l))) )
	  (t (cons object l)) ))
