%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% File:         PXU:MULS.SL
% Description:  Support for BIGNUM package with double INUM operations
%               Lap source for SPARC processor
% Author:       H. Melenk, W. Neun
% Created:      25 January 1989
% Modified:
% Mode:         Lisp
% Package:      Utilities
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Sun4 version by W. Neun, ZIB Berlin
%
% 14 Aug 1992 (Winfried Neun)
% Added SpuerSparc support and removed surplus wdivide from wquotientdouble
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the 68020 processor.
% These are the open codes which embed the extended arithmetic
% operations such that they are accessible from compiled LISP top level.
% They map the 2*30 bit operations needed by BIG to the 2*32 bit operations
% of the processor by shifting etc.

(fluid '(*second-value*))

(lap '((*entry wwwtimesdouble expr 2)
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The hig order bits are placed in fluid variable.
       (*alloc 0)
	(smul (reg 1) (reg 2) (reg 1))
        (rdy (reg 0) (reg 0) (reg 3))
      % now we have 32 low bits in REG 1, 32 high bits in REG 3
       (*WSHIFT (reg 3) 2)
       (*MOVE (reg 1) (reg 2))
       (*WSHIFT (reg 2) -30)
       (*WOR (reg 3)(reg 2))
       (*MOVE (reg 3)($FLUID *second-value*))
       (*WSHIFT (reg 1) 2)
       (*WSHIFT (reg 1) -2)
       (*exit 0)))


(lap '((*entry wwWxTimes2 expr 2)   % extended times
	(smul (reg 1) (reg 2) (reg 1))
	(rdy (reg 0) (reg 0) (reg 2))
	(*move (reg 2) ($fluid *second-value*))
	(*exit 0)))

(de supersparc() (copyd 'WxTimes2 'wwWxTimes2)
		 (copyd 'wquotientdouble 'wwwquotientdouble)
		 (copyd 'wtimesdouble 'wwwtimesdouble))
(commentoutcode
(put ' oldwquotientdouble 'opencode
      % called with a double length number in params 1 and 2
      % and a single length number in par 3.
      % Result is the single length quotient.
      % the remainder is placed in a fluid variable.
      '(%%(*alloc 3)  % adjusting the words first
        (*move 1 (reg 5))
        (*wshift (reg 5) 30)
        (*move (reg 1) (displacement (reg sp) -4))
        (*move (reg 2) (displacement (reg sp) -8))
        (*move (reg 5) (displacement (reg sp) -12))
        (*move (reg 3) (displacement (reg sp) -16))
        (ld (displacement (reg sp) -4) (reg f0))
        (ld (displacement (reg sp) -8) (reg f1))
        (ld (displacement (reg sp) -12) (reg f2))
        (fitox (reg f0) (reg f4))
        (fitox (reg f1) (reg f8))
        (fitox (reg f2) (reg f12))
        (ld (displacement (reg sp) -16) (reg f3))
        (fmulx (reg f4) (reg f12) (reg f16))
        (faddx (reg f8) (reg f16) (reg f20))
        (fitox (reg f3) (reg f24))
        (fdivx (reg f20) (reg f24) (reg f28))
        (fxtoi (reg f28) (reg f1)) % the quotient
        (fitox (reg f1) (reg f4))
        (fmulx (reg f4) (reg f24) (reg f8))
        (fsubx (reg f20) (reg f8) (reg f12))
        (fxtoi (reg f12) (reg f2)) % the remainder
        (st (reg f1) (displacement (reg sp) -4))
        (st (reg f2) (displacement (reg sp) -8))
        (ld (displacement (reg sp) -4) (reg 1))
        (ld (displacement (reg sp) -8) (reg 2))
        (*move (reg g6) (displacement (reg sp) -4))
        (*move (reg g6) (displacement (reg sp) -8))
        (*move (reg 5) (displacement (reg sp) -12))
        (*move (reg 3) (displacement (reg sp) -16))
        (*MOVE (reg 2) ($FLUID *second-value*))
        ))
)
(fluid '(*second-value*))

(compiletime (progn

 (defcmacro *wrshift3)

 (de *wrshift3 (a1 a2 a3)
   (prog (resultingcode*)
     (setq a1 (resolveoperand '(reg t1) a1))
     (setq a2 (resolveoperand '(reg t1) a2))
     (setq a3 (resolveoperand '(reg t1) a3))
     (return `((srl ,a2 ,a3 ,a1)))))
))

(lap '((*entry wwwquotientdouble expr 3)
              % adjusting the words first
        (*MOVE (reg 1) (reg 4))
        (*WSHIFT (reg 1) -2)
        (*WSHIFT (reg 4) 30)
        (*WOR (reg 2) (reg 4))
        (wry (reg 1) (reg 0) (reg 0))
          % now we can divide and spread the results
        (SDIV (reg 2) (reg 3) (reg 1)) % quotient is in REG 1
        (smul (reg 1) (reg 3) (reg 4))
        (*wdifference (reg 2) (reg 4))
        (*MOVE (reg 2) ($FLUID *second-value*))
        (*exit 0)))

(lap '((*entry wquotientdouble expr 3)
       (*alloc 6)
       (*move (reg 1) (frame 1)) % arg1
       (*move (reg 2) (frame 2)) % arg2
       (*move (reg 3) (frame 3)) % arg3
%       (*move (reg 3) (reg 2))
%       (*link wdivide expr 2)
%       (*move (reg 1) (frame 4))  % quot
%       (*move ($fluid *second-value*) (frame 5)) % rem
       (*move (quote 0)  (frame 4))
       (*move (reg 1) (frame 5))
       (*wshift (frame 2) 2)
                                   30 times!
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab1) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab1
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab2) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab2
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab3) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab3
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab4) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab4
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab5) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab5
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab6) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab6
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab7) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab7
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab8) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab8
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab9) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab9
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab10) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab10
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab11) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab11
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab12) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab12
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab13) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab13
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab14) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab14
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab15) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab15
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab16) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab16
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab17) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab17
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab18) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab18
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab19) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab19
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab20) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab20
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab21) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab21
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab22) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab22
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab23) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab23
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab24) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab24
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab25) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab25
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab26) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab26
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab27) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab27
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab28) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab28
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab29) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab29
       (*wshift (frame 4) 1)
       (*wrshift3 (reg 2) (frame 2) 31) % carrybit
       (*wshift (frame 2) 1)
       (*wshift (frame 5) 1)
       (*wplus2 (frame 5) (reg 2))
       (*jumpwlessp (label lab30) (frame 5) (frame 3))
       (*wdifference (frame 5) (frame 3))
       (*wplus2 (frame 4) 1)
    lab30                           %done
       (*move (frame 5) ($fluid *second-value*))
       (*move (frame 4) (reg 1))
       (*exit 6)))

(put 'addcarry 'opencode
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WPLUS2 (reg 1)(reg 2))
         %(*MOVE (reg 1)(reg 3)) (*WSHIFT (reg 3) -30)
	 (srl (reg 1) 30 (reg 3))
         (*MOVE (reg 3) ($FLUID carry*))
         (*WSHIFT (reg 1) 2)
         (*WSHIFT (reg 1) -2)))

(put 'subcarry 'opencode
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WDIFFERENCE (reg 1)(reg 2))
         %(*MOVE (reg 1) (reg 2)) (*WSHIFT (reg 2) -31)
         (srl (reg 1) 31 (reg 2))
         (*MOVE (reg 2) ($FLUID carry*))
         (*WSHIFT (reg 2) 30)
         (*WPLUS2 (reg 1) (reg 2))
       ))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   functions supporting word arithmetic with bbase*
%   bbase* is too large to be stored in a fluid cell;
%   so we perform operations with bbase* by shift and mask
%   instructions.

%   get the value for bbase* into a register
(put 'bbase** 'opencode
%     '((*MOVE 1 (reg 1))
%       (*WSHIFT (reg 1) 30)))
      '((sethi 16#100000 (reg 1))))

%   get the value for logicalbits* into a register
(put 'logicalbits** 'opencode
      '((*MOVE -1 (reg 1))
        (*WSHIFT (reg 1) -2)))

%   calculate (remainder x bbase*)
(put 'remainder-bbase 'opencode
      '((*WSHIFT (reg 1) 2)
        (*WSHIFT (reg 1) -2)))

%   calculate (quotient x bbase*)
(put 'quotient-bbase 'opencode
      '((*WSHIFT (reg 1) -30)))

(put 'bbase** 'destroys '((reg 1)))
(put 'logicalbits** 'destroys '((reg 1)))
(put 'remainder-bbase 'destroys '((reg 1)))
(put 'quotient-bbase 'destroys '((reg 1)))
