%
% NAME_SYSCALLS.RED - calls to the apollo naming server
%
% Author:      G. Q. Maguire Jr.
%              Computer Science Dept.
%              University of Utah
% Date:        2 November 1982
% Copyright (c) 1982 G. Q. Maguire Jr.
%
% 14-Mar-1984 JW Peterson, Fixed handling of return of status.
%
% 14-June-1983 Mark R. Swanson
% Added NAME_Get_Path for pathname utilities (and NMODE)

% ---------------------------------------------------------------------
% NAME related routines
% ---------------------------------------------------------------------

(de name_add_link (linkname linktext) 
% Create the LinkName with the value LinkText
% The Apollo uses textual substitution for links to allows versions of
% files to change.
(prog (linknamelength linktextlength status) 
  (setq linknamelength (stringlength linkname)) 
  (setq linktextlength (stringlength linktext)) 

  (putarg 1 linkname '(string))         % ARRAY of CHAR
  (putarg 2 linknamelength '(integer16)) 	% 16 bit integer name length
  (putarg 3 linktext '(string))         % ARRAY of CHAR
  (putarg 4 linktextlength '(integer16)) 	% 16 bit integer name length
  (cleararg 5) 

  (syscall 'name_$add_link 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return linkname))) (t 
    (return (apolloerror status))))))



(de name_create_directory (directoryname) 
% Create the directory DirectoryName
(prog (directorynamelength status) 
  (setq directorynamelength (stringlength directoryname)) 


  (putarg 1 directoryname '(string))    % ARRAY of CHAR
  (putarg 2 directorynamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$create_directory 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return directoryname))) (t 
    (return (apolloerror status))))))


(de name_create_file (filename) 
% Create the File FileName
(prog (filenamelength status) 
  (setq filenamelength (stringlength filename)) 


  (putarg 1 filename '(string))         % ARRAY of CHAR
  (putarg 2 filenamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$create_file 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return filename))) (t 
    (return (apolloerror status))))))


(de name_delete_directory (directoryname) 
% Delete the directory DirectoryName
(prog (directorynamelength status) 
  (setq directorynamelength (stringlength directoryname)) 


  (putarg 1 directoryname '(string))    % ARRAY of CHAR
  (putarg 2 directorynamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$delete_directory 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return directoryname))) (t 
    (return (apolloerror status))))))


(de name_delete_file (filename) 
% Delete the File FileName
(prog (filenamelength status) 
  (setq filenamelength (stringlength filename)) 


  (putarg 1 filename '(string))         % ARRAY of CHAR
  (putarg 2 filenamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$delete_file 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return filename))) (t 
    (return (apolloerror status))))))


% Symbolic Procedure Extract_Data_From_Directory(DirectoryName);
% Not Implemented YET

(de name_get_ndir nil 
% Returns the Naming directory as a LISP string.
(prog (blankstring length status pathname) 
  (setq blankstring (mkstring 256 32)) 

  (putarg 1 blankstring '(string))      % ARRAY [1..256] of CHAR
  (cleararg 2)                          % 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$get_ndir 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (setq length (getarg 2 '(integer16) '(inum))) 	% 16 bit integer
                                                        % name length
            (apolloresetstatus) 
           % convert to LISP string
            (setq pathname (getarg 1 (cons 'string length) '(string))) 

            (return pathname) nil)) (t 
         
     (return (apolloerror status))))))


(de name_set_ndir (pathname) 
% Accepts the Naming directory as a LISP string
% and set it to be the current anming directory.
(prog (pathnamelength length status) 
  (setq pathnamelength (stringlength pathname)) 

  (putarg 1 pathname '(string))         % ARRAY of CHAR
  (putarg 2 pathnamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$set_ndir 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return pathname))) (t 
    (return (apolloerror status))))))


(de name_get_path (pathname) 
% Accepts the Working directory as a LISP string
% and set it to be the current anming directory.
(prog (blankstring pathnamelength length status) 
  (setq pathnamelength (stringlength pathname)) 
  (setq blankstring (mkstring 256 32)) 

  (putarg 1 pathname '(string))         % ARRAY of CHAR
  (putarg 2 pathnamelength '(integer16)) 	% 16 bit integer name length
  (putarg 3 blankstring '(string))      % ARRAY [1..256] of CHAR
  (cleararg 4) 
  (cleararg 5) 

  (syscall 'name_$get_path 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (setq length (getarg 4 '(integer16) '(inum))) 	% 16 bit integer name length
           % convert to LISP string
            (apolloresetstatus) 
            (setq pathname (getarg 3 (cons 'string length) '(string))) 

            (return pathname) nil)) (t 
         
    (return (apolloerror status))))))



(de name_get_wdir nil 
% Returns the Working directory as a LISP string.
(prog (blankstring length status pathname) 
  (setq blankstring (mkstring 256 32)) 

  (putarg 1 blankstring '(string))      % ARRAY [1..256] of CHAR
  (cleararg 2)                          % 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$get_wdir 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (setq length (getarg 2 '(integer16) '(inum))) 	% 16 bit integer name length
           % convert to LISP string
            (apolloresetstatus) 
            (setq pathname (getarg 1 (cons 'string length) '(string))) 

            (return pathname) nil)) (t 
         
     (return (apolloerror status))))))


(de name_set_wdir (pathname) 
% Accepts the Working directory as a LISP string
% and set it to be the current anming directory.
(prog (pathnamelength length status) 
  (setq pathnamelength (stringlength pathname)) 

  (putarg 1 pathname '(string))         % ARRAY of CHAR
  (putarg 2 pathnamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$set_wdir 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return pathname))) (t 
    (return (apolloerror status))))))


(de name_drop_link (linkname) 
% Create the LinkName with the value LinkText
% The Apollo uses textual substitution for links to allows versions of
% files to change.
(prog (linknamelength linktextlength status) 
  (setq linknamelength (stringlength linkname)) 

  (putarg 1 linkname '(string))         % ARRAY of CHAR
  (putarg 2 linknamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 3) 

  (syscall 'name_$drop_link 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return linkname))) (t 
    (return (apolloerror status))))))


% Symbolic Procedure Read_Directory(DirectoryName);
% Not Implemented YET

(de name_read_link (linkname) 
% Read the LinkName and return the value LinkText
% The Apollo uses textual substitution for links to allows versions of
% files to change.
(prog (linknamelength linktextlength linktext status) 
  (setq linknamelength (stringlength linkname)) 

  (putarg 1 linkname '(string))         % ARRAY of CHAR
  (putarg 2 linknamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 3) 
  (cleararg 4) 
  (cleararg 5) 

  (syscall 'name_$read_link 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn 
            (setq linktextlength (getarg 4 '(integer16))) 	% 16 bit integer
                                                        % name length
            (apolloresetstatus) 
            (setq linktextlength (stringlength linktext)) 
            (setq linktext (getarg 3 (cons 'string linktextlength) '(string))) 
                                                        % ARRAY of CHAR
            (return linkname) nil)) (t 
         
    (return (apolloerror status))))))

