%
% STREAM_SYSCALLS.RED - Stream related system calls.
%
% Author:      G. Q. Maguire Jr.
%              Computer Science Dept.
%              University of Utah
% Date:        2 November 1982
% Copyright (c) 1982 G. Q. Maguire Jr.
%
% 5-Apr-87, Hucka.  Un-rlisp'ed and converted to 3.4.
% 14-MAR-84, JW-peterson, fixed handling of returned status

% ---------------------------------------------------------------------
% STREAM related routines
% ---------------------------------------------------------------------

(de stream_close (streamid) 
  (prog (status) 

	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(cleararg 2) 

	(syscall 'stream_$close 2) 

	(setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn
		 (apolloresetstatus)
		 (return status)))
	      (t (return (apolloerror status))))))


(de stream_create (pathname access concurrency) 
  (prog (pathnamelength streamid status) 

	(setq pathnamelength (stringlength pathname)) 

	(putarg 1 pathname '(string))         % ARRAY of CHAR
	(putarg 2 pathnamelength '(integer16)) 	% 16 bit integer name length
	(putarg 3 access '(integer16))        % Integer*2
	(putarg 4 concurrency '(integer16))   % Integer*2
	(cleararg 5) 
	(cleararg 6) 

	(syscall 'stream_$create 6) 

	(setq status (getarg 6 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn 
		 (apolloresetstatus) 
		 (setq streamid (getarg 5 '(integer16) '(inum))) 
		 (return streamid)))
	      (t (return (apolloerror status))))))

(de stream_create_here (pathname access concurrency locationname) 
  (prog (pathnamelength locationnamelength streamid status) 

	(setq pathnamelength (stringlength pathname)) 
	(setq locationnamelength (stringlength locationname)) 

	(putarg 1 pathname '(string))         % ARRAY of CHAR
	(putarg 2 pathnamelength '(integer16)) 	% 16 bit integer name length
	(putarg 3 access '(integer16))        % Integer*2
	(putarg 4 concurrency '(integer16))   % Integer*2

	(putarg 5 locationnamelength '(integer16)) % 16 bit integer name length
	(putarg 6 locationname '(string))     % ARRAY of CHAR
	(cleararg 7) 
	(cleararg 8) 

	(syscall 'stream_$create 8) 

	(setq status (getarg 8 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn 
		 (apolloresetstatus) 
		 (setq streamid (getarg 7 '(integer16) '(inum))) 
		 (return streamid)))
	      (t (return (apolloerror status))))))

(de stream_delete (streamid) 
  (prog (status) 

	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(cleararg 2) 

	(syscall 'stream_$delete 2) 

	(setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn
		 (apolloresetstatus)
		 (return status)))
	      (t (return (apolloerror status))))))

(de stream_get_conditional (streamid bufferpointer bufferlength) 
  (prog (returnpointer returnlength seek_key status) 
	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(putarg 2 bufferpointer '(integer32)) % 32 bit pointer
	(putarg 3 bufferlength '(integer32))  % 32 bit length
	(cleararg 4) 
	(cleararg 5) 
	(cleararg 6) 
	(cleararg 7) 

	(syscall 'stream_$get_conditional 7) 

	(setq status (getarg 7 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn
		 (apolloresetstatus) 
		 (setq returnpointer (getarg 4 '(integer32) '(inum))) 	%32 bit pointer
		 (setq returnlength (getarg 5 '(integer32) '(inum))) 	% 32 bit length
		 (setq seek_key (getarg 6 '(integer32) '(inum))) 
                                                  % 16 bit integer name length
		 (return (list returnpointer returnlength seek_key)) nil))
	      (t (return (apolloerror status))))))

(de stream_get_ec (streamid streamkey) 
  (prog (eventcounterpointer status) 
	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(putarg 2 streamkey '(integer32))     % 32 bit pointer
	(cleararg 3) 
	(cleararg 4) 

	(syscall 'stream_$get_ec 4) 
	
	(setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn
		 (apolloresetstatus) 
		 (setq eventcounterpointer
		       (getarg 3 '(integer16 . 3) '(integer16 . 3))) nil))
	      (t (return (apolloerror status))))))

(de stream_get_prior_rec (streamid bufferpointer bufferlength) 
  (prog (returnpointer returnlength seek_key status) 
	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(putarg 2 bufferpointer '(integer32)) % 32 bit pointer
	(putarg 3 bufferlength '(integer32))  % 32 bit length
	(cleararg 4) 
	(cleararg 5) 
	(cleararg 6) 
	(cleararg 7) 

	(syscall 'stream_$get_prior_rec 7) 

	(setq status (getarg 7 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn
		 (apolloresetstatus) 
		 (setq returnpointer (getarg 4 '(integer32) '(inum))) 	%32 bit pointer
		 (setq returnlength (getarg 5 '(integer32) '(inum))) 	% 32 bit length
		 (setq seek_key (getarg 6 '(integer32) '(inum)))           % 16 bit integer name length
						  
		 (return (list returnpointer returnlength seek_key)) nil))
	      (t (return (apolloerror status))))))

(de stream_get_rec (streamid bufferpointer bufferlength) 
  (prog (returnpointer returnlength seek_key status) 
	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(putarg 2 bufferpointer '(integer32)) % 32 bit pointer
	(putarg 3 bufferlength '(integer32))  % 32 bit length
	(cleararg 4) 
	(cleararg 5) 
	(cleararg 6) 
	(cleararg 7) 

	(syscall 'stream_$get_rec 7) 

	(setq status (getarg 7 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn
		 (apolloresetstatus) 
		 (setq returnpointer (getarg 4 '(integer32) '(inum))) 	%32 bit pointer
		 (setq returnlength (getarg 5 '(integer32) '(inum))) 	% 32 bit length
		 (setq seek_key (getarg 6 '(integer32) '(inum))) 
                                                  % 16 bit integer name length

                 (return (list returnpointer returnlength seek_key)) nil))
	      (t (return (apolloerror status))))))


% NOT IMPLEMENTED YET
%Symbolic Procedure Stream_Inquire(InputMask,
%                                  Inquiry_Type,
%                                  Attributes);
%
%begin scalar Error_Mask,
%             status;

%  SysCall('STREAM_!$INQUIRE,5);

%  Status := GetArg(5,  '(Integer32),     '(FixN));   % 32 bit status code

%  if Status eq 0
%    then ???
%    else return ApolloError(Status);
%end;


(de stream_open (pathname access concurrency) 
  (prog (pathnamelength streamid status) 

	(setq pathnamelength (stringlength pathname)) 

	(putarg 1 pathname '(string))         % ARRAY of CHAR
	(putarg 2 pathnamelength '(integer16)) 	% 16 bit integer name length
	(putarg 3 access '(integer16))        % Integer*2
	(putarg 4 concurrency '(integer16))   % Integer*2
	(cleararg 5) 
	(cleararg 6) 

	(syscall 'stream_$create 6) 

	(setq status (getarg 6 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn 
		 (apolloresetstatus) 
		 (setq streamid (getarg 5 '(integer16) '(inum))) 
		 (return streamid)))
	      (t (return (apolloerror status))))))


(de stream_put_chr (streamid bufferpointer bufferlength) 
  (prog (seek_key status) 

	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(putarg 2 bufferpointer '(integer32)) % 32 bit pointer
	(putarg 3 bufferlength '(integer32))  % 32 bit length
	(cleararg 4) 
	(cleararg 5) 

	(syscall 'stream_$put_chr 5) 

	(setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn 
		 (apolloresetstatus) 
		 (setq seek_key (getarg 4 '(integer32) '(inum))) 
                                                  % 16 bit integer name length

                 (return (list seek_key)) nil))
	      (t (return (apolloerror status))))))


(de stream_put_rec (streamid bufferpointer bufferlength) 
  (prog (seek_key status) 

	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(putarg 2 bufferpointer '(integer32)) % 32 bit pointer
	(putarg 3 bufferlength '(integer32))  % 32 bit length
	(cleararg 4) 
	(cleararg 5) 

	(syscall 'stream_$put_rec 5) 

	(setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn 
		 (apolloresetstatus) 
		 (setq seek_key (getarg 4 '(integer32) '(inum))) 
                                                  % 16 bit integer name length

                 (return (list seek_key)) nil))
	      (t (return (apolloerror status))))))


% NOT IMPLEMENTED YET
%Symbolic Procedure Stream_Redefine(InputMask,
%                                   Inquiry_Type,
%                                   Attributes);
%
%begin scalar Error_Mask,
%             status;

%  SysCall('STREAM_!$REDEFINE,5);

%  Status := GetArg(5,  '(Integer32),     '(FixN));   % 32 bit status code

%  if Status eq 0
%    then ???
%    else return ApolloError(Status);
%end;

(de stream_replace (streamid bufferpointer bufferlength) 
  (prog (seek_key status) 

	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(putarg 2 bufferpointer '(integer32)) % 32 bit pointer
	(putarg 3 bufferlength '(integer32))  % 32 bit length
	(cleararg 4) 
	(cleararg 5) 

	(syscall 'stream_$replace 5) 

	(setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn 
		 (apolloresetstatus) 
		 (setq seek_key (getarg 4 '(integer32) '(inum))) 
                                                  % 16 bit integer name length

                 (return (list seek_key)) nil))
	      (t (return (apolloerror status))))))


(de stream_seek (streamid searchbase searchtype seek_key) 
  (prog (status) 

	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(putarg 2 searchbase '(integer16))    % first seek param
	(putarg 3 searchtype '(integer16))    % second seek param
	(putarg 4 seek_key '(integer32))      % Seek pointer or signed offset
	(cleararg 5) 

	(syscall 'stream_$seek 5) 

	(setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn
		 (apolloresetstatus)
		 (return status)))
	      (t (return (apolloerror status))))))

(de stream_truncate (streamid) 
  (prog (status) 

	(putarg 1 streamid '(integer16))      % Stream_$ID_T
	(cleararg 2) 

	(syscall 'stream_$truncate 2) 

	(setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

	(cond ((eq status 0) 
	       (progn
		 (apolloresetstatus)
		 (return status)))
	      (t (return (apolloerror status))))))
