%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         double.sl
% Title:        double precision float arithmetic
%               mostly translated into LISP/lap from Cray's SCILIB routines
% Author:       Winfried Neun, ZIB Berlin
% Created:      20 December 1987
% Status:       Experimental
% Mode:         Lisp / lap
%
% (c) Copyright 1987, Konred-Zuse-Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% double's are double precision floatnums. They occupy two words of heap
% (plus headerword) in contrary to normal floats (with one word.
% This interface is a provsionary.
% The coercing and related stuff is under users reponsibility.
% they are flagged as floats and in some funtions they are equivalent,
% such as MINUSP, ZEROP, LESSP and GREATERP. Double can be printed like
% floats too.
 
% allocators
 
(de gtdouble ()
      (prog (x)
         (setq x (gtheap 3))
         (putmem x (mkitem hwords-tag 1))
         (return x)))
 
(compiletime
 (put 'mkdouble 'opencode '((*mkitem (Reg 1) floatnum-tag ))))
 
% conversion routines
 
(de float2double(fl)
   (prog (x)
    (setq x (gtdouble))
    (putmem (aplus2 x 1) (getmem (aplus2 fl 1)))
    (putmem (aplus2 x 2) 0)
    (return (mkdouble x))))
 
(de double2float (dou)
   (prog (x)
    (setq x (gtfltn))
    (putmem (aplus2 x 1) (getmem (aplus2 dou 1)))
    (return (mkfltn x))))
 
% double arithmetic (from Cray Research Inc.)
 
(lap '(
(*entry double-times2 expr 2)
(*alloc 3)
(*move (Reg 1) (Frame 1))
(*move (Reg 2) (Frame 2))
(*link gtdouble expr 0)
(*move (reg 1) (Frame 3))    % heap space for result
(*move (Frame 1) (Reg 1))
(*move (Frame 2) (Reg 2))
 
(*move (memory (reg 2) 1) (reg 3)) % adapt the registers
(*move (memory (reg 2) 2) (Reg 4)) % to call conventions
(*move (memory (reg 1) 2) (reg 2))
(*move (memory (reg 1) 1) (Reg 1))
 
(*link double-times2-intern expr 4)
 
(*Move (Frame 3) (Reg a7))    %store the result
(*move (reg 1) (memory (Reg a7) 1))
(*move (reg 2) (memory (Reg a7) 2))
(*move (frame 3) (reg 1))
(*link mkdouble expr 1)
(*EXIT 3)
))
 
(lap '(
(*entry double-times2-intern expr 2)
%        DMSS - Double precision floating point multiply
%
%        entry     (Reg 1) and (Reg 2) = 1st argument words 1 and 2
%                  (Reg 3) and (Reg 4) = 2nd argument words 1 and 2
%        exit      (Reg 1)            result
%
%        Algorithm and terminology:
%
%        Each of the input registers is broken up into two 24 bit
%        fields.  The exponent is stripped off and handled separately.
%        The 8 resulting 24 bit fields are all individually multiplied
%        using integer multiplication.  The results of the multiplies
%        are summed using the method outlined in the following table.
%        The results are then recombined and rounded to produce the
%        double precision product.
%
%        Operands:
%                 ((Reg 1) (Reg 2))        X0     X1     X2     X3
%                 ((Reg 3) (Reg 4))        Y0     Y1     Y2     Y3
%                         ______________________________
%
%                             X0*Y3  X1*Y3  X2*Y3  X3*Y3
%                      X0*Y2  X1*Y2  X2*Y2  X3*Y2
%               X0*Y1  X1*Y1  X2*Y1  X3*Y1
%        X0*Y0  X1*Y0  X2*Y0  X3*Y0
%        _______________________________________________
%        part7  part6  part5  part4  part3  part2  part1
%
%        :  24     24: :  24     24:    24     24     24
%
%        Result:
%
%        :    MSX    : :     LSX   :
%
%
% X0      Frame 1
% X1            2
% X2            3
% X3            4
% Y0            5
% Y1            6
% Y2            7
% Y3            8
%
(*alloc 8)
(fmul (Reg S7) (reg 1) (reg 3)) %  Force f.p. error if overflow argument
(rmask (Reg 5) 48)
(or (Reg S6) (Reg 1) (Reg 2))   %  Fold first argument to see if all zero
(and (Reg S0) (Reg 5) (Reg S6))
(or  (Reg S6) (Reg 3) (Reg 4))  %  Fold 2nd argument to see if all zero
(and (Reg S6) (Reg 5) (Reg S6))
(onepop (Reg A0) (Reg S6))
(or (Reg S6) (Reg 2) (Reg 4))   %  Combine second words for zero check
(JSZ   ZEROARG)                 %  If multiplier zero
(and (Reg S0) (Reg S6) (Reg 5)) %  Mask off top of second words
(JAZ   ZEROARG)                 %  If multiplicand zero
(rmask (Reg S7)  24)            %  Build mask to extract 24 bit chunks
(move (Reg 5) (Reg 1))
(move (Reg S6) (Reg 3))
(lshift (Reg S7) (Reg S7) 24)   %  Position to most significant part
(rshift (Reg 5) (Reg 5) 48)     %  Extract exponent of x
(rshift (Reg S6) (Reg S6) 48)   %  Extract exponent of y
(move  (Reg A2) (Reg 5))
(move  (Reg A5) (Reg S6))
(and (Reg 5) (Reg 1) (Reg S7))
(and (Reg S6) (Reg 3) (Reg S7))
(*move (reg 5) (Frame 1))        %  x0
(*move (reg S6) (Frame 5))       %  y0
(cmpand (Reg 5) (Reg 1) (Reg S7))
(cmpand (Reg S6) (Reg 3) (Reg S7))
(lshift (Reg 5) (Reg 5) 24)
(lshift (Reg S6) (Reg S6) 24)
(JSZ    BOTZERO)                % If bottom words are zero then skip
(*move (Reg 5) (Frame 2))       % x1
(*move (reg S6) (Frame 6))      % y1
(and (Reg 3) (Reg 2) (Reg S7))
(*move (reg 3) (Frame 3))       % x2
(cmpand (Reg 1) (Reg 2) (Reg S7))
(cmpand (Reg 2) (Reg 4) (Reg S7))
(lshift (Reg 1) (Reg 1) 24)
(lshift (Reg 2) (Reg 2) 24)
(*move (Reg 1) (Frame 4))       % x3
(*move (Reg 2) (Frame 8))       % y3
(and (Reg 4) (Reg 4) (Reg S7))
(*move (Reg 4) (Frame 7))       % y2
%
%        Begin computation of product by multiplying by parts
%
(fmul (Reg S7) (Reg 1) (Reg 2)) % part1 = x3*y3
(fmul (Reg S6) (Reg 1) (Reg S6))% x3*y1
(fmul (Reg 5) (Reg 5) (Reg 2))  % x1*y3
(fmul (Reg 1) (Reg 1) (Reg 4))  % x3*y2
(fmul (Reg 2) (Reg 3) (Reg 2))  % x2*y3
(fmul (Reg 3) (Reg 3) (Reg 4))  % x2*y2
(*move (Frame 1) (Reg 4))
(rshift (Reg S7) (Reg S7) 24)   % Shift part1 = x3*y3 factor for add
(add  (Reg 5) (Reg S6) (Reg 5)) % x3*y1+x1*y3
(*move (Frame 8) (Reg S6))
(add (Reg 1) (Reg 2) (Reg 1))   % part2 = x2*y3+x3*y2
(fmul (Reg 2) (Reg 4) (Reg S6)) % x0*y3
(add  (Reg S6) (Reg 5) (Reg 3)) % part3 = x3*y1+x1*y3+x2*y2
(*move (Frame 2) (Reg 3))
(add (Reg S7) (Reg 1) (Reg S7)) % part2a = part2 + shift(part1)
(*move (Frame 7) (Reg 4))
(fmul (Reg 3) (Reg 3) (Reg 4))  % x1*y2
(rshift (Reg S7) (Reg S7) 24)   % Shift part2a
(*move (Frame 3) (Reg 4))
(add (Reg S7) (Reg S6) (Reg S7))% part3a = part3 + shift(part2a)
(*move (Frame 6) (Reg 5))
(fmul (Reg 4) (Reg 4) (Reg 5))  % x2*y1
(rshift (Reg S7) (Reg S7) 24)   % Shift part3a
(add (Reg S6) (Reg 2) (Reg 3))  % x0*y3+x1*y2
(*move (Frame 4) (Reg 2))
(*move (Frame 5) (Reg 1))
(fmul (Reg 2) (Reg 2) (Reg 1))  % x3*y0
(add (Reg S6) (Reg S6) (Reg 4)) % x0*y3+x1*y2+x2*y1
(*move (Frame 3) (Reg 3))
(fmul (Reg 3) (Reg 3) (Reg 1))  % x2*y0
(*move (Frame 2) (Reg 4))
(fmul (Reg 5) (Reg 4) (Reg 1))  % x1*y0
(*move (Frame 6)(Reg 1))
(add (Reg S6) (Reg S6) (Reg 2)) % part4 = x0*y3+x1*y2+x2*y1+x3*y0
(fmul (Reg 4) (Reg 4) (Reg 1))  % x1*y1
(*move (Frame 1) (Reg 2))
(add (Reg S7) (Reg S6) (Reg S7))% part4a = part4 + shift(part3a)
(*move (Frame 7) (Reg 1))
(fmul (Reg 1) (Reg 2) (Reg 1))  % x0*y2
(*move (Frame 6) (Reg S6))
(fmul (Reg S6) (Reg 2) (Reg S6)) % x0*y1
(add  (Reg 4) (Reg 3) (Reg 4))   % x2*y0+x1*y1
(*move (Frame 5) (Reg 3))
(fmul (Reg 2) (Reg 2) (Reg 3))   % part7 = x0*y0
(rshift (Reg S7) (Reg S7) 22)    % Shift part4 two bits high
(move (Reg 3) 1)
(add (Reg S7) (Reg 3) (Reg S7))  % Round part4 one bit low
(add (Reg 4) (Reg 4) (Reg 1))    % part5 = x2*y0+x1*y1+x0*y2
(add (Reg 5) (Reg 5) (Reg S6))   % part6 = x1*y0+x0*y1
(and (Reg 1) (Reg S7) (Reg 3))   % Determine if carry during rounding
(rshift (Reg S7) (Reg S7) 1)     % Shift rnd(part4) one bit high
(*move (reg 1) (Frame 8))       % Save round bit for later normalization
(lshift (Reg 4) (Reg 4) 1)       % Shift part5 one bit high
(lshift (Reg 5) (Reg 5) 1)       % Shift part6 one bit higher
(add (Reg 4) (Reg 4) (Reg S7))   % part5a = part5 + shift(rnd(part4a))
(rmask  (Reg 1) 24)
(lshift (Reg 2) (Reg 2) 1)       % Shift part7 one bit higher
(and  (Reg S6) (Reg 1) (Reg 4))  % Extract lower part of part5a
(rshift (Reg 4) (Reg 4) 24)      % Shift part5a
(add (Reg A2) (Reg A2) (Reg A5)) % Compute exponent value
(add (Reg 4) (Reg 4) (Reg 5))    % part6a = part6 + shift(part5a)
(move (Reg A6) 8#40001)          % Get exponent offset value
(and (Reg 5) (Reg 4) (Reg 1))    % Save lower part of part6a
(rshift (Reg 4) (Reg 4) 24)      % Get top of part6a
(lshift (Reg 5) (Reg 5) 24)      % Position bottom of part6a
(add (Reg 1) (Reg 2) (Reg 4))    % Result word 1 = part7 + shift(part6a
(or (Reg 2) (Reg 5) (Reg S6))    % Result word 2 = bot(part6a)+part5a
(lmask (Reg 3) 16)               % Get mask to check for normalization
(and  (Reg S0) (Reg 1) (Reg 3))  % Check if normalization needed
(*move (Frame 8) (Reg 3))        % Recover bit from prior round
(add  (Reg 4) (Reg 2) (Reg 3))   % Round word 2 of result for normalize
(sub  (Reg A2) (Reg A2) (Reg A6))% Remove offset bias from exponent
(JSZ lab1)                  % $IF SN If normalization needed
(move  (Reg 2)  (Reg 4))
(rshift (Reg 4)  (Reg 4) 48)     % Check for overflow on round of word
(add  (Reg 1) (Reg 1) (Reg 4))   % Add possible overflow to word 1
(lshift (Reg 2) (Reg 2) 16)      % Prepare to normalize second word
(add  (Reg A2) (Reg A2) 1)       % Increment exponent
(drshift (Reg 2) (Reg 1) 1)      % Shift off bit in second word
(rshift  (Reg 1) (Reg 1) 1)      % Shift off bit in first word
(rshift  (Reg 2) (Reg 2) 16)     % Reposition second word
lab1                         % $ENDIF
(move  (Reg 3)  (Reg A2))        % Pick up the exponent
(move  (Reg 4)   8#20001)        % Get underflow exponent
(add   (Reg S0) (Reg 4) (Reg 3)) % Check if exponent overflows
(lshift (Reg 3) (Reg 3) 48)      % Position to exponent field
(lshift (Reg S0) (Reg S0) 49)    % Shift off sign for underflow check
(or (Reg 1) (Reg 1) (Reg 3))     % Stick in the exponent
(JSP   ZEROARG)                  % If result underflows then return zero
(*EXIT 8)
%
%        Shortened computation when the second words of the arguments
%        are both zero.  Multiplication is performed by multiplying
%        by parts.
%
BOTZERO
(fmul (Reg S7) (Reg 5) (Reg S6)) % part5 = x1*y1
(*move (Frame 1) (Reg 3))
(*move (Frame 5) (Reg 4))
(fmul (Reg S6) (Reg 3) (Reg S6)) % x0*y1
(fmul (Reg 5)  (Reg 5) (Reg 4))  % x1*y0
(fmul (Reg 4)  (Reg 3) (Reg 4))  % part7 = x0*y0
(rmask (Reg 1) 24)
(and  (Reg 2)  (Reg 1) (Reg S7)) % Extract lower part of part5a
(rshift (Reg S7) (Reg S7) 24)    % Shift part5a
(add (Reg A2) (Reg A2) (Reg A5)) % Compute exponent value
(add  (Reg 5) (Reg 5) (Reg S6))  % part6 = x1*y0+x0*y1
(move (Reg A6)  8#40000)         % Get exponent offset value
(add  (Reg S7) (Reg S7) (Reg 5)) % part6a = part6 + shift(part5a)
(lmask (Reg 3) 17)               % Get mask to check for normalization
(and  (Reg 5) (Reg S7) (Reg 1))  % Save lower part of part6a
(rshift (Reg S7) (Reg S7) 24)    % Get top of part6a
(lshift (Reg 5)  (Reg 5) 24)     % Position bottom of part6a
(add  (Reg 1)  (Reg 4) (Reg S7)) % Result word 1 = part7 + shift(part6a
(or  (Reg 2)  (Reg 5) (Reg 2))   % Result word 2 = bot(part6a)+part5a
(sub (Reg A2) (Reg A2) (Reg A6)) % Remove offset bias from exponent
(and  (Reg S0)  (Reg 1) (Reg 3)) % Check if normalization needed
(move (Reg 4)  8#20001)          % Get underflow exponent
(move (Reg 3) (Reg A2))          % Pick up the exponent
(JSN lab2)                   %  $IF SZ If normalization needed
(lshift (Reg 2) (Reg 2) 16)      % Prepare to normalize second word
(sub  (Reg A2)  (Reg A2) 1)      % Decrement exponent
(dlshift (Reg 1) (Reg 2) 1)      % Shift bit into first word
(lshift (Reg 2) (Reg 2) 1)       % Shift off bit in second word
(move  (Reg 3)  (Reg A2))        % Pick up new exponent
(rshift (Reg 2)  (Reg 2) 16)     % Reposition second word
lab2                          %  $ENDIF
(add  (Reg S0) (Reg 4) (Reg 3))  % Check if exponent overflows
(lshift (Reg 3) (Reg 3) 48)      % Position to exponent field
(lshift (Reg S0) (Reg S0) 49)    % Shift off sign for underflow check
(or   (Reg 1) (Reg 1) (Reg 3))   % Stick in the exponent
(JSP    ZEROARG )                % If result underflows then return zero
(*EXIT 8)
%
ZEROARG
(move (Reg 1) 0)  % return zero result
(move (Reg 2) 0)
(*EXIT 8)
))
%        IDENT     $DMSS
%        DMSS - Double precision floating point multiply
%
%        DMSS%   (call by value)
%        entry     (S1) and (S2) = 1st argument words 1 and 2
%                  (S3) and (S4) = 2nd argument words 1 and 2
%        exit      (S1)            result
%
%        Algorithm and terminology:
%
%        Each of the input registers is broken up into two 24 bit
%        fields.  The exponent is stripped off and handled separately.
%        The 8 resulting 24 bit fields are all individually multiplied
%        using integer multiplication.  The results of the multiplies
%        are summed using the method outlined in the following table.
%        The results are then recombined and rounded to produce the
%        double precision product.
%
%        Operands:
%                 (S1 S2)        X0     X1     X2     X3
%                 (S3 S4)        Y0     Y1     Y2     Y3
%                         ______________________________
%
%                             X0*Y3  X1*Y3  X2*Y3  X3*Y3
%                      X0*Y2  X1*Y2  X2*Y2  X3*Y2
%               X0*Y1  X1*Y1  X2*Y1  X3*Y1
%        X0*Y0  X1*Y0  X2*Y0  X3*Y0
%        _______________________________________________
%        part7  part6  part5  part4  part3  part2  part1
%
%        :  24     24: :  24     24:    24     24     24
%
%        Result:
%
%        :    MSX    : :     LSX   :
%
%
%X0       DEFT      TEMP    Frame 1
%X1       DEFT      TEMP          2
%X2       DEFT      TEMP          3
%X3       DEFT      TEMP          4
%Y0       DEFT      TEMP          5
%Y1       DEFT      TEMP          6
%Y2       DEFT      TEMP          7
%Y3       DEFT      TEMP          8
%
%DMSS%    ENTER     NP=2,MODE=BASELVL,TYPE=VALUE,ARGSIZE=TWOWORD,ALIGN=
%         S7        S1*FS3         Force f.p. error if overflow argumen
%         S5        <D'48
%         S6        S1]S2          Fold first argument to see if all ze
%         S0        S5&S6
%         S6        S3]S4          Fold 2nd argument to see if all zero
%         S6        S5&S6
%         A0        PS6
%         S6        S2]S4          Combine second words for zero check
%         JSZ       ZEROARG        If multiplier zero
%         S0        S6&S5          Mask off top of second words
%         JAZ       ZEROARG        If multiplicand zero
%         S7        <D'24          Build mask to extract 24 bit chunks
%         S5        S1
%         S6        S3
%         S7        S7<D'24        Position to most significant part
%         S5        S5>D'48        Extract exponent of x
%         S6        S6>D'48        Extract exponent of y
%         A2        S5
%         A3        S6
%         S5        S1&S7
%         S6        S3&S7
%         T.X0      S5             x0
%         T.Y0      S6             y0
%         S5        #S7&S1
%         S6        #S7&S3
%         S5        S5<D'24
%         S6        S6<D'24
%         JSZ       BOTZERO        If bottom words are zero then skip
%         T.X1      S5             x1
%         T.Y1      S6             y1
%         S3        S2&S7
%         T.X2      S3             x2
%         S1        #S7&S2
%         S2        #S7&S4
%         S1        S1<D'24
%         S2        S2<D'24
%         T.X3      S1             x3
%         T.Y3      S2             y3
%         S4        S4&S7
%         T.Y2      S4             y2
%
%        Begin computation of product by multiplying by parts
%
%         S7        S1*FS2         part1 = x3*y3
%         S6        S1*FS6         x3*y1
%         S5        S5*FS2         x1*y3
%         S1        S1*FS4         x3*y2
%         S2        S3*FS2         x2*y3
%         S3        S3*FS4         x2*y2
%         S4        T.X0
%         S7        S7>D'24        Shift part1 = x3*y3 factor for add
%         S5        S6+S5          x3*y1+x1*y3
%         S6        T.Y3
%         S1        S2+S1          part2 = x2*y3+x3*y2
%         S2        S4*FS6         x0*y3
%         S6        S5+S3          part3 = x3*y1+x1*y3+x2*y2
%         S3        T.X1
%         S7        S1+S7          part2a = part2 + shift(part1)
%         S4        T.Y2
%         S3        S3*FS4         x1*y2
%         S7        S7>D'24        Shift part2a
%         S4        T.X2
%         S7        S6+S7          part3a = part3 + shift(part2a)
%         S5        T.Y1
%         S4        S4*FS5         x2*y1
%         S7        S7>D'24        Shift part3a
%         S6        S2+S3          x0*y3+x1*y2
%         S2        T.X3
%         S1        T.Y0
%         S2        S2*FS1         x3*y0
%         S6        S6+S4          x0*y3+x1*y2+x2*y1
%         S3        T.X2
%         S3        S3*FS1         x2*y0
%         S4        T.X1
%         S5        S4*FS1         x1*y0
%         S1        T.Y1
%         S6        S6+S2          part4 = x0*y3+x1*y2+x2*y1+x3*y0
%         S4        S4*FS1         x1*y1
%         S2        T.X0
%         S7        S6+S7          part4a = part4 + shift(part3a)
%         S1        T.Y2
%         S1        S2*FS1         x0*y2
%         S6        T.Y1
%         S6        S2*FS6         x0*y1
%         S4        S3+S4          x2*y0+x1*y1
%         S3        T.Y0
%         S2        S2*FS3         part7 = x0*y0
%         S7        S7>D'22        Shift part4 two bits high
%         S3        1
%         S7        S3+S7          Round part4 one bit low
%         S4        S4+S1          part5 = x2*y0+x1*y1+x0*y2
%         S5        S5+S6          part6 = x1*y0+x0*y1
%         S1        S7&S3          Determine if carry during rounding
%         S7        S7>1           Shift rnd(part4) one bit high
%         T.Y3      S1             Save round bit for later normalizati
%         S4        S4<1           Shift part5 one bit high
%         S5        S5<1           Shift part6 one bit higher
%         S4        S4+S7          part5a = part5 + shift(rnd(part4a))
%         S1        <D'24
%         S2        S2<1           Shift part7 one bit higher
%         S6        S1&S4          Extract lower part of part5a
%         S4        S4>D'24        Shift part5a
%         A2        A2+A3          Compute exponent value
%         S4        S4+S5          part6a = part6 + shift(part5a)
%         A4        O'40001        Get exponent offset value
%         S5        S4&S1          Save lower part of part6a
%         S4        S4>D'24        Get top of part6a
%         S5        S5<D'24        Position bottom of part6a
%         S1        S2+S4          Result word 1 = part7 + shift(part6a
%         S2        S5]S6          Result word 2 = bot(part6a)+part5a
%         S3        >D'16          Get mask to check for normalization
%         S0        S1&S3          Check if normalization needed
%         S3        T.Y3           Recover bit from prior round
%         S4        S2+S3          Round word 2 of result for normalize
%         A2        A2-A4          Remove offset bias from exponent
%         $IF       SN             If normalization needed
%           S2        S4
%           S4        S4>D'48      Check for overflow on round of word
%           S1        S1+S4        Add possible overflow to word 1
%           S2        S2<D'16      Prepare to normalize second word
%           A2        A2+1         Increment exponent
%           S2        S1,S2>1      Shift off bit in second word
%           S1        S1>1         Shift off bit in first word
%           S2        S2>D'16      Reposition second word
%         $ENDIF
%         S3        A2             Pick up the exponent
%         S4        O'20001        Get underflow exponent
%         S0        S4+S3          Check if exponent overflows
%         S3        S3<D'48        Position to exponent field
%         S0        S0<D'49        Shift off sign for underflow check
%         S1        S1]S3          Stick in the exponent
%         JSP       ZEROARG        If result underflows then return zer
%         EXIT      MODE=BASELVL
%
%        Shortened computation when the second words of the arguments
%        are both zero.  Multiplication is performed by multiplying
%        by parts.
%
%BOTZERO  =         *
%         S7        S5*FS6         part5 = x1*y1
%         S3        T.X0
%         S4        T.Y0
%         S6        S3*FS6         x0*y1
%         S5        S5*FS4         x1*y0
%         S4        S3*FS4         part7 = x0*y0
%         S1        <D'24
%         S2        S1&S7          Extract lower part of part5a
%         S7        S7>D'24        Shift part5a
%         A2        A2+A3          Compute exponent value
%         S5        S5+S6          part6 = x1*y0+x0*y1
%         A4        O'40000        Get exponent offset value
%         S7        S7+S5          part6a = part6 + shift(part5a)
%         S3        >D'17          Get mask to check for normalization
%         S5        S7&S1          Save lower part of part6a
%         S7        S7>D'24        Get top of part6a
%         S5        S5<D'24        Position bottom of part6a
%         S1        S4+S7          Result word 1 = part7 + shift(part6a
%         S2        S5]S2          Result word 2 = bot(part6a)+part5a
%         A2        A2-A4          Remove offset bias from exponent
%         S0        S1&S3          Check if normalization needed
%         S4        O'20001        Get underflow exponent
%         S3        A2             Pick up the exponent
%         $IF       SZ             If normalization needed
%           S2        S2<D'16      Prepare to normalize second word
%           A2        A2-1         Decrement exponent
%           S1        S1,S2<1      Shift bit into first word
%           S2        S2<1         Shift off bit in second word
%           S3        A2           Pick up new exponent
%           S2        S2>D'16      Reposition second word
%         $ENDIF
%         S0        S4+S3          Check if exponent overflows
%         S3        S3<D'48        Position to exponent field
%         S0        S0<D'49        Shift off sign for underflow check
%         S1        S1]S3          Stick in the exponent
%         JSP       ZEROARG        If result underflows then return zer
%         EXIT      MODE=BASELVL
%
%ZEROARG  =         *
%         S1        0              return zero result
%         S2        0
%         EXIT      MODE=BASELVL
%
%         END
(lap '(
(*entry double-quotient expr 2)
(*alloc 3)
(*move (Reg 1) (Frame 1))
(*move (Reg 2) (Frame 2))
(*link gtdouble expr 0)
(*move (reg 1) (Frame 3))    % heap space for result
(*move (Frame 1) (Reg 1))
(*move (Frame 2) (Reg 2))
 
(*move (memory (reg 2) 1) (reg 3)) % adapt the registers
(*move (memory (reg 2) 2) (Reg 4)) % to call conventions
(*move (memory (reg 1) 2) (reg 2))
(*move (memory (reg 1) 1) (Reg 1))
 
(*link double-quotient-intern expr 4)
 
(*Move (Frame 3) (Reg a7))    %store the result
(*move (reg 1) (memory (Reg a7) 1))
(*move (reg 2) (memory (Reg a7) 2))
(*move (frame 3) (reg 1))
(*link mkdouble expr 1)
(*EXIT 3)
))
 
(lap '(
(*entry double-quotient-intern expr 4)
(*alloc 3)
 
%*   $DDSS - Double precision floating point divide
%
%    DDSS%   (call by value)
%    entry   (Reg S1) and (Reg S2) = numerator words 1 and 2
%            (Reg S3) and (Reg S4) = divisor words 1 and 2
%    exit  (Reg S1) quotient
%
%NUM1   Frame 1
%NUM2   2
%APPROX 3
%
(move (Reg 5) (Reg 4))      % Make copy of divisor word 2
(rshift (Reg 5) (Reg 5) 47) % Extract high order bit of second wor
(add (Reg 5)(Reg 3)(Reg 5)) % Round the first word of the divisor
(sbit (Reg S6))
(rshift (Reg S6) (Reg S6) 16)
(or (Reg 5) (Reg 5) (Reg S6)) % Add back norm bit in case of overflo
(hrecip (Reg S6) (Reg 5))     % Compute reciprocal approximation
(*move (Reg 1) (Frame 1))     % Save numerator word 1
(*move (Reg 2) (Frame 2))     % ...and word 2
(rmul (Reg S7) (Reg S6) (Reg 5))  % Multiply approx * denom
(rmul (Reg S7) (Reg S6) (Reg S7)) % Approx * ( approx * denom )
(fsub (Reg S7) (Reg S6) (Reg S7)) % Approx - approx * (approx * denom)
(fadd (Reg 1) (Reg S6) (Reg S7))  % 2 * approx - approx * (approx * deno
(move  (Reg 2) 0)
(*move (Reg 1) (Frame 3))         %Save approximation
(*link double-times2-intern expr 4) % Compute correction = divisor * app
%
%     Perform a second newton iteration in double precision
%
%*********************************************************************
%
%    The following code computes 2.0 - x.The method assumes
%    that x will always be 'close' to 1.0 where 'close' is defined
%    as within 32 bits of precision.This allows the algorithm
%    to effectively ignore the top 32 bits and to hold the
%    significant part of the factor in 64 bits.  The method is
%    based on the observation that 2.0 - x will simply be -x
%    with the exponent properly adjusted.  For values of x
%    exceeding 1.0 the exponent of 2.0 - x will be O'040000
%    and for values of x less than 1.0 the exponent of 2.0 - x
%    will be O'040001.  If x is greater than 1.0 than a left
%    shift of the mantissa is performed otherwise a right shift
%    is performed.  When the right shift is required then the
%    value is rounded before the shift is performed.
%
%*********************************************************************
%
(lshift (Reg 1) (Reg 1) 48)  % Get bottom 16 bits of first word
(lmask (Reg 4)  32)          % Build mask of all ones for middle
(or (Reg 1) (Reg 1) (Reg 2)) % Put bottom 16 bit of 1st word on 2nd
(rshift (Reg 4) (Reg 4) 16)  % Position mask to middle of word
(sub (Reg S0) (Reg S0) (Reg 1))
(sub (Reg 1) (Reg S0) (Reg 1)) % Negate 64 bits of correction factor
(sconst04 (Reg 3))             % Get exponent for values < 1.0
(or (Reg 3) (Reg 3) (Reg 4))   % Fill middle with all ones
(lshift (Reg 1) (Reg 1) 1)     % Position factor for exponent < 1.0
(jsm lab1)                   %   $IF  SP
(move (Reg 4) 2)
(add  (Reg 1) (Reg 1) (Reg 4)) % Round the correction factor
(sconst10  (Reg 3))            % Get exponent for factor => 1.0
(rshift  (Reg 1) (Reg 1) 2)    % Normalize the correction factor
lab1                         %   $ENDIF
(rmask (Reg 4) 48)
(and  (Reg 2) (Reg 1) (Reg 4)) % Extract lower 48 bits for 2nd word
(rshift (Reg 1) (Reg 1) 48)    % Get upper 16 bits for 1st word
(or (Reg 1) (Reg 1) (Reg 3))   % Tack on proper exponent for correcti
(*move (Frame 3) (Reg 3))      % Reload approx
(move  (Reg 4) 0)
(*link double-times2-intern expr 4) % Compute correction = divisor * app
(*move (Frame 1) (Reg 3))      %   Restore numerator word 1
(*move (Frame 2) (Reg 4))      % ...and word 2
(*link double-times2-intern expr 4)% Compute correction = divisor * app
(*EXIT 3)
))
 
%         IDENT     $DDSS
%         ENTRY     $DIV
%*       $DDSS -   Double precision floating point divide
%
%        DDSS%     (call by value)
%        entry     (S1) and (S2) = numerator words 1 and 2
%                  (S3) and (S4) = divisor words 1 and 2
%        exit      (S1) and (S2) = quotient words 1 and 2
%
%NUM1     DEFT
%NUM2     DEFT
%APPROX   DEFT
%
%DDSS%    ENTER     MODE=USER,TYPE=VALUE,ALIGN=ON
%$DIV     =         DDSS%          Old entry name
%         S5        S4             Make copy of divisor word 2
%         S5        S5>D'47        Extract high order bit of second wor
%         S5        S3+S5          Round the first word of the divisor
%         S6        SB
%         S6        S6>D'16
%         S5        S5]S6          Add back norm bit in case of overflo
%         S6        /HS5           Compute reciprocal approximation
%         T.NUM1    S1             Save numerator word 1
%         T.NUM2    S2             ...and word 2
%         S7        S6*RS5         Multiply approx * denom
%         S7        S6*RS7         Approx * ( approx * denom )
%         S7        S6-FS7         Approx - approx * (approx * denom)
%         S1        S6+FS7         2 * approx - approx * (approx * deno
%         S2        0
%         T.APPROX  S1             Save approximation
%         CALLV     DMSS%,,STKPTR=A7 Compute correction = divisor * app
%
%        Perform a second newton iteration in double precision
%
%*********************************************************************
%
%        The following code computes 2.0 - x.  The method assumes
%        that x will always be 'close' to 1.0 where 'close' is defined
%        as within 32 bits of precision.  This allows the algorithm
%        to effectively ignore the top 32 bits and to hold the
%        significant part of the factor in 64 bits.  The method is
%        based on the observation that 2.0 - x will simply be -x
%        with the exponent properly adjusted.  For values of x
%        exceeding 1.0 the exponent of 2.0 - x will be O'040000
%        and for values of x less than 1.0 the exponent of 2.0 - x
%        will be O'040001.  If x is greater than 1.0 than a left
%        shift of the mantissa is performed otherwise a right shift
%        is performed.  When the right shift is required then the
%        value is rounded before the shift is performed.
%
%*********************************************************************
%
%         S1        S1<D'48        Get bottom 16 bits of first word
%         S4        >D'32          Build mask of all ones for middle
%         S1        S1]S2          Put bottom 16 bit of 1st word on 2nd
%         S4        S4>D'16        Position mask to middle of word
%         S0        -S1
%         S1        -S1            Negate 64 bits of correction factor
%         S3        0.4            Get exponent for values < 1.0
%         S3        S3]S4          Fill middle with all ones
%         S1        S1<1           Position factor for exponent < 1.0
%         $IF       SP
%           S4        2
%           S1        S1+S4        Round the correction factor
%           S3        1.           Get exponent for factor => 1.0
%           S1        S1>2         Normalize the correction factor
%         $ENDIF
%         S4        <D'48
%         S2        S1&S4          Extract lower 48 bits for 2nd word
%         S1        S1>D'48        Get upper 16 bits for 1st word
%         S1        S1]S3          Tack on proper exponent for correcti
%         S3        T.APPROX       Reload approx
%         S4        0
%         CALLV     DMSS%,,STKPTR=A7 Compute approx * (2-divisor*approx
%         S3        T.NUM1         Restore numerator word 1
%         S4        T.NUM2         ...and word 2
%         CALLV     DMSS%,,STKPTR=A7 Compute quotient
%         EXIT      MODE=USER
%
%         END
 
(lap
'((*entry double-difference expr 2)
%  IDENT     $DASS
%*$DASS   - Double precision floating point add and subtract
%
%  DASS%     (call by value)
%  entry     ((Reg 1)) and ((Reg 2)) = 1st argument words 1 and 2
%            ((Reg 3)) and ((Reg 4)) = 2nd argument words 1 and 2
%  exit      (Reg 1) result
%
%  DSSS%     (call by value)
%  entry     ((Reg 1)) and ((Reg 2)) = 1st argument words 1 and 2
%            ((Reg 3)) and ((Reg 4)) = 2nd argument words 1 and 2
% exit       (Reg 1) result
%
(*alloc 3)
(*move (Reg 1) (Frame 1))
(*move (Reg 2) (Frame 2))
(*link gtdouble expr 0)
(*move (Reg 1) (Frame 3))  % heap space for result
(*move (Frame 1) (Reg 1))
(*move (Frame 2) (Reg 2))
 
                                   % adapt the registers
(*move (memory (Reg 2) 1) (Reg 3))
(xor (Reg 3) (Reg 3) (Reg S0))      % Change sign if subtraction
(*move (memory (reg 2) 2) (Reg 4)) % to call conventions
(*move (memory (reg 1) 2) (reg 2))
(*move (memory (reg 1) 1) (Reg 1))
(*link double-plus2-intern expr 4)
(*Move (Frame 3) (Reg a7))
(*move (reg 1) (memory (Reg a7) 1))
(*move (reg 2) (memory (Reg a7) 2))
(*move (frame 3) (reg 1))
(*link mkdouble expr 1)
(*EXIT 3)
))
 
(lap '(
(*entry double-plus2 expr 2)
(*alloc 3)
(*move (Reg 1) (Frame 1))
(*move (Reg 2) (Frame 2))
(*link gtdouble expr 0)
(*move (Reg 1) (Frame 3))  % heap space for result
(*move (Frame 1) (Reg 1))
(*move (Frame 2) (Reg 2))
 
(*move (memory (reg 2) 1) (reg 3)) % adapt the registers
(*move (memory (reg 2) 2) (Reg 4)) % to call conventions
(*move (memory (reg 1) 2) (reg 2))
(*move (memory (reg 1) 1) (Reg 1))
(*link double-plus2-intern expr 4)
(*Move (Frame 3) (Reg a7))
(*move (reg 1) (memory (Reg a7) 1))
(*move (reg 2) (memory (Reg a7) 2))
(*move (frame 3) (reg 1))
(*link mkdouble expr 1)
(*EXIT 3)
))
 
(lap '(
(*entry double-plus2-intern expr 2)
(*alloc 0)
 
% Determine larger of augend and addend
 
(cmpand (Reg 5)  (Reg 1) (Reg S0)) % abs(x)
(cmpand (Reg S6) (Reg 3) (Reg S0)) % abs(y)
(sub (Reg S0) (Reg S6) (Reg 5))
(fadd (Reg S7) (Reg 1) (Reg 3))    % Force f.p. error if overflow argument
(rshift (Reg 5) (Reg 5) 48)        % Extract exponents
(rshift (Reg S6) (Reg S6) 48)
(move (Reg A2) 8#60000)            % Set up value to check for overflow
%
% Put larger argument in (Reg 1) & (Reg 2) and smaller in (Reg 3) & (Reg 4).  The
% larger value will now be referred to as X and the smaller as
% Y.  The first word of X is X1 and the second is X2.
%
(JSM DASS2)            %  $IFSP    If addend word 1 => augend word 1
(JSN lab1)              %  $IFSZ   If addend word 1 = augend word 1
(sub (Reg S0) (Reg 2) (Reg 4))    % Compare 2nd words of addend and augend
(JSP DASS2)          % $JUMP  DAS2,SP   Skip if augend word 2 => addend word
lab1                    %  $ENDIF
(move (Reg S7) (Reg 3))     % Swap addend and augend
(move (Reg 3)  (Reg 1))
(move (Reg 1)  (Reg S7))
(move (Reg S7) (Reg 4))
(move (Reg 4)  (Reg 2))
(move (Reg 2)  (Reg S7))
(move (Reg S7) (Reg S6))
(move (Reg S6) (Reg 5))
(move (Reg 5)  (Reg S7))
%  $ENDIF
DASS2
%
% Determine the amount to shift Y in order to align coefficient
% for addition.
%
(xor (Reg S0) (Reg 1) (Reg 3))  % Check for signs of arguments difference
(move (Reg A5) (Reg 5))         % Get exponent of X
(move (Reg A6) (Reg S6))        % Get exponent of Y
(sub (Reg A7) (Reg A5) (Reg A6)) % Exponent difference
(sub (Reg A0) (Reg A5) (Reg A2)) % Check for arg exponent over O'60000
(rmask (Reg 5) 48)               % Mask to extract coefficient
(and   (Reg 3) (Reg 3) (Reg 5))  % Extract coefficient of Y1
(lshift (Reg 4) (Reg 4) 16)      % Move second word coefficient to top
(JAP OFLOW)                     % If exponent exceeds O'60000 then err
%
% Shift the smaller of the arguments (Y) to be aligned with
% the larger argument.
%
(drshift (Reg 4) (Reg 3) (Reg A7)) % Shift the second word of Y
(rshift  (Reg 3) (Reg 3) (Reg A7)) % Shift the first word
(rshift  (Reg 4) (Reg 4) 16)       % Reposition the second word
(and (Reg S7) (Reg 1) (Reg 5))     % Extract the coefficient of X
%
% If the signs of the arguments differ then negate Y.
%
(JSP lab2)     %      $IF   SM If signs differ
(sub (Reg S6) (Reg S0) (Reg 4)) % Negate second word of Y
(not (Reg 3)  (Reg 3))          % Complement the first word
(sub (Reg 4) (Reg S0) (Reg 4))
(xor (Reg S6) (Reg S6) (Reg S0))% Sign will be 0 if Y2<>0, 1 if Y2=0
(rshift (Reg S6) (Reg S6) 63)   % Extract a 1 if Y2 was zero
(and (Reg 4) (Reg 4) (Reg 5))   % Extract just coefficient of Y2
(add (Reg 3) (Reg 3) (Reg S6))  % Carry up into the first word
lab2           %      $ENDIF
%
% Perform the addition by doing integer addition of the aligned
% coefficients.
%
(add (Reg 4) (Reg 2) (Reg 4))   % Add coefficients of second words
(add (Reg S7) (Reg 3) (Reg S7)) % Add coefficients of first words
(and (Reg 2) (Reg 4) (Reg 5))   % Extract coefficient of second word sum
(rshift (Reg 4) (Reg 4) 48)     % Recover overflow of second word sum
(add (Reg A5) (Reg a5) 1)       % Increase exponent in case of overflow
(add (Reg S7) (Reg 4) (Reg S7)) % Add overflow of second word to first
(lshift (Reg 2) (Reg 2) 16)     % Move coefficient of second word to top
(and (Reg 1) (Reg 1) (Reg S0))  % Extract the sign of X
%
% Normalize the result of the addition by shifting off any
% leading zeroes.
%
(onepop (Reg A0) (Reg 2))       % Check for second word all zero
(move  (Reg S0) (Reg S7))       % Check if first word of result is zero
(zeropop (Reg A7) (Reg 2))      % Compute leading zeros in second word
(zeropop (Reg A2) (Reg S7))     % Compute shift count for normalization
(jsn  lab3)                %  $IFSZ  If first word is zero then
(JAZ  UFLOW)                    % If second word is zero then return zero
(add (Reg A2) (Reg A7) (Reg A2))% Add D'64 to normalization shift valu
lab3                       %  $ENDIF
(move (Reg A7) 15)
(sub (Reg A6) (Reg A2) (Reg A7)) % Determine how far to shift result
(move (Reg A2) 8#20000)          % Value of minimum allowable exponent
(dlshift (Reg S7)(Reg 2)(Reg A6))% Shift first word of result
(sub (Reg A5) (Reg A5) (Reg A6)) % Subtract amount of shift from expone
(lshift (Reg 2) (Reg 2)(Reg A6)) % Shift second word of result
(move  (Reg A6) 8#57777)         % Value of maximum allowable exponent
(sub (Reg A0) (Reg A5) (Reg A2)) % Check for exponent underflow
(drshift (Reg 2) (Reg S7) 1)     % Perform one shift of one more positi
(rshift  (Reg S7) (Reg S7) 1)    % to avoid shift of -1 on overflow
%
% Add the exponent and sign to the coefficient of the result
% and check for overflow or underflow.
%
(rshift (Reg 2) (Reg 2) 16)      % Position the second word of the result
(JAM    UFLOW)                   % If underflow then return a zero
(sub  (Reg A0) (Reg A6) (Reg A5))% Check for exponent overflow
(move  (Reg S6) (Reg A5))        % Recover the value of the exponent
(or (Reg 1) (Reg 1) (Reg S7))    % Merge sign and coefficient
(lshift  (Reg S6) (Reg S6) 48)   % Position the exponent
(JAM    OFLOW)                   % If overflow then return invalid value
(or (Reg 1) (Reg 1) (Reg S6))    % Insert exponent
(*Exit 0)
%
OFLOW
(move  (Reg 1) 8#600004)         % Return overflow result
(move  (Reg 2) 0)
(lshift (Reg 1) (Reg 1) 45)      % Position the overflow result
(add (Reg S0) (Reg S0) (Reg 1))  % Force f.p. error if overflow result
(*EXIT 3)
%
UFLOW
(move (Reg 2) 0)                 % Return zero result
(move (Reg 1) 0)                 % Return zero result
(*EXIT 3)
))
%         IDENT     $DASS
%*       $DASS   - Double precision floating point add and subtract
%
%        DASS%     (call by value)
%        entry     (S1) and (S2) = 1st argument words 1 and 2
%                  (S3) and (S4) = 2nd argument words 1 and 2
%        exit      (S1) and (S2) = result words 1 and 2
%
%        DSSS%     (call by value)
%        entry     (S1) and (S2) = 1st argument words 1 and 2
%                  (S3) and (S4) = 2nd argument words 1 and 2
%        exit      (S1) and (S2) = result words 1 and 2
%
%DSSS%    ENTER     MODE=BASELVL,TYPE=VALUE,ALIGN=ON
%$SUB     =         DSSS%          Old entry name
%         S3        S3\SB          Change sign if subtraction
%         J         DASS1          And use addition routine
%
%DASS%    ENTER     MODE=BASELVL,TYPE=VALUE,SHARED=DSSS%
%$ADD     =         DASS%          Old entry name
%DASS1    =         *
%
%        Determine larger of augend and addend
%
%         S5        #SB&S1         abs(x)
%         S6        #SB&S3         abs(y)
%         S0        S6-S5
%         S7        S1+FS3         Force f.p. error if overflow argumen
%         S5        S5>D'48        Extract exponents
%         S6        S6>D'48
%         A2        O'60000        Set up value to check for overflow
%
%        Put larger argument in S1 & S2 and smaller in S3 & S4.  The
%        larger value will now be referred to as X and the smaller as
%        Y.  The first word of X is X1 and the second is X2.
%
%         $IF       SP             If addend word 1 => augend word 1
%           $IF       SZ           If addend word 1 = augend word 1
%             S0        S2-S4      Compare 2nd words of addend and auge
%             $JUMP     DASS2,SP   Skip if augend word 2 => addend word
%           $ENDIF
%           S7        S3           Swap addend and augend
%           S3        S1
%           S1        S7
%           S7        S4
%           S4        S2
%           S2        S7
%           S7        S6
%           S6        S5
%           S5        S7
%         $ENDIF
%DASS2    =         *
%
%        Determine the amount to shift Y in order to align coefficient
%        for addition.
%
%         S0        S1\S3          Check for signs of arguments differe
%         A5        S5             Get exponent of X
%         A3        S6             Get exponent of Y
%         A4        A5-A3          Exponent difference
%         A0        A5-A2          Check for arg exponent over O'60000
%         S5        <D'48          Mask to extract coefficient
%         S3        S3&S5          Extract coefficient of Y1
%         S4        S4<D'16        Move second word coefficient to top
%         JAP       OFLOW          If exponent exceeds O'60000 then err
%
%        Shift the smaller of the arguments (Y) to be aligned with
%        the larger argument.
%
%         S4        S3,S4>A4       Shift the second word of Y
%         S3        S3>A4          Shift the first word
%         S4        S4>D'16        Reposition the second word
%         S7        S1&S5          Extract the coefficient of X
%
%        If the signs of the arguments differ then negate Y.
%
%         $IF       SM             If signs differ
%           S6        -S4          Negate second word of Y
%           S3        #S3          Complement the first word
%           S4        -S4
%           S6        S6\SB        Sign will be 0 if Y2<>0, 1 if Y2=0
%           S6        S6>D'63      Extract a 1 if Y2 was zero
%           S4        S4&S5        Extract just coefficient of Y2
%           S3        S3+S6        Carry up into the first word
%         $ENDIF
%
%        Perform the addition by doing integer addition of the aligned
%        coefficients.
%
%         S4        S2+S4          Add coefficients of second words
%         S7        S3+S7          Add coefficients of first words
%         S2        S4&S5          Extract coefficient of second word s
%         S4        S4>D'48        Recover overflow of second word sum
%         A5        A5+1           Increase exponent in case of overflo
%         S7        S4+S7          Add overflow of second word to first
%         S2        S2<D'16        Move coefficient of second word to t
%         S1        S1&SB          Extract the sign of X
%
%        Normalize the result of the addition by shifting off any
%        leading zeroes.
%
%         A0        PS2            Check for second word all zero
%         S0        S7             Check if first word of result is zer
%         A4        ZS2            Compute leading zeros in second word
%         A2        ZS7            Compute shift count for normalizatio
%         $IF       SZ             If first word is zero then
%           JAZ       UFLOW        If second word is zero then return z
%           A2        A4+A2        Add D'64 to normalization shift valu
%         $ENDIF
%         A4        D'15
%         A3        A2-A4          Determine how far to shift result
%         A2        O'20000        Value of minimum allowable exponent
%         S7        S7,S2<A3       Shift first word of result
%         A5        A5-A3          Subtract amount of shift from expone
%         S2        S2<A3          Shift second word of result
%         A3        O'57777        Value of maximum allowable exponent
%         A0        A5-A2          Check for exponent underflow
%         S2        S7,S2>1        Perform one shift of one more positi
%         S7        S7>1             to avoid shift of -1 on overflow
%
%        Add the exponent and sign to the coefficient of the result
%        and check for overflow or underflow.
%
%         S2        S2>D'16        Position the second word of the result
%         JAM       UFLOW          If underflow then return a zero
%         A0        A3-A5          Check for exponent overflow
%         S6        A5             Recover the value of the exponent
%         S1        S1]S7          Merge sign and coefficient
%         S6        S6<D'48        Position the exponent
%         JAM       OFLOW          If overflow then return invalid value
%         S1        S1]S6          Insert exponent
%         EXIT      MODE=BASELVL
%
%OFLOW    =         *
%         S1        O'600004       Return overflow result
%         S2        0
%         S1        S1<D'45        Position the overflow result
%         S0        +FS1           Force f.p. error if overflow result
%         EXIT      MODE=BASELVL
%
%UFLOW    =         *
%         S1        0              Return zero result
%         S2        0
%         EXIT      MODE=BASELVL
%
%         END
