%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PU:S-STRINGS.SL 
% Description:  Useful String Functions (with lots of error checking) 
% Author:       Alan Snyder 
% Created:      17 September 1982 
% Modified:     26 Jan 1984 1603-PST (Brian Beach) 
% Package:      NMODE Utilities 
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 26 Jan 1984 1603-PST (Brian Beach)
%  Corrected file name in header.
% 30 Nov 83 (Cris Perdue)
%   Added assign-op for string-fetch.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Defines the following functions:
%
% (string-fetch s i)
% (string-store s i ch)
% (string-length s)
% (string-upper-bound s)
% (string-empty? s)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de string-fetch (s i)
  (cond ((not (StringP s)) (NonStringError s 'String-Fetch))
	((not (FixP i)) (NonIntegerError i 'String-Fetch))
	(t (indx s i))
	))

(put 'string-fetch 'assign-op 'string-store)
% Apparently not setf-safe, at least not in the fast case. /csp

(de string-store (s i c)
  (cond ((not (StringP s)) (NonStringError s 'String-Store))
	((not (FixP i)) (NonIntegerError i 'String-Store))
	((not (FixP c)) (NonCharacterError c 'String-Store))
	(t (setindx s i c))
	))

(de string-length (s)
  (cond ((not (StringP s)) (NonStringError s 'String-Length))
	(t (Plus2 (size s) 1))
	))

(de string-upper-bound (s)
  (cond ((not (StringP s)) (NonStringError s 'String-Upper-Bound))
	(t (size s))
	))

(de string-empty? (s)
  (cond ((not (StringP s)) (NonStringError s 'String-Empty?))
	(t (EqN (size s) -1))
	))

% End of file.

